--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local Action = define ("Action")

Action.kStateInit = 0
Action.kStateRunning = 1
Action.kStatePaused = 2
Action.kStateCompleted = 3
Action.kStateCanceled = 4

function Action:Initialize(name, listener)
    self.name = name
    self.state = Action.kStateInit
    self.listener = listener
    self:_OnInitialize()
end

function Action:IsRunning()
    return self.state == Action.kStateRunning
end

function Action:IsCompleted()
    return self.state == Action.kStateCompleted
end

function Action:IsCanceled()
    return self.state == Action.kStateCanceled
end

function Action:IsPaused()
    return self.state == Action.kStatePaused
end

function Action:Cancel()
    self.request_cancel = true
end

function Action:Pause()
    self.request_pause = true
end

function Action:Resume()
    self.request_resume = true
end

function Action:Update(dt)
    if self.request_cancel and not self:IsCanceled() then
        self.state = Action.kStateCanceled
        if self.listener ~= nil and self.listener.OnCancel ~= nil then
            self.listener:OnCancel(self)
            self:Reset(Action.kStateCanceled)
        end
        self.request_cancel = false
    end

    if self.request_pause and not self:IsPaused() then
        self.state = Action.kStatePaused
        if self.listener ~= nil and self.listener.OnPause ~= nil then
            self.listener:OnPause(self)
        end
        self.request_pause = false
    end

    if self.request_resume and not self:IsRunning() then
        self.state = Action.kStateRunning
        if self.listener ~= nil and self.listener.OnResume ~= nil then
            self.listener:OnResume(self)
        end
        self.request_resume = false
    end

    if self:IsCanceled() or self:IsCompleted() or self:IsPaused() then
        return
    end

    if self.elapse_time == nil then
        self.elapse_time = 0
        if self.listener ~= nil and self.listener.OnBegin ~= nil then
            self.listener:OnBegin(self)
        end
        self.state = Action.kStateRunning
    else
        self.elapse_time = self.elapse_time + dt
    end

    if self:IsRunning() then
        self:_OnUpdate(dt)
    end
end

function Action:Reset(new_state)
    self:_OnReset(new_state)
    self.elapse_time = nil
    self.request_cancel = false
    self.request_pause = false
    self.request_resume = false
    self.state = new_state
end

function Action:_OnInitialize()
end

function Action:_OnUpdate(dt)
end

function Action:_OnReset(new_state)
end

local FixedTimeAction = define ("FixedTimeAction", Action)

function FixedTimeAction.Construct(name, duration, interpolator, listener)
    local new_action = FixedTimeAction:New()
    new_action:Initialize(name, duration, interpolator, listener)
    return new_action
end

function FixedTimeAction:Initialize(name, duration, interpolator, listener)
    CHECK(duration > 0, "FixedTimeAction: Construct: invalid duration:", duration, "!!!")

    Action.Initialize(self, name, listener)
    self.duration = duration
    self.interpolator = interpolator
end

function FixedTimeAction:_OnUpdate(dt)
    self.progress = self.elapse_time/self.duration

    if self.elapse_time >= self.duration then
        self.progress = 1.0
        self:Step(self.progress)

        if self.listener ~= nil and self.listener.OnEnd ~= nil then
            self.listener:OnEnd(self)
        end

        self:Reset(Action.kStateCompleted)
    else
        self:Step(self.progress)
    end
end

function FixedTimeAction:Step(progress)
    local interpolated_progress = progress
    if self.interpolator ~= nil then
        interpolated_progress = self.interpolator:interpolate(progress)
    end

    if self.listener ~= nil and self.listener.OnStep ~= nil then
        self.listener:OnStep(interpolated_progress)
    end

    self:_OnStep(interpolated_progress)
end

function FixedTimeAction:_OnStep(progress)
end

function FixedTimeAction:_OnReset(new_state)
    self.progress = 0.0
end

local RepeatAction = define ("RepeatAction", Action)

RepeatAction.kRepeatForever = 0

function RepeatAction.Construct(name, action, repeat_time, listener)
    local new_action = RepeatAction:New()
    new_action:Initialize(name, action, repeat_time, listener)

    return new_action
end

function RepeatAction:Initialize(name, action, repeat_time, listener)
    CHECK(repeat_time >= 0, "RepeatAction: Construct: invalid repeat time:", repeat_time, "!!!")

    Action.Initialize(self, name, action, listener)

    self.repeat_time = repeat_time
    self.inner_action = action
end

function RepeatAction:_OnUpdate(dt)
    if self.repeated_time == nil then
        self.repeated_time = 0
    end

    self.inner_action:Update(dt)

    if self.inner_action:IsCompleted() or self.inner_action:IsCanceled() then
        self.repeated_time = self.repeated_time + 1
        self.inner_action:Reset(Action.kStateInit)
        LOGD("RepeatAction: Action:", self.inner_action.name, "is repeated", self.repeated_time, "times ...")
    end

    if self.repeat_time ~= RepeatAction.kRepeatForever and self.repeated_time >= self.repeat_time then
        self:Reset(Action.kStateCompleted)
    end
end

function RepeatAction:_OnReset(new_state)
    self.repeated_time = nil
    self.inner_action:Reset(new_state)
end

local SequenceAction = define ("SequenceAction", Action)

function SequenceAction.Construct(name, actions, listener)
    local new_action = SequenceAction:New()
    new_action:Initialize(name, actions, listener)

    return new_action
end

function SequenceAction:Initialize(name, actions, listener)
    CHECK(#actions > 0, "SequenceAction: Construct: invalid action numbers:", #actions, "!!!")

    Action.Initialize(self, name, listener)

    self.inner_actions = actions
    self.inner_action_num = #self.inner_actions
    self.cur_action_pos = 1
    self.cur_action = self.inner_actions[self.cur_action_pos]
end

function SequenceAction:_OnUpdate(dt)
    self.cur_action:Update(dt)

    if self.cur_action:IsCompleted() or self.cur_action:IsCanceled() then
        self.cur_action_pos = self.cur_action_pos + 1
    end

    if self.cur_action_pos > self.inner_action_num then
        self:Reset(Action.kStateCompleted)
    else
        self.cur_action = self.inner_actions[self.cur_action_pos]
    end
end

function SequenceAction:_OnReset(new_state)
    self.cur_action_pos = 1
    self.inner_action_num = 0
    self.cur_action = nil
    for _, inner_action in ipairs(self.inner_actions) do
        inner_action:Reset(new_state)
    end
end

local ParallelAction = define ("ParallelAction", Action)

function ParallelAction.Construct(name, actions, listener)
    local new_action = ParallelAction:New()
    new_action:Initialize(name, actions, listener)

    return new_action
end

function ParallelAction:Initialize(name, actions, listener)
    CHECK(#actions > 0, "ParallelAction: Construct: invalid action numbers:", #actions, "!!!")

    Action.Initialize(self, name, listener)

    self.inner_actions = actions
    self.has_running_action = false
end

function ParallelAction:_OnUpdate(dt)
    self.has_running_action = false
    for _, inner_action in ipairs(self.inner_actions) do
        inner_action:Update(dt)
        if not inner_action:IsCompleted() and not inner_action:IsCanceled() then
            self.has_running_action = true
        end
    end

    if not self.has_running_action then
        self:Reset(Action.kStateCompleted)
    end
end

function ParallelAction:_OnReset(new_state)
    self.has_running_action = false

    for _, inner_action in ipairs(self.inner_actions) do
        inner_action:Reset(new_state)
    end
end

local DelayAction = define ("DelayAction", Action)

function DelayAction.Construct(name, action, delay_time, listener)
    local new_action = DelayAction:New()
    new_action:Initialize(name, action, delay_time, listener)

    return new_action
end

function DelayAction:Initialize(name, action, delay_time, listener)
    CHECK(action ~= nil, "DelayAction: Construct: action is nil !!!")

    Action.Initialize(self, name, listener)

    self.delay_time = delay_time
    self.inner_action = action
end

function DelayAction:_OnUpdate(dt)
    if self.delay_elapsed_time == nil then
        self.delay_elapsed_time = 0
    else
        self.delay_elapsed_time = self.delay_elapsed_time + dt
    end

    if self.delay_elapsed_time > self.delay_time then
        if self.delay_timeout == nil then
            self.delay_timeout = true
            self.inner_action:Update(self.delay_elapsed_time - self.delay_time)
            LOGD("DelayAction: delay time out, Action:", self.inner_action.name, "begin times ...")
        else
            self.inner_action:Update(dt)
        end
    end

    if self.inner_action:IsCompleted() or self.inner_action:IsCanceled() then
        self:Reset(Action.kStateCompleted)
    end
end

function DelayAction:_OnReset(new_state)
    self.delay_timeout = nil
    self.delay_elapsed_time = nil

    self.inner_action:Reset(new_state)
end

local ActorAction = define ("ActorAction", FixedTimeAction)

function ActorAction.Construct(name, duration, interpolator, listener, actor)
    local new_action = ActorAction:New()
    new_action:Initialize(name, duration, interpolator, listener, actor)
    return new_action
end

function ActorAction:Initialize(name, duration, interpolator, listener, actor)
    FixedTimeAction.Initialize(self, name, duration, interpolator, listener)

    CHECK(actor ~= nil, "ActorAction: Construct: actor is nil !!!")
    self.actor = actor
    self.transform = actor:GetTransform()
    self.initial_pos = self.transform:GetLocalPosition()
    self.initial_rotation = self.transform:GetLocalRotation()
    self.initial_scale = self.transform:GetLocalScale()
end

function ActorAction:Rotate(begin_rotation, end_rotation)
    self.begin_rotation = begin_rotation
    self.end_rotation = end_rotation
end

function ActorAction:Scale(begin_scale, end_scale)
    self.begin_scale = begin_scale
    self.begin_scale = end_scale
end

function ActorAction:Translate(begin_pos, end_pos)
    self.begin_pos = begin_pos
    self.end_pos = end_pos
end

function ActorAction:RotateSpeed(speed)
    self.rotate_speed = speed
end

function ActorAction:ScaleSpeed(speed)
    self.scale_speed = speed
end

function ActorAction:TranslateSpeed(speed)
    self.translate_speed = speed
end

function ActorAction:_OnStep(progress)
    if self.rotate_speed ~= nil then
        self.transform:SetLocalRotation(self.transform:GetLocalRotation()*(self.rotate_speed*self.elapse_time))
    elseif self.begin_rotation ~= nil then
        self.transform:SetLocalRotation(hl.Quat_Slerp(self.begin_rotation, self.end_rotation, progress))
    end

    if self.scale_speed ~= nil then
        self.transform:SetLocalScale(self.transform:GetLocalScale() + self.scale_speed*self.elapse_time)
    elseif self.begin_scale ~= nil then
        self.transform:SetLocalScale(self.begin_scale*(1.0 - progress) + self.end_scale*progress)
    end

    if self.translate_speed ~= nil then
        self.transform:SetLocalPosition(self.transform:GetLocalPosition() + self.translate_speed*self.elapse_time)
    elseif self.begin_pos ~= nil then
        self.transform:SetLocalPosition(self.begin_pos*(1.0 - progress) + self.end_pos*progress)
    end
end

function ActorAction:_OnReset(new_state)
    self.transform:SetLocalRotation(self.initial_rotation)
    self.transform:SetLocalScale(self.initial_scale)
    self.transform:SetLocalPosition(self.initial_pos)
    FixedTimeAction._OnReset(self, new_state)
end

local SpriteAction = define ("SpriteAction", ActorAction)

function SpriteAction.Construct(name, duration, interpolator, listener, actor, sprite)
    local new_action = SpriteAction:New()
    new_action:Initialize(name, duration, interpolator, listener, actor, sprite)
    return new_action
end

function SpriteAction:Initialize(name, duration, interpolator, listener, actor, sprite)
    ActorAction.Initialize(self, name, duration, interpolator, listener, actor)

    CHECK(sprite ~= nil, "SpriteAction: Initialize: sprite is nil !!!")
    self.sprite = sprite
    self.material = sprite:GetMaterial()
    self.frame_num = 1
    self.frame_row_num = 1
    self.frame_row_stride = 1.0/self.frame_row_num
    self.frame_column_num = 1
    self.frame_column_stride = 1.0/self.frame_column_num
    self.cur_frame_index = 0
    self.cur_frame_row_index = 0
    self.cur_frame_column_index = 0
end

function SpriteAction:ConfigureUVAnimation(frame_num, frame_row_num, frame_column_num)
    self.frame_num = frame_num
    self.frame_row_num = frame_row_num
    self.frame_column_num = frame_column_num
    self.frame_row_stride = 1.0/self.frame_row_num
    self.frame_column_stride = 1.0/self.frame_column_num
end

function SpriteAction:_OnStep(progress)
    if self.frame_num > 1 then
        self.cur_frame_index = math.floor(progress*self.frame_num)
        self.cur_frame_row_index = math.floor(self.cur_frame_index/self.frame_column_num)
        self.cur_frame_column_index = math.floor(self.cur_frame_index%self.frame_column_num)
        LOGD("Sprite1", self.cur_frame_index, self.cur_frame_row_index, self.cur_frame_column_index)
        self.material:SetDefaultSetting(hl.Vec4(self.cur_frame_column_index*self.frame_column_stride, self.cur_frame_row_index*self.frame_row_stride,
                self.frame_column_stride, self.frame_row_stride))
        LOGD("Sprite2", self.cur_frame_column_index*self.frame_column_stride,
                self.cur_frame_row_index*self.frame_row_stride,
                self.frame_column_stride,
                self.frame_row_stride)
    end
end

function SpriteAction:_OnReset(new_state)
    self.cur_frame_index = 0
    self.cur_frame_row_index = 0
    self.cur_frame_column_index = 0

    ActorAction._OnReset(self, new_state)
end

local ActionSystem = define ("ActionSystem")

function ActionSystem:Initialize()
    self.actions_map = {}
    self.ordered_actions = {}

    self.tmp_actions = {}
end

function ActionSystem:AddAction(new_action)
    if self.actions_map[new_action.name] ~= nil then
        CHECK(false, "Action: AddAction:", new_action.name, "already added !!!")
        return
    end

    new_action:Reset(Action.kStateInit)
    table.insert(self.ordered_actions, new_action)
    self.actions_map[new_action.name] = {action = new_action, pos = #self.ordered_actions}

    LOGD("ActionSystem: AddAction: add action:", new_action.name)
end

function ActionSystem:RemoveAction(removed_action)
    local action_info = self.actions_map[removed_action.name]

    if action_info == nil or action_info.action ~= removed_action then
        CHECK(false, "Action: RemoveAction:", removed_action.name, "is not added before !!!")
        return
    end

    table.remove(self.ordered_actions, action_info.pos)
    self.actions_map[removed_action.name] = nil

    LOGD("ActionSystem: RemoveAction: remove action:", removed_action.name)
end

function ActionSystem:Update(dt)
    for _, action in ipairs(self.ordered_actions) do
        action:Update(dt)

        if action:IsCanceled() or action:IsCompleted() then
            table.insert(self.tmp_actions, action)
        end
    end

    for _, stopped_action in ipairs(self.tmp_actions) do
        self:RemoveAction(stopped_action)
    end

    self.tmp_actions = {}
end