--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local Service = import "Service"

local FaceReshapeService = define ("FaceReshapeService", Service)

LOGD("FaceReshapeService has been defined ..., parent is", Service)

function FaceReshapeService:_OnInitialize()
    self.service_type = Service.SERVICE_FACE_RESHAPE
    self.input_texture = nil
    self.output_texture = nil
    self.external_output_texture = nil

    self.intensity = 1.0
    self.face_distorter = hl.FaceDistortion_New()
    self.face_distorter_inited = false

    self.configure_dirty = false

    self.frame_view_id_list = hl.U16Array()

    hl.Class_Retain(self.face_distorter)
end

function FaceReshapeService:_Prepare()
    if self.configure_dirty == false then
        return
    end

    if self.face_distorter_inited == false then
        self.face_distorter:InitWithSetting(self.input_texture,
                self.external_output_texture or self.output_texture, self.intensity)
        self.face_distorter_inited = true
    else
        self.face_distorter:SetInput(self.input_texture)
        self.face_distorter:SetOutput(self.external_output_texture or self.output_texture)
        self.face_distorter:SetIntensity(self.intensity)
    end

    self.configure_dirty = false
end

function FaceReshapeService:_OnStop()
    self:EnableService(Service.SERVICE_FACE_TRACK, false)
end

function FaceReshapeService:_OnStart()
    self:EnableService(Service.SERVICE_FACE_TRACK, true)
end

function FaceReshapeService:_OnUpdate(dt)
    self:_Prepare()

    local face_infos = self.attach_node.node_context.node_manager:GetService(
            Service.SERVICE_FACE_TRACK):GetFaceInfos()

    local face_num = 0

    if face_infos ~= nil then
        face_num = face_infos.face_num
    end

    self.face_distorter:UpdateFaceCount(face_num);

    for i, face_id in ipairs(face_infos.valid_face_ids) do
        self.face_distorter:UpdateFromFaceInfo(i, face_infos.faces[face_id].original_face_data)
    end

    if self.face_distorter:PrepareRender() then
        local count = self.face_distorter:GetRenderViewCount()
        self.frame_view_id_list:clear()
        for k = 1, count do
            self.frame_view_id_list:push_back(hl.GPRenderDrive_AllocFrameView())
        end
        self.face_distorter:Render(self.frame_view_id_list)
    end
end

function FaceReshapeService:_OnNodeEvent(msg, message_source)
    local param_num = msg:PopInt32()

    CHECK(param_num > 0, "FaceReshapeService: _OnNodeEvent: param num:", param_num, "<= 0 !!!")

    local param_key = nil
    local param_value = nil
    for i = 1, param_num do
        param_key = msg:PopInt32()
        if param_key == hl.Hl_MSG_SCRIPT_SERVICE_FACE_RESHAPE_INTENSITY then
            param_value = msg:PopFloat()
            self.intensity = param_value
            LOGD("FaceReshapeService: _OnNodeEvent: intensity ->", self.intensity)
        else
            CHECK(false, "FaceReshapeService: _OnNodeEvent: unknown param key:", param_key, "!!!")
        end
    end

    self.configure_dirty = true
end

function FaceReshapeService:SetInputTexture(texture)
    CHECK(texture ~= nil, "FaceReshapeService: SetInputTexture: input texture is nil !!!")
    self.input_texture = texture
    if self.output_texture == nil or
            self.output_texture:GetWidth() ~= texture:GetWidth() or
            self.output_texture:GetHeight() ~= texture:GetHeight() or
            self.output_texture:GetType() ~= texture:GetType() or
            self.output_texture:GetFormat() ~= texture:GetFormat() then

        if self.output_texture ~= nil then
            hl.Class_Release(self.output_texture)
        end

        self.output_texture = hl.RenderTexture_New()
        self.output_texture:SetWidth(self.input_texture:GetWidth())
        self.output_texture:SetHeight(self.input_texture:GetHeight())
        self.output_texture:SetType(self.input_texture:GetType())
        self.output_texture:SetFlags(hl.GPTextureFlag_kDefault2DRT)
        self.output_texture:SetFormat(self.input_texture:GetFormat())
        hl.Class_Retain(self.output_texture)
    end

    self.configure_dirty = true
end

function FaceReshapeService:SetExternalOutputTexture(texture)
    self.external_output_texture = texture

    self.configure_dirty = true
end

function FaceReshapeService:_OnShutdown()
    if self.output_texture ~= nil then
        hl.Class_Release(self.output_texture)
        self.output_texture = nil
    end

    hl.Class_Release(self.face_distorter)
    self.face_distorter_inited = false
    self.face_distorter = nil
end

function FaceReshapeService:GetOutputTexture()
    return self.external_output_texture or self.output_texture
end

return FaceReshapeService