--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local GameFramework = import "GameFramework"
local GamePlayRecord = import "GamePlayRecord"
local InstanceCall = import "InstanceCall"
local Service = import "Service"

local FaceTracker = define ("FaceTracker")

FaceTracker.kNone = 0
FaceTracker.kEngineFaceTrackerType = 1
FaceTracker.kRecordFaceTrackerType = 2

FaceTracker.kMouthScaleMin = 0.15
FaceTracker.kMouthScaleMax = 0.4
FaceTracker.kMouthScaleStride = FaceTracker.kMouthScaleMax - FaceTracker.kMouthScaleMin
FaceTracker.kEyeScaleMin = 0.03
FaceTracker.kEyeScaleMax = 0.05
FaceTracker.kEyeScaleStride = FaceTracker.kEyeScaleMax - FaceTracker.kEyeScaleMin

local function GetRatio(begin_value, end_value, cur_value, value_stride)
    if cur_value <= begin_value then
        return 0.0
    elseif cur_value >= end_value then
        return 1.0
    else
        return (cur_value - begin_value)/value_stride
    end
end

function FaceTracker:Initialize(name, node)
    self.initialized = true
    self.name = name
    self.attach_node = node
    self.stopped = true
    self.type = FaceTracker.kNone
    self.face_infos = {faces = {}, face_num = 0, valid_face_ids = {}}
    self.active = false

    self:InvalidateFaceInfos()

    LOGD("FaceTracker:", self.name, "Initialize")
    self:_OnInitialize()
end

function FaceTracker:Active(active)
    self.active = active
    self:_OnActive(active)
end

function FaceTracker:IsActive()
    return self.active
end

function FaceTracker:GetFaceInfos()
    return self.face_infos
end

function FaceTracker:InvalidateFaceInfos()
    for _, face in pairs(self.face_infos.faces) do
        face.can_smooth = false
        face.appear = false
    end
    self.face_infos.face_num = 0
    self.face_infos.valid_face_ids = {}
end

function FaceTracker:Start()
    if false == self.stopped then
        return
    end

    local start_ok = self:_OnStart()

    if false == start_ok then
        CHECK(false, "FaceTracker", self.name, "start fail !!!")
        return
    end

    LOGD("FaceTracker", self.name, "start ...")
    self.stopped = false
end

function FaceTracker:Stop()
    if self.stopped then
        return
    end

    self:_OnStop()
    self:InvalidateFaceInfos()
    LOGD("FaceTracker", self.name, "stop ...")
    self.stopped = true
end

function FaceTracker:Detect()
    local detect_success = self:_OnDetect()

    if not detect_success then
        return false
    end

    local mouth_scale = 1.0
    local eye_scale = 1.0

    local face_width = 0
    local mouth_height = 0
    local eye_height = 0

    for _, face in pairs(self.face_infos.faces) do
        if face.appear then
            face_width = hl.FaceTracker_GetFaceWidth(face.original_face_data)
            if face_width <= 0.001 then
                face.mouth_scale = 0.0
                face.eye_scale = 0.0
            else
                mouth_height = hl.FaceTracker_GetMouthHeight(face.original_face_data)
                eye_height = hl.FaceTracker_GetEyeHeight(face.original_face_data)

                mouth_scale = mouth_height/face_width

                mouth_scale = GetRatio(FaceTracker.kMouthScaleMin, FaceTracker.kMouthScaleMax,
                        mouth_scale, FaceTracker.kMouthScaleStride)

                eye_scale = eye_height/face_width

                eye_scale = GetRatio(FaceTracker.kEyeScaleMin, FaceTracker.kEyeScaleMax,
                        eye_scale, FaceTracker.kEyeScaleStride)

                face.mouth_scale = mouth_scale
                face.eye_scale = eye_scale
            end
        end
    end

    return true
end

function FaceTracker:GetType()
    return self.type
end

function FaceTracker:Shutdown()
    self:_OnShutdown()

    self.initialized = false
    self.face_infos = nil
end

function FaceTracker:_OnInitialize()
end

function FaceTracker:_OnStart()
end

function FaceTracker:_OnActive(active)
end

function FaceTracker:_OnStop()
end

function FaceTracker:_OnDetect()
    return true
end

function FaceTracker:_OnShutdown()
end

local EngineFaceTracker = define ("EngineFaceTracker", FaceTracker)

function EngineFaceTracker:_OnInitialize()
    self.type = FaceTracker.kEngineFaceTrackerType

    self.inner_face_tracker = hl.FaceTracker_New()
    self.inner_face_tracker:SetTrackingTexture(self.attach_node.node_context.node_manager:GetFixedTexture(
            self.attach_node.node_context.node_manager.INPUT_TEXTURE_0, true))
    --self.inner_face_tracker:SetTrackingFlags(hl.FaceTracker_kFacePoint | hl.FaceTracker_kHeadYaw)
    self.inner_face_tracker:SetTrackingMode(hl.FaceTracker_kVideo)
end

function EngineFaceTracker:EnablePose3D(enable)
    LOGD("FaceTrackService: EngineFaceTracker: enable pose 3d:", enable)
    self.inner_face_tracker:SetEnableFace3DPoseEstimation(enable)
end

function EngineFaceTracker:_OnStart()
    return self.inner_face_tracker:Start()
end

function EngineFaceTracker:_OnStop()
    self.inner_face_tracker:Stop()
end

local kFaceTrackerRotateScaleSmoothParam = hl.Vec4(0.2, 0.2, 0.2, 0.1)
function EngineFaceTracker:_OnDetect()
    local detect_success = self.inner_face_tracker:Update()
    if false == detect_success then
        self:InvalidateFaceInfos()
        return false
    end

    local face_num = self.inner_face_tracker:GetFaceCount()
    if face_num == 0 then
        self:InvalidateFaceInfos()
        return true
    end

    self.face_infos.face_num = face_num
    self.face_infos.valid_face_ids = {}

    for _, face in pairs(self.face_infos.faces) do
        face.appear = false
        face.original_face_data = nil
    end

    local cur_face_data
    local face_id
    local cur_face
    for i = 0, face_num - 1 do
        cur_face_data = self.inner_face_tracker:GetFace(i)
        face_id = cur_face_data.face_id

        table.insert(self.face_infos.valid_face_ids, face_id)

        cur_face = self.face_infos.faces[face_id]

        if cur_face == nil then
            cur_face = {can_smooth = false, appear = true}
            self.face_infos.faces[face_id] = cur_face
        else
            cur_face.appear = true
        end
        cur_face.original_face_data = cur_face_data

        local rot_quat = cur_face_data.rot2d
        local scale = self.inner_face_tracker:GetFaceScale(i)

        local new_rotation_scale = hl.Vec4(rot_quat:EulerAngle(), scale)

        if cur_face.can_smooth then
            cur_face.smooth_rotation_scale = hl.FaceTracker_SmoothRotationAndUnifScale(
                    cur_face.smooth_rotation_scale, new_rotation_scale, kFaceTrackerRotateScaleSmoothParam)
            cur_face.smooth_scale = cur_face.smooth_rotation_scale.w
            cur_face.smooth_rotation = hl.Quat(cur_face.smooth_rotation_scale:CastToVec3())
        else
            cur_face.can_smooth = true
            cur_face.smooth_rotation_scale = new_rotation_scale
            cur_face.smooth_scale = scale
            cur_face.smooth_rotation = rot_quat
        end
    end

    for _, face in pairs(self.face_infos.faces) do
        if false == face.appear then
            face.can_smooth = false
        end
    end

    cur_face_data = nil
    face_id = nil
    cur_face = nil

    return true
end

function EngineFaceTracker:_OnShutdown()
    hl.Class_Discard(self.inner_face_tracker)
    self.inner_face_tracker = nil
end

local function ReadFaceInfo(face_infos, record_message)
    face_infos.face_num = record_message:PopUint32()

    if face_infos.face_num == 0 then
        return
    end

    for _, face in pairs(face_infos.faces) do
        face.appear = false
    end

    local face_id
    local cur_face
    face_infos.valid_face_ids = {}

    for i = 1, face_infos.face_num do
        table.insert(face_infos.valid_face_ids, record_message:PopInt32())
    end

    for i = 1, face_infos.face_num do
        face_id = record_message:PopInt32()
        cur_face = face_infos.faces[face_id]
        if cur_face == nil then
            face_infos.faces[face_id] = {original_face_data = hl.FaceInfo()}
            cur_face = face_infos.faces[face_id]
        end

        cur_face.original_face_data:ReadFromMessage(record_message)
        cur_face.smooth_scale = record_message:PopFloat()
        cur_face.smooth_rotation = record_message:PopQuat()
        cur_face.appear = true
    end
    cur_face = nil
end

local function WriteFaceInfo(face_infos, record_message)
    record_message:PushUint32(face_infos.face_num)

    if face_infos.face_num == 0 then
        return
    end

    for _, valid_face_id in ipairs(face_infos.valid_face_ids) do
        record_message:PushInt32(valid_face_id)
    end

    for face_id, face in pairs(face_infos.faces) do
        if face.appear then
            record_message:PushInt32(face_id)
            face.original_face_data:WriteToMessage(record_message)
            record_message:PushFloat(face.smooth_scale)
            record_message:PushQuat(face.smooth_rotation)
        end
    end
end

local RecordPlayFaceTracker = define ("RecordPlayFaceTracker", FaceTracker)

local FaceTrackerManager = define ("FaceTrackerManager")

function FaceTrackerManager:Initialize(attach_node)
    self.engine_face_tracker = nil
    self.record_play_face_tracker = nil

    self.attach_node = attach_node

    local this = self

    self.merge = false
    self.merged_face_infos = nil

    self.pose3DSwitch = InstanceCall:New()

    self.pose3DSwitch:Initialize(function ()
        this:_EnablePose3DImpl(true)
    end,
            function ()
                this:_EnablePose3DImpl(false)
            end)
end

function FaceTrackerManager:RefreshFaceTracker(state)
    if state == GamePlayRecord.kState_Preview or
            state == GamePlayRecord.kState_Recording then
        if self.engine_face_tracker == nil then
            LOGD("FaceTrackerManager: new state(Preview/Recording):", state, "need create engine face tracker")
            self.engine_face_tracker = EngineFaceTracker:New()
            self.engine_face_tracker:Initialize("EngineFaceTracker", self.attach_node)
            if self.pose3DSwitch:GetRefCount() > 0 then
                LOGD("FaceTrackerManager: new state(Preview/Recording):", state, "new engine face tracker created, need enable pose 3d")
                self.engine_face_tracker:EnablePose3D(true)
            end
        end
        self.engine_face_tracker:Active(true)

        if self.record_play_face_tracker ~= nil then
            self.record_play_face_tracker:Active(false)
            self.record_play_face_tracker:Stop()
        end
    elseif state == GamePlayRecord.kState_Playing then
        if self.record_play_face_tracker == nil then
            LOGD("FaceTrackerManager: new state(Playing):", state, "need create record play face tracker")
            self.record_play_face_tracker = RecordPlayFaceTracker:New()
            self.record_play_face_tracker:Initialize("RecordPlayFaceTracker", self.attach_node)
        end
        self.record_play_face_tracker:Active(true)

        if self.engine_face_tracker ~= nil then
            self.engine_face_tracker:Active(false)
            self.engine_face_tracker:Stop()
        end
    elseif state == GamePlayRecord.kState_PlayRecord then
        if self.engine_face_tracker == nil then
            LOGD("FaceTrackerManager: new state(PlayRecord):", state, "need create engine face tracker")
            self.engine_face_tracker = EngineFaceTracker:New()

            self.engine_face_tracker:Initialize("EngineFaceTracker", self.attach_node)
            if self.pose3DSwitch:GetRefCount() > 0 then
                LOGD("FaceTrackerManager: new state(Preview/Recording):", state, "new engine face tracker created, need enable pose 3d")
                self.engine_face_tracker:EnablePose3D(true)
            end
        end

        if self.record_play_face_tracker == nil then
            LOGD("FaceTrackerManager: new state(PlayRecord):", state, "need create record play face tracker")
            self.record_play_face_tracker = RecordPlayFaceTracker:New()
            self.record_play_face_tracker:Initialize("RecordPlayFaceTracker", self.attach_node)
        end

        self.engine_face_tracker:Active(true)
        self.record_play_face_tracker:Active(true)
    else
        CHECK(false, "FaceTrackerManager: unsupported state", state, "!!!")
    end
end

function FaceTrackerManager:Shutdown()
    if self.engine_face_tracker ~= nil then
        self.engine_face_tracker:Shutdown()
        self.engine_face_tracker = nil
    end

    if self.record_play_face_tracker ~= nil then
        self.record_play_face_tracker:Shutdown()
        self.record_play_face_tracker = nil
    end
end

function FaceTrackerManager:Stop()
    if self.record_play_face_tracker ~= nil then
        self.record_play_face_tracker:Stop()
    end

    if self.engine_face_tracker ~= nil then
        self.engine_face_tracker:Stop()
    end
end

function FaceTrackerManager:Merge(merge)
    self.merge = merge
end

function FaceTrackerManager:Update(dt)
    local detect_ok_face_tracker = nil
    if self.record_play_face_tracker ~= nil and self.record_play_face_tracker:IsActive() then
        self.record_play_face_tracker:Start()
        if self.record_play_face_tracker:Detect() then
            detect_ok_face_tracker = self.record_play_face_tracker
            if self.engine_face_tracker ~= nil then
                self.engine_face_tracker:Stop()
            end
        end
    end

    if nil == detect_ok_face_tracker or self.merge then
        if self.engine_face_tracker ~= nil and self.engine_face_tracker:IsActive() then
            self.engine_face_tracker:Start()
            if self.engine_face_tracker:Detect() then
                detect_ok_face_tracker = self.engine_face_tracker
            end
        end
    end

    if self.merge then

    else
        self.merged_face_infos = nil
    end
end

function FaceTrackerManager:OnEvent(msg, message_source)
    if self.record_play_face_tracker ~= nil then
        self.record_play_face_tracker:NodeEvent(msg, message_source)
    end
end

function FaceTrackerManager:GetFaceInfos()
    if self.merge then
        return self.merged_face_infos
    end

    if self.record_play_face_tracker ~= nil
            and self.record_play_face_tracker:IsActive()
            and self.record_play_face_tracker:IsPlaying() then
        return self.record_play_face_tracker:GetFaceInfos()
    end

    if self.engine_face_tracker ~= nil and self.engine_face_tracker:IsActive() then
        return self.engine_face_tracker:GetFaceInfos()
    end

    return nil
end

function FaceTrackerManager:EnablePose3D(enable)
    if enable then
        self.pose3DSwitch:Retain()
    else
        if self.pose3DSwitch:GetRefCount() > 0 then
            self.pose3DSwitch:Release()
        end
    end

    LOGD("FaceTrackerManager: EnablePose3D: try enable/disable pose 3d:", enable, ", pose 3d's switch ref count:", self.pose3DSwitch:GetRefCount())
end

function FaceTrackerManager:_EnablePose3DImpl(enable)
    LOGD("FaceTrackerManager: _EnablePose3DImpl: enable/disable pose 3d:", enable)
    if self.engine_face_tracker ~= nil then
        self.engine_face_tracker:EnablePose3D(enable)
    end
end

local FaceTrackService = define ("FaceTrackService", Service)

FaceTrackService.kStart = 1
FaceTrackService.kStop = 2
FaceTrackService.kData = 3


function RecordPlayFaceTracker:_OnInitialize()
    self.type = FaceTracker.kRecordFaceTrackerType
    self.playing = false
end

function RecordPlayFaceTracker:_OnDetect()
    return self:IsPlaying()
end

function RecordPlayFaceTracker:_OnActive(active)
    if not active then
        self.playing = false
    end
end

function RecordPlayFaceTracker:IsPlaying()
    return self.playing
end

function RecordPlayFaceTracker:NodeEvent(msg, message_source)
    if message_source == GameFramework.KMessage_Recorded then
        local id = msg:PopInt32()
        if id == FaceTrackService.kData then
            self.playing = true
            ReadFaceInfo(self.face_infos, msg)
        elseif id == FaceTrackService.kStart then
            LOGD("RecordPlayFaceTracker: record data start playing ...")
            self.playing = true
        elseif id == FaceTrackService.kStop then
            LOGD("RecordPlayFaceTracker: record data stop playing ...")
            self.playing = false
        end

        msg:PopEnd()
    end
end

function FaceTrackService:EnablePose3D(enable)
    self.face_tracker_manager:EnablePose3D(enable)
end

function FaceTrackService:_OnInitialize()
    self.service_type = Service.SERVICE_FACE_TRACK

    self.face_tracker_manager = FaceTrackerManager:New()
    self.face_tracker_manager:Initialize(self.attach_node)
end

function FaceTrackService:GetFaceInfos()
    return self.face_tracker_manager:GetFaceInfos()
end

function FaceTrackService:_OnStart()
    local state = self.attach_node.node_context.state
    if state == GamePlayRecord.kState_Recording or
            state == GamePlayRecord.kState_PlayRecord then
        local record_message = self:AllocRecordMessage()
        record_message:PushInt32(FaceTrackService.kStart)
        record_message:PushEnd()
    end
end

function FaceTrackService:_OnStop()
    self.face_tracker_manager:Stop()
    local state = self.attach_node.node_context.state
    if state == GamePlayRecord.kState_Recording or
            state == GamePlayRecord.kState_PlayRecord then
        local record_message = self:AllocRecordMessage()
        record_message:PushInt32(FaceTrackService.kStop)
        record_message:PushEnd()
    end
end

function FaceTrackService:_OnSleep()
    LOGD("FaceTrackService: sleep, stop the face trackers ...")
    self.face_tracker_manager:Stop()
end

function FaceTrackService:_OnNodeEvent(msg, message_source)
    self.face_tracker_manager:OnEvent(msg, message_source)
end

function FaceTrackService:_OnUpdate(dt)
    self.face_tracker_manager:Update(dt)

    local face_info = self.face_tracker_manager:GetFaceInfos()
    if face_info ~= nil then
        local state = self.attach_node.node_context.state
        if state == GamePlayRecord.kState_Recording or
                state == GamePlayRecord.kState_PlayRecord then
            local record_message = self:AllocRecordMessage()
            record_message:PushInt32(FaceTrackService.kData)
            WriteFaceInfo(face_info, record_message)
            record_message:PushEnd()
        end
    end
end

function FaceTrackService:_OnSwitchState(new_state)
    self.face_tracker_manager:RefreshFaceTracker(new_state)
end

function FaceTrackService:_OnShutdown()
    self.face_tracker_manager:Shutdown()
end

return FaceTrackService