--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local Service = import "Service"

local FilterService = define ("FilterService", Service)

LOGD("FilterService has been defined ..., parent is", Service)

function FilterService:_OnInitialize()
    self.service_type = Service.SERVICE_LOOKUP_TABLE
    self.input_texture = nil
    self.output_texture = nil
    self.external_output_texture = nil

    self.intensity = 1.0
    self.grid_count = 8

    self.lookup_textures = {}
    self.lookup_texture = nil

    self.color_filter = hl.ColorFilter_New()
    self.color_filter_inited = false

    self.configure_dirty = false

    self.frame_view_id_list = hl.U16Array()

    hl.Class_Retain(self.color_filter)
end

function FilterService:_GetLookupTexture(name)
    if self.lookup_textures[name] == nil then
        local lookup_texture_path = self.attach_node:GetRuntimeBuiltinDir() / ("filters/" .. name .. ".png")
        LOGD("FilterService: _GetLookupTexture: lookup_texture_path:", lookup_texture_path:CStr())
        self.lookup_textures[name] = self.attach_node:AllocImageTexture(hl.Named(name),
                lookup_texture_path, false, false)
    end

    return self.lookup_textures[name]
end

function FilterService:_Prepare()
    if self.configure_dirty == false then
        return
    end

    if self.color_filter_inited == false then
        self.color_filter:InitWithSetting(self.input_texture,
                self.external_output_texture or self.output_texture, self.lookup_texture,
                self.grid_count, self.intensity)
        self.color_filter_inited = true
    else
        self.color_filter:SetInput(self.input_texture)
        self.color_filter:SetOutput(self.external_output_texture or self.output_texture)
        self.color_filter:SetLookupGridCount(self.grid_count)
        self.color_filter:SetIntensity(self.intensity)
        CHECK(self.lookup_texture ~= nil, "FilterService: lookup texture is nil !!!")
        self.color_filter:SetLookupTexture(self.lookup_texture)
    end

    self.configure_dirty = false
end

function FilterService:_OnStop()
    self:EnableService(Service.SERVICE_SHOW_BG, false)
end

function FilterService:_OnStart()
    self:EnableService(Service.SERVICE_SHOW_BG, true)
end

function FilterService:_OnUpdate(dt)
    self:_Prepare()

    if self.color_filter:PrepareRender() then
        local count = self.color_filter:GetRenderViewCount()
        self.frame_view_id_list:clear()
        for k = 1, count do
            self.frame_view_id_list:push_back(hl.GPRenderDrive_AllocFrameView())
        end
        self.color_filter:Render(self.frame_view_id_list)
    end
end

function FilterService:_OnNodeEvent(msg, message_source)
    local param_num = msg:PopInt32()

    CHECK(param_num > 0, "FilterService: _OnNodeEvent: param num:", param_num, "<= 0 !!!")

    local param_key = nil
    local param_value = nil
    for i = 1, param_num do
        param_key = msg:PopInt32()
        if param_key == hl.Hl_MSG_SCRIPT_SERVICE_FILTER_INTENSITY then
            param_value = msg:PopFloat()
            self.intensity = param_value
            LOGD("FilterService: _OnNodeEvent: intensity ->", self.intensity)
        elseif param_key == hl.Hl_MSG_SCRIPT_SERVICE_FILTER_LOOKUP_IMAGE then
            param_value = msg:PopString()
            self.lookup_texture = self:_GetLookupTexture(param_value)
            LOGD("FilterService: _OnNodeEvent: lookup_texture ->", param_value)
            CHECK(self.lookup_texture ~= nil, "FilterService: _OnNodeEvent: lookup_texture:", param_value, "is nil !!!")
        else
            CHECK(false, "FilterService: _OnNodeEvent: unknown param key:", param_key, "!!!")
        end
    end

    self.configure_dirty = true
end

function FilterService:SetInputTexture(texture)
    CHECK(texture ~= nil, "FilterService: SetInputTexture: input texture is nil !!!")
    self.input_texture = texture
    if self.output_texture == nil or
            self.output_texture:GetWidth() ~= texture:GetWidth() or
            self.output_texture:GetHeight() ~= texture:GetHeight() or
            self.output_texture:GetType() ~= texture:GetType() or
            self.output_texture:GetFormat() ~= texture:GetFormat() then

        if self.output_texture ~= nil then
            hl.Class_Release(self.output_texture)
        end

        self.output_texture = hl.RenderTexture_New()
        self.output_texture:SetWidth(self.input_texture:GetWidth())
        self.output_texture:SetHeight(self.input_texture:GetHeight())
        self.output_texture:SetType(self.input_texture:GetType())
        self.output_texture:SetFlags(hl.GPTextureFlag_kDefault2DRT)
        self.output_texture:SetFormat(self.input_texture:GetFormat())
        hl.Class_Retain(self.output_texture)
    end

    self.configure_dirty = true
end

function FilterService:SetExternalOutputTexture(texture)
    self.external_output_texture = texture

    self.configure_dirty = true
end

function FilterService:_OnShutdown()
    if self.output_texture ~= nil then
        hl.Class_Release(self.output_texture)
        self.output_texture = nil
    end

    hl.Class_Release(self.color_filter)
    self.color_filter_inited = false
    self.color_filter = nil
end

function FilterService:GetOutputTexture()
    return self.external_output_texture or self.output_texture
end

return FilterService