--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local Service = import "Service"
local Node = import "Node"

local FontService = define ("FontService", Service)
local GameFramework = import "GameFramework"
local RotationGestureDetector = import "RotationGestureDetector"
local ScaleGestureDetector = import "ScaleGestureDetector"

local kMsgSetText       = 1

local kTextStyleNormal  = 0
local kTextStyleBold    = 1 << 0
local kTextStyleItalic  = 1 << 1
local kTextStyleStroke  = 1 << 2

LOGD("FontService has been defined ..., parent is", Service)

local kInvalidTouchedTextID = -10000

function FontService:_OnInitialize()
    self.service_type = Service.SERVICE_FONT

    self.text_actor_list = {}
end

function FontService:_AllocTextActor(id, font_file, text, style, color, size, width)
    CHECK(self.text_actor_list[id] == nil or self.text_actor_list[id].actor == nil, "FontService: Text id already exists", id)

    local skew, border, embolden
    if style & kTextStyleItalic ~= 0 then
        skew = hl.Vec2(0.5, 0.0)
    else
        skew = hl.Vec2(0.0, 0.0)
    end

    if style & kTextStyleBold ~= 0 then
        embolden = 1
    else
        embolden = 0
    end

    if style & kTextStyleStroke ~= 0 then
        border = 2
    else
        border = 0
    end

    size = size or 54
    width = width or hl.RuntimeWidth() - 100

    local font = hl.Font_New()
    if font:InitWithSource(self:GetRuntimeBuiltinDir() / font_file,
            size, 512, border, embolden, hl.Vec2_kOne, skew) then
        local utf16 = hl.U16Array()
        if hl.UTF8ToUTF16(text, utf16) then
            local name = hl.Named("FontActor")
            local text_actor = self:AllocActor(name)
            local texter = hl.Texter_New()
            local font_mtl = hl.Font_CreateSimpleMaterial(color)
            texter:SetMaterial(font_mtl)
            texter:SetWidth(width)
            texter:SetEnableAutoSize(true)
            texter:SetFont(font)
            texter:SetText(utf16)
            self:SetRenderLayer(texter, self.attach_node.LAYER_OVERLAY_2D)
            text_actor:AddComponent(texter)
            self:AddToWorld(text_actor)

            hl.Class_Retain(text_actor)
            if self.text_actor_list[id] then
                hl.Class_Release(self.text_actor_list[id].actor)
            end

            local desc = {}
            desc.actor = text_actor
            desc.style = style
            desc.color = color
            self.text_actor_list[id] = desc

            return text_actor
        end
    end
end

function FontService:_UpdateText(id, text, color)
    CHECK(self.text_actor_list[id].actor ~= nil, "FontService: Not find text id", id)

    if string.len(text) == 0 then
        self:RemoveFromWorld(self.text_actor_list[id].actor)
        hl.Class_Release(self.text_actor_list[id].actor)
        self.text_actor_list[id] = nil
    else
        local utf16 = hl.U16Array()
        if hl.UTF8ToUTF16(text, utf16) then
            local texter = hl.Texter_Cast(self.text_actor_list[id].actor:QueryComponent(hl.Texter_ID()))
            texter:SetWidth(hl.RuntimeWidth() - 100)
            texter:SetText(utf16)
            texter:GetMaterial():SetColor(color or hl.Vec4(1.0))
        end
    end
end

function FontService:_OnNodeEvent(msg, msg_src)
    local param_num = msg:PopInt32()
    local option = msg:PopInt32()
    CHECK(option == kMsgSetText)
    local id = msg:PopInt32()
    --CHECK(id == 0)
    local style = msg:PopInt32()
    local R = msg:PopFloat()
    local G = msg:PopFloat()
    local B = msg:PopFloat()
    local A = msg:PopFloat()
    local color = hl.Vec4(R, G, B, A)

    local text = msg:PopString()

    if self.text_actor_list[id] and self.text_actor_list[id].actor then
        if style ~= self.text_actor_list[id].style then
            self:_UpdateText(id, "")
            self:_AllocTextActor(
                    id,
                    "FangSong_GB2312.ttf",
                    text,
                    style,
                    color
            )
        else
            self:_UpdateText(id, text, color)
        end
    else
        self:_AllocTextActor(
                id,
                "FangSong_GB2312.ttf",
                text,
                style,
                color
        )
    end
end

function FontService:_OnStart()
    if hl.IsPlatformMobile() then
        self:AddListener(hl.Hl_MSG_TOUCH, self._OnTouchEvent, self)

        self.sgd = ScaleGestureDetector:New()
        self.sgd:Initialize(self)

        self.rgd = RotationGestureDetector:New()
        self.rgd:Initialize(self)
    else
        self:AddListener(hl.Hl_MSG_MOUSE, self._OnMouseEvent, self)
    end
end

function FontService:_OnStop()
    for _, desc in pairs(self.text_actor_list) do
        if desc then
            self.attach_node:RemoveFromWorld(desc.actor)
            hl.Class_Release(desc.actor)
        end
    end
    self.text_actor_list = {}
end

function FontService:_OnMouseEvent(action, x, y)
    if action == hl.Hl_ACTION_MOUSE_ENTER_LEFT then
        self.touched_text_id = self:_RaycastTest(x, y)
    elseif action == hl.Hl_ACTION_MOUSE_LEAVE_LEFT then
        self.touched_text_id = kInvalidTouchedTextID
    end
end

function FontService:_OnTouchEvent(action, touches)
    if action == hl.Hl_ACTION_TOUCH_BEGAN then
        for id, touch in ipairs(touches) do
            if touch.id == 0 then
                self.touched_text_id = self:_RaycastTest(touch.x, touch.y)
                self.x = touch.x
                self.y = touch.y
                break
            end
        end
    elseif action == hl.Hl_ACTION_TOUCH_MOVED then
        for id, touch in ipairs(touches) do
            if touch.id == 0 then
                if self.touched_text_id ~= kInvalidTouchedTextID then
                    local actor = self.text_actor_list[self.touched_text_id].actor
                    local pos = actor:GetTransform():GetLocalPosition()
                    pos.x = pos.x + (touch.x - self.x)
                    pos.y = pos.y + (self.y - touch.y)
                    self.x = touch.x
                    self.y = touch.y
                    actor:GetTransform():SetLocalPosition(pos)
                end
                break
            end
        end
    elseif action == hl.Hl_ACTION_TOUCH_ENDED then
        self.touched_text_id = kInvalidTouchedTextID
    end

    self.sgd:TouchEvent(action, touches)
    self.rgd:TouchEvent(action, touches)
end

function FontService:_RaycastTest(x, y)
    local camera2d = GameFramework:GetNodeManager():GetDefaultCamera2d()
    local ray2d = camera2d:ScreenToRay(hl.Vec2(x, y))
    for id, desc in pairs(self.text_actor_list) do
        local texter = hl.Texter_Cast(desc.actor:QueryComponent(hl.Texter_ID()))
        if texter:RayCastTest(ray2d) then
            return id
        end
    end
    return kInvalidTouchedTextID
end

function FontService:OnScaleBegin(sgd)
    return self.touched_text_id ~= kInvalidTouchedTextID
end

function FontService:OnScale(sgd)
    if self.touched_text_id ~= kInvalidTouchedTextID then
        local scale = sgd:GetScaleFactor()
        local desc = self.text_actor_list[self.touched_text_id]
        local trans = desc.actor:GetTransform()
        trans:SetLocalScale(trans:GetLocalScale() * scale)
    end
end

function FontService:OnScaleEnd(sgd)
end

function FontService:OnRotateBegin(rgd)
    return self.touched_text_id ~= kInvalidTouchedTextID
end

function FontService:OnRotate(rgd)
    if self.touched_text_id ~= kInvalidTouchedTextID then
        local rotate = rgd:GetRotation()
        local desc = self.text_actor_list[self.touched_text_id]
        local trans = desc.actor:GetTransform()
        trans:SetLocalRotation(trans:GetLocalRotation() * hl.Quat(rotate, hl.Vec3_kAxisZ))
    end
end

function FontService:OnRotateEnd(rgd)
end

return FontService