--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local GameFramework = define "GameFramework"
local GamePlayRecord = define "GamePlayRecord"
local NodeManager = import "NodeManager"
local NetworkManager = import "NetworkManager"

GameFramework.kMessage_A2S			= 0
GameFramework.kMessage_N2S			= 1
GameFramework.KMessage_Recorded		= 2

function GameFramework:GetNodeManager()
    return self.node_manager
end

function GameFramework:Initialize()
	self.node_manager = nil
	self.play_record = nil
	self.scripts = {}

	local this = self

	local MouseEventProc = function(msg)
		local action = msg:PopInt32()
		local x = msg:PopFloat()
		local y = msg:PopFloat()
		return action, x, y
	end

	local TouchEventProc = function(msg)
		local touches = {}
		local action = msg:PopInt32()
		local num = msg:PopInt32()
		local id, x, y
		for i = 1, num do
			id = msg:PopInt32()
			x = msg:PopFloat()
			y = msg:PopFloat()
			table.insert(touches, {id = id, x = x, y = y})
		end
		return action, touches
	end

	local ViewEventProc = function(msg)
		local object = msg:PopObject()
		local action = msg:PopInt32()
		return object, action
	end

	local AppEventProc = function(msg)
		local action = msg:PopInt32()
		return action, msg
	end

	local KeyEventProc = function(msg)
		local action = msg:PopInt32()
		local key = msg:PopInt32()
		return action, key
	end

	local CharEventProc = function(msg)
		local character = msg:PopUint32()
		return character
	end

	local InputTextProc = function(msg)
		local input_text = msg:PopString()
		return input_text
	end

	local PlayerActionProc = function(msg)
		local action = msg:PopInt32()
		return action, msg
	end

	local PlayerEventProc = function(msg)
		local play_id = msg:PopInt32()
		local event = msg:PopInt32()
		return play_id, event, msg
	end

	self.MsgProc = {
		[hl.Hl_N2S_RUNTIME_INITIALIZE] = function(msg)
			hl.GPRenderDrive_SetDebug(hl.GPRenderDebugFlag_kShowRenderLog) --| hl.GPRenderDebugFlag_kShowRenderInfo

			CHECK(this.node_manager == nil, "Runtime already initialize")
			this.play_record = GamePlayRecord:New()
			this.play_record:Initialize()
			this.node_manager = NodeManager:New()
			this.node_manager:Initialize(this.play_record)
			this.node_manager:SetState(this.play_record:GetState())
			this.play_record:SetMessageListener(this.node_manager)

			if hl.IsEditorMode() then
				Ed.gui = hl.GUI_New()
				Ed.gui:SetFilterMask(this.node_manager:GetDefaultCamera2d():GetFilterMask())
				Ed.gui:SetLayer(0x0FFFFFFF)
				local gui_actor = hl.Actor_New()
				gui_actor:SetName(hl.NamedView("_Hulua_Editor_GUI_"))
				gui_actor:AddComponent(Ed.gui)
				hl.World_Query(0):AddActor(gui_actor)
			end

			LOGD("Hl_N2S_RUNTIME_INITIALIZE")
			return 0
		end,
		[hl.Hl_N2S_RUNTIME_SHUTDOWN] = function(msg)
			CHECK(this.node_manager ~= nil, "Runtime not initialize")
			this.play_record:SetMessageListener(nil)
			this.node_manager:Shutdown()
			this.node_manager = nil
			this.play_record:Shutdown()
			this.play_record = nil
			LOGD("Hl_N2S_RUNTIME_SHUTDOWN")
			return 0
		end,
		[hl.Hl_MSG_RUNTIME_RESET] = function(msg)
			CHECK(this.node_manager ~= nil, "Runtime not initialize")
			local w = msg:PopInt32()
			local h = msg:PopInt32()
			hl.GPRenderDrive_Reset(w, h)
			LOGD("Hl_MSG_RUNTIME_RESET", w, h)
			return w, h
		end,
		[hl.Hl_MSG_WORKDIR_CHANGE] = function(msg)
			local workdir = msg:PopString()
			hl.SetWorkDir(hl.Path(workdir))
			LOGD("Hl_MSG_WORKDIR_CHANGE", workdir)
			return workdir
		end,
		[hl.Hl_MSG_SCRIPT_STATE_PREVIEW] = function(msg)
			this.play_record:SetState(GamePlayRecord.kState_Preview, msg)
			self.node_manager:SetState(GamePlayRecord.kState_Preview)
			LOGD("Hl_MSG_SCRIPT_STATE_PREVIEW")
		end,
		[hl.Hl_MSG_SCRIPT_STATE_RECORDING] = function(msg)
			this.play_record:SetState(GamePlayRecord.kState_Recording, msg)
			this.node_manager:SetState(GamePlayRecord.kState_Recording)
			this.node_manager:SetMessageRecorder(this.play_record:GetRecorder())
			LOGD("Hl_MSG_SCRIPT_STATE_RECORDING")
		end,
		[hl.Hl_MSG_SCRIPT_STATE_PLAYING] = function(msg)
			this.play_record:SetState(GamePlayRecord.kState_Playing, msg)
			this.node_manager:SetState(GamePlayRecord.kState_Playing)
			LOGD("Hl_MSG_SCRIPT_STATE_PLAYING")
		end,
		[hl.Hl_MSG_SCRIPT_STATE_PLAY_RECORD] = function(msg)
			this.play_record:SetState(GamePlayRecord.kState_PlayRecord, msg)
			this.node_manager:SetState(GamePlayRecord.kState_PlayRecord)
			this.node_manager:SetMessageRecorder(this.play_record:GetRecorder())
			LOGD("Hl_MSG_SCRIPT_STATE_PLAY_RECORD")
		end,
		[hl.Hl_MSG_SCRIPT_STATE_EDITING] = function(msg)
			-- this.play_record:SetState(GamePlayRecord.kState_Editing)
			-- this.node_manager:SetState(GamePlayRecord.kState_Editing)
			LOGD("Hl_MSG_SCRIPT_STATE_EDITING")
		end,
		[hl.Hl_MSG_SCRIPT_RECORD_SAVE] = function(msg)
			local save_file = msg:PopString()
			this.play_record:SaveRecordData(save_file)
			LOGD("Hl_MSG_SCRIPT_RECORD_SAVE", save_file)
		end,
		[hl.Hl_MSG_SCRIPT_RECORD_REMOVE_BACK] = function(msg)
			local remove_begin_time_ms = msg:PopFloat()
			this.play_record:RemoveBack(remove_begin_time_ms)
			LOGD("Hl_MSG_SCRIPT_RECORD_REMOVE_BACK", remove_begin_time_ms)
		end,
		[hl.Hl_MSG_SCRIPT_INITIALIZE] = function(msg)
			return this.node_manager:OnA2SEvent(msg)
		end,
		[hl.Hl_MSG_SCRIPT_SHUTDOWN] = function(msg)
			return this.node_manager:OnA2SEvent(msg)
		end,
		[hl.Hl_MSG_SCRIPT_RESET] = function(msg)
			return this.node_manager:OnA2SEvent(msg)
		end,
		--[hl.Hl_MSG_SCRIPT_HIERARCHY_CHANGED] = function(msg)
		--	this.node_manager:OnA2SEvent(msg)
		--end,
		[hl.Hl_MSG_SCRIPT_TRANSFORM_CHANGED] = function(msg)
			this.node_manager:OnA2SEvent(msg)
		end,
		[hl.Hl_MSG_SCRIPT_POSITION_CHANGED] = function(msg)
			this.node_manager:OnA2SEvent(msg)
		end,
		[hl.Hl_MSG_SCRIPT_ROTATION_CHANGED] = function(msg)
			this.node_manager:OnA2SEvent(msg)
		end,
		[hl.Hl_MSG_SCRIPT_SCALE_CHANGED] = function(msg)
			this.node_manager:OnA2SEvent(msg)
		end,
		[hl.Hl_MSG_SCRIPT_NODE_LOGIC] = function(msg)
			this.node_manager:OnA2SEvent(msg)
		end,
		[hl.Hl_MSG_FILE_DIALOG_OPEN_RESULT] = function(msg)
			local ok = msg:PopInt32()
			if ok == 1 then
				local path = msg:PopString()
				LOGD("Hl_MSG_FILE_DIALOG_OPEN_RESULT", path)
			end
		end,
		[hl.Hl_MSG_TOUCH] = TouchEventProc,
		[hl.Hl_MSG_MOUSE] = MouseEventProc,
		[hl.Hl_MSG_INPUT_KEY] = KeyEventProc,
		[hl.Hl_MSG_INPUT_CHAR] = CharEventProc,
		[hl.Hl_N2S_MSG_TOUCH] = ViewEventProc,
		[hl.Hl_A2S_MSG_LOGIC] = AppEventProc,
		[hl.Hl_MSG_SCRIPT_RECEIVE_NETWORK_RESPONSE] = function(msg)
			local request_id = msg:PopInt64()
			local return_code = msg:PopInt32()
			local response = msg:PopString()
			NetworkManager:OnResponse(request_id, return_code, response)
		end,
		[hl.Hl_MSG_SCRIPT_INPUT_TEXT] = InputTextProc,
		[hl.Hl_MSG_SCRIPT_PLAYER_ACTION] = PlayerActionProc,
		[hl.Hl_MSG_SCRIPT_PLAYER_EVENT] = PlayerEventProc,

		[hl.Hl_MSG_LIVE2D_MODEL_CREATE] = function(msg)
			local user_id = msg:PopInt32()
			local config = msg:PopString()
			return user_id, config
		end,
		[hl.Hl_MSG_LIVE2D_LOAD_PHYSICS] = function(msg)
			local user_id = msg:PopInt32()
			local path = hl.Path(msg:PopString())
			return user_id, path
		end,
		[hl.Hl_MSG_LIVE2D_LOAD_POSE] = function(msg)
			local user_id = msg:PopInt32()
			local path = hl.Path(msg:PopString())
			return user_id, path
		end,
		[hl.Hl_MSG_LIVE2D_LOAD_TEXTURE] = function(msg)
			local user_id = msg:PopInt32()
			local num = msg:PopUint32()
			local tex_ids = hl.StrArray()
			local tex_paths = hl.PathArray()
			for i = 1, num do
				local id = msg:PopString()
				local path = msg:PopString()
				tex_ids:push_back(id)
				tex_paths:push_back(hl.Path(path))
			end
			return user_id, tex_ids, tex_paths
		end,
		[hl.Hl_MSG_LIVE2D_LOAD_EXPRESSION] = function(msg)
			local user_id = msg:PopInt32()
			local num = msg:PopUint32()
			local expressions = hl.PathArray()
			for i = 1, num do
				local path = msg:PopString()
				expressions:push_back(hl.Path(path))
			end
			return user_id, expressions
		end,
		[hl.Hl_MSG_LIVE2D_LOAD_MOTION] = function(msg)
			local user_id = msg:PopInt32()
			local num = msg:PopString()
			local motions = hl.PathArray()
			for i = 1, num do
				local path = msg:PopString()
				motions:push_back(hl.Path(path))
			end
			return user_id, motions
		end,
		[hl.Hl_MSG_LIVE2D_START_MOTION] = function(msg)
			local user_id = msg:PopInt32()
			local num_motions = msg:PopUint32()
			local pa = hl.PathArray()
			for i = 1, num_motions do
				local motion = hl.Path(msg:PopString())
				pa:push_back(motion)
			end
			local queue_index = msg:PopUint32()
			local looper = msg:PopInt32()
			local force = msg:PopInt32()
			return user_id, pa, queue_index, looper == 1, force == 1
		end,
		[hl.Hl_MSG_LIVE2D_START_PER_MOTION] = function(msg)
			local user_id = msg:PopInt32()
			local num_motions = msg:PopUint32()
			local pa = hl.PathArray()
			local pc = hl.I32Array()
			for i = 1, num_motions do
				local motion = hl.Path(msg:PopString())
				local play_count = msg:PopInt32()
				pa:push_back(motion)
				pc:push_back(play_count)
			end
			local queue_index = msg:PopUint32()
			local looper = msg:PopInt32()
			local force = msg:PopInt32()
			return user_id, pa, pc, queue_index, looper == 1, force == 1
		end,
		[hl.Hl_MSG_LIVE2D_STOP_MOTION] = function(msg)
			local user_id = msg:PopInt32()
			local stop_all = msg:PopInt32() == 1
			local motion
			if not stop_all then
				motion = msg:PopUint32()
			end
			return user_id, stop_all, motion
		end,
		[hl.Hl_MSG_LIVE2D_SET_EXPRESSTION] = function(msg)
			local user_id = msg:PopInt32()
			local expression = hl.Path(msg:PopString())
			return user_id, expression
		end,
		[hl.Hl_MSG_LIVE2D_REPLACE_TEXTURE] = function(msg)
			local user_id = msg:PopInt32()
			local num = msg:PopUint32()
			local tex_ids = hl.StrArray()
			local tex_paths = hl.PathArray()
			for i = 1, num do
				local id = msg:PopString()
				local path = msg:PopString()
				tex_ids:push_back(id)
				tex_paths:push_back(hl.Path(path))
			end
			return user_id, tex_ids, tex_paths
		end,
		[hl.Hl_MSG_LIVE2D_REMOVE_TEXTURE] = function(msg)
			local user_id = msg:PopInt32()
			local num = msg:PopUint32()
			local tex_ids = hl.StrArray()
			for i = 1, num do
				local id = msg:PopString()
				tex_ids:push_back(id)
			end
			return user_id, tex_ids
		end,
		[hl.Hl_MSG_LIVE2D_SET_PARAMETER] = function(msg)
			local user_id = msg:PopInt32()
			local id = msg:PopString()
			local value = msg:PopFloat()
			return user_id, id, value
		end,
		[hl.Hl_MSG_LIVE2D_SET_ACTIONS] = function(msg)
			local user_id = msg:PopInt32()
			local json_cfg = msg:PopString()
			return user_id, json_cfg
		end,
		[hl.Hl_MSG_LIVE2D_SET_TEXT] = function(msg)
			local user_id = msg:PopInt32()
			local show = msg:PopInt32()
			local text = ""
            local time_sec = 0
			if show == 1 then
				text = msg:PopString()
                time_sec = msg:PopFloat()
			end
			return user_id, show, text, time_sec
		end,
		[hl.Hl_MSG_LIVE2D_SET_BACKGROUND] = function(msg)
			local user_id = msg:PopInt32()
			local path = msg:PopString()
			return user_id, path
		end,
		[hl.Hl_MSG_LIVE2D_SET_MODE] = function(msg)
			local user_id = msg:PopInt32()
			local mode = msg:PopInt32()
			return user_id, mode
		end,
		[hl.Hl_MSG_LIVE2D_SET_STYLE] = function(msg)
			local user_id = msg:PopInt32()
			local style = msg:PopInt32()
			return user_id, style
		end,
		[hl.Hl_MSG_LIVE2D_MODEL_RELEASE] = function(msg)
			local user_id = msg:PopInt32()
			return user_id
		end,
		[hl.Hl_MSG_LIVE2D_SET_MODEL_VISIBILE] = function(msg)
			local user_id = msg:PopInt32()
			local visibile = msg:PopInt32()
			return user_id, visibile
		end,
		[hl.Hl_MSG_LIVE2D_SET_MOTION_STREAK] = function(msg)
			local user_id = msg:PopInt32()
			local enabled = msg:PopInt32()
			local config_path
			if enabled == 1 then
				config_path = msg:PopString()
			end
			return user_id, enabled == 1, config_path
		end,
		[hl.Hl_MSG_LIVE2D_SET_MODEL_SLOT] = function(msg)
			local user_id = msg:PopInt32()
			local slot = msg:PopInt32()
			return user_id, slot
		end,
		[hl.Hl_MSG_LIVE2D_SET_ENABLE_BREATH] = function(msg)
			local user_id = msg:PopInt32()
			local enable = msg:PopInt32()
			return user_id, enable == 1
		end,
		[hl.Hl_MSG_LIVE2D_SET_ENABLE_EYEBLINK] = function(msg)
			local user_id = msg:PopInt32()
			local enable = msg:PopInt32()
			return user_id, enable == 1
		end,
		[hl.Hl_MSG_LIVE2D_ENABLE_FACE_TRACKER] = function(msg)
			local user_id = msg:PopInt32()
			local enable = msg:PopInt32()
			return user_id, enable == 1
		end,
		[hl.Hl_MSG_LIVE2D_ENABLE_AVATAR_TRACKER] = function(msg)
			local user_id = msg:PopInt32()
			local enable = msg:PopInt32()
			return user_id, enable == 1
		end,
		[hl.Hl_MSG_LIVE2D_ENABLE_WATERMARK] = function(msg)
			local user_id = msg:PopInt32()
			local enable = msg:PopInt32()
			return user_id, enable == 1
		end,
		[hl.Hl_MSG_IV_PLAYER_LOAD] = function(msg)
			local json_path = msg:PopString()
			return json_path
		end,
		[hl.Hl_MSG_IV_PARAMETER_UPDATE] = function(msg)
			local num_param = msg:PopUint32()
			local parameters = {}
			local n, v
			for i = 1, num_param do
				n = msg:PopString()
				v = msg:PopInt32()
				table.insert(parameters, { name = n, value = v })
			end
			return parameters
		end,
		[hl.Hl_MSG_IV_PLAYER_PLAY] = function(msg)
			local scene_index = msg:PopInt32()
			local node_index = msg:PopInt32()
			return scene_index, node_index
		end,
	}

	self.MsgRec = {}

	for k, _ in pairs(self.MsgProc) do
		self.MsgRec[k] = {}
	end

	return true
end

function GameFramework:Shutdown()
	LOGD("GameFramework:Shutdown")
end

function GameFramework:AddListener(event, listener, holder)
	table.insert(self.MsgRec[event], { l = listener, h = holder})
end

function GameFramework:RemoveListener(event, listener)
	local t = self.MsgRec[event]
	for k, v in ipairs(t) do
		if v.l == listener then
			table.remove(t, k)
			break
		end
	end
end

function GameFramework:Event(msg)
	-- self.node_manager:OnN2SEvent(msg)

	local id = msg:PopBegin()
	local result = { self.MsgProc[id](msg) }
	msg:PopEnd()

    for _, v in ipairs(self.MsgRec[id]) do
        if v.h then
            v.l(v.h, table.unpack(result))
        else
            v.l(table.unpack(result))
		end
	end

	return 1
end

function GameFramework:SimulationEffectEditor()
	self.playing = self.playing or false
	self.recording = self.recording or false
	self.playing_var = self.playing_var or hl.Variant(false)
	self.recording_var = self.recording_var or hl.Variant(false)
	self.playing_var:Set(self.playing)
	self.recording_var:Set(self.recording)
	Ed.gui:SetNextWindowSize(hl.Vec2(200, 200))
	Ed.gui:SetNextWindowPos(hl.Vec2(0, 0))
	Ed.gui:Begin("EffectEditor", hl.GUI_kNoMove)
	if Ed.gui:VCheckBox("Playing", self.playing_var) then
		self.playing = self.playing_var:GetBool()
		LOGD("Playing", self.playing)
		local msg = hl.MessageQueue_MakeAsync()
		msg:PushBegin(hl.Hl_MSG_NATIVE_FILE_DIALOG_OPEN)
		msg:PushUint32(0)
		msg:PushInt32(0)
		msg:PushEnd()
		hl.MessageQueue_PostAndFreeAsync(msg)
	end
	if Ed.gui:VCheckBox("Recording", self.recording_var) then
		self.recording = self.recording_var:GetBool()
		LOGD("Recording", self.recording)
		self.aaa = true
	end
	Ed.gui:End()
	-- self.select = Ed.Show(self.node_manager:GetWorld(), self.select)
	-- if self.select then
	-- 	Ed.Show(self.select)
	-- end
end

function GameFramework:ProcAsyncMessage()
	local msg = hl.MessageQueue_AcquireAsync()
	local id
	while(msg)
	do
		local clone_msg = hl.MessageQueue_Make()
		hl.MessageQueue_PostAndFree(clone_msg:CopyFrom(msg))
		id = msg:PopBegin()
		if self.MsgProc[id] then
			local result = { self.MsgProc[id](msg) }
			for _, v in ipairs(self.MsgRec[id]) do
				if v.h then
					v.l(v.h, table.unpack(result))
				else
					v.l(table.unpack(result))
				end
			end
		end
		hl.MessageQueue_FreeAsync(msg)
		msg = hl.MessageQueue_AcquireAsync()
	end
end

function GameFramework:ProcSyncMessage()
	local msg = hl.MessageQueue_AcquireSync()
	local id
	while(msg)
	do
		local clone_msg = hl.MessageQueue_Make()
		hl.MessageQueue_PostAndFree(clone_msg:CopyFrom(msg))
		id = msg:PopBegin()
		local result = { self.MsgProc[id](msg) }
		for _, v in ipairs(self.MsgRec[id]) do
			if v.h then
				v.l(v.h, table.unpack(result))
			else
				v.l(table.unpack(result))
			end
		end
		hl.MessageQueue_FreeSync(msg)
		msg = hl.MessageQueue_AcquireSync()
	end
end


local mega_to_byte = 1024 * 1024
local byte_to_mega = 1 / mega_to_byte
local gc_threshold = 5 * mega_to_byte
function GameFramework:Update(dt)

	local need_update = self.node_manager:NeedUpdate()
	if need_update then
		hl.GPRenderDrive_BeginFrame()
		hl.World_Prepare(dt)

		if hl.IsEditorMode() then
			Ed.gui:BeginFrame()
		end
	end

	self.play_record:BeginFrame(dt)
	self:ProcSyncMessage()
	self:ProcAsyncMessage()
	self.play_record:UpdateFrame(dt)
	
	if self.node_manager:NeedUpdate() then
		if not need_update then
			hl.GPRenderDrive_BeginFrame()
			hl.World_Prepare(dt)
			if hl.IsEditorMode() then
				Ed.gui:BeginFrame()
			end
		end
		-- if hl.IsEditorMode() then
		-- 	self:SimulationEffectEditor()
		-- end

		self.node_manager:UpdateFrame(dt)
		need_update = true
	end

	self.play_record:EndFrame(dt)

	if need_update then
		if hl.IsEditorMode() then
			Ed.gui:EndFrame()
		end

		hl.World_Update(dt)
		hl.GPRenderDrive_EndFrame()
		hl.World_Post(dt)
	end

	if self.aaa then
		hl.RenderData_UnloadAll()
		self.aaa = false
	end

	if need_update then
		return hl.ScriptVM_kRuning
	else
		return hl.ScriptVM_kNoLogic
	end
end

return GameFramework
