--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local GamePlayRecord = define "GamePlayRecord"

GamePlayRecord.kState_Preview 		= 0
GamePlayRecord.kState_Recording 	= 1
GamePlayRecord.kState_Playing 		= 2
GamePlayRecord.kState_PlayRecord 	= 3
GamePlayRecord.kState_Editing 		= 4


GamePlayRecord.kMessageVersion0		= 0

GamePlayRecord.kMessageType0		= 0
GamePlayRecord.kMessageType1		= 1

local class_rw_ctx = hl.ClassRWContext()
class_rw_ctx.version = hl.Hl_ENGINE_VERSION
class_rw_ctx.min_version = hl.Hl_ENGINE_MINIMUM_VERSION
class_rw_ctx.package_name = hl.Named(hl.Hl_RECORD_DEFAULT_FOLDER)
local function GetRecordDataCtx()
	class_rw_ctx.work_dir = hl.GetWorkDir()
	return class_rw_ctx
end

function GamePlayRecord:Initialize()
	self.player = hl.MessagePlayer_New()
	self.recorder = hl.MessageRecorder_New()
	self.state = GamePlayRecord.kState_Preview
	self.play_version = nil
	self.play_type = nil
	self.record_version = GamePlayRecord.kMessageVersion0
	self.record_type = GamePlayRecord.kMessageType1
	self.start = false
	self.msg_listener = nil
end

function GamePlayRecord:Shutdown()
	if self.player then
		hl.Class_Discard(self.player)
		self.player = nil
	end
	if self.recorder then
		hl.Class_Discard(self.recorder)
		self.recorder = nil
	end
	self.state = GamePlayRecord.kState_Preview
	self.start = false
	self.msg_listener = nil
end

function GamePlayRecord:IsWorking()
	return self.start
end

function GamePlayRecord:GetPlayer()
	return self.player
end

function GamePlayRecord:GetRecorder()
	return self.recorder
end

function GamePlayRecord:GetState()
	return self.state
end

function GamePlayRecord:GetPlayVersion()
	return self.play_version
end

function GamePlayRecord:GetRecordVersion()
	return self.record_version
end

function GamePlayRecord:GetPlayType()
	return self.play_type
end

function GamePlayRecord:GetRecordType()
	return self.record_type
end

function GamePlayRecord:GetMessageListener()
	return self.msg_listener
end

function GamePlayRecord:SetMessageListener(listener)
	self.msg_listener = listener
end

function GamePlayRecord:SetState(state, msg)
	if state == GamePlayRecord.kState_Preview then
		self.state = GamePlayRecord.kState_Preview
	elseif state == GamePlayRecord.kState_Recording then
		self.state = GamePlayRecord.kState_Recording
		local action = msg:PopInt32()
		if action == hl.Hl_RECORD_ACTION_RESET then
			LOGD("GamePlayRecord kState_Recording: Hl_RECORD_ACTION_RESET")
			self.recorder:Reset()
			local msg_store = hl.MessageStore_New()
			msg_store:SetUserData(0, self.record_version)
			msg_store:SetUserData(1, self.record_type)
			self.recorder:SetTarget(msg_store)
		else
			LOGD("GamePlayRecord kState_Recording: Hl_RECORD_ACTION_APPEND")
			if not self.recorder:GetTarget() then
				local msg_store = hl.MessageStore_New()
				msg_store:SetUserData(0, self.record_version)
				msg_store:SetUserData(1, self.record_type)
				self.recorder:SetTarget(msg_store)
			end
		end
	elseif state == GamePlayRecord.kState_Playing then
		local name = msg:PopString()
		LOGD("GamePlayRecord kState_Playing: ", name)
		self.state = GamePlayRecord.kState_Playing
		self.player:Reset()
		local data = hl.MessageStore_Cast(hl.Class_Load(hl.MessageStore_ID(), hl.NamedView(name), GetRecordDataCtx()))
		local count = data:GetUserDataCount()
		if count >= 2 then
			self.play_version = data:GetUserData(0)
			self.play_type = data:GetUserData(1)
		end
		self.player:SetTarget(data)
	elseif state == GamePlayRecord.kState_PlayRecord then
		local name = msg:PopString()
		LOGD("GamePlayRecord kState_PlayRecord: ", name)
		self.state = GamePlayRecord.kState_PlayRecord
		self.player:Reset()
		local data = hl.MessageStore_Cast(hl.Class_Load(hl.MessageStore_ID(), hl.NamedView(name), GetRecordDataCtx()))
		local count = data:GetUserDataCount()
		if count >= 2 then
			self.play_version = data:GetUserData(0)
			self.play_type = data:GetUserData(1)
		end
		self.player:SetTarget(data)
		self.recorder:Reset()
		local msg_store = hl.MessageStore_New()
		msg_store:SetUserData(0, self.record_version)
		msg_store:SetUserData(1, self.record_type)
		self.recorder:SetTarget(msg_store)
	elseif state == GamePlayRecord.kState_Editing then
		self.state = GamePlayRecord.kState_Editing
	else
		CHECK(false, "Invalid state: ", state)
	end
	self.start = false
end

function GamePlayRecord:SaveRecordData(save_file)
	LOGD("GamePlayRecord:SaveRecordData:", save_file)
	local store = self.recorder:GetTarget()
	local ok
	if store then
		store:SetName(hl.NamedView(save_file))
		ok = hl.Class_Save(store, GetRecordDataCtx())
	end

	local send_msg = hl.MessageQueue_MakeAsync()
	send_msg:PushBegin(hl.Hl_MSG_NATIVE_SCRIPT_RECORD_SAVE)
	if ok then
		send_msg:PushInt32(1)
		LOGD("GamePlayRecord save record file success: ", save_file)
	else
		send_msg:PushInt32(0)
		LOGW("GamePlayRecord save record file failed: ", save_file)
	end
	send_msg:PushEnd()
	hl.MessageQueue_PostAndFreeAsync(send_msg)

	return ok
end

function GamePlayRecord:RemoveBack(abs_begin_ms)
	if self.recorder:GetTarget() then
		self.recorder:Remove(abs_begin_ms)
	end
end

function GamePlayRecord:_BeginFrame(dt)
	CHECK(self.msg_listener ~= nil and type(self.msg_listener.OnRecordedEvent) == "function", "Invalid message listener")
		
	if self.state == GamePlayRecord.kState_Playing then
		self.player:BeginFrame(dt)
		local msg = self.player:IterationMessage()
		while msg do
			self.msg_listener:OnRecordedEvent(msg)
			msg = self.player:IterationMessage()
		end
		self.start = true
	elseif self.state == GamePlayRecord.kState_Recording then
		self.recorder:BeginFrame(dt)
		self.start = true
	elseif self.state == GamePlayRecord.kState_PlayRecord then
		self.recorder:BeginFrame(dt)
		self.player:BeginFrame(dt)
		local msg = self.player:IterationMessage()
		while msg do
			self.msg_listener:OnRecordedEvent(msg)
			msg = self.player:IterationMessage()
		end
		self.start = true
	end
end

function GamePlayRecord:_EndFrame(dt)
	if self.state == GamePlayRecord.kState_Playing then
		self.player:EndFrame(dt)
	elseif self.state == GamePlayRecord.kState_Recording then
		self.recorder:EndFrame(dt)
	elseif self.state == GamePlayRecord.kState_PlayRecord then
		self.recorder:EndFrame(dt)
		self.player:EndFrame(dt)
	end
end

function GamePlayRecord:BeginFrame(dt)
	if not self.start then
		self:_BeginFrame(dt)
	end
end

function GamePlayRecord:UpdateFrame(dt)
	if not self.start then
		self:_BeginFrame(dt)
	end
end

function GamePlayRecord:EndFrame(dt)
	if self.start then
		self:_EndFrame(dt)
		self.start = false
	end
end

return GamePlayRecord