--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================

if hl.IsConfigureDebug() then
    hl.ScriptVM_SetFlags(hl.ScriptVM_kForceEnableDebugLog)
end

local function Hulua_FileLoad(filename)
  local file
  if filename == nil then
      file = io.stdin
  else
      local err
      file, err = io.open(filename, "rb")
      if file == nil then
          error(("Unable to read '%s': %s"):format(filename, err))
      end
  end
  local data = file:read("*a")

  if filename ~= nil then
      file:close()
  end

  if data == nil then
      error("Failed to read " .. filename)
  end

  return data
end

local function Hulua_TableToString(t)  
    local cache = {}
    local result = ''
    local function _print_impl(t,indent)
        if cache[tostring(t)] then
            result = result .. indent.."*"..tostring(t) .. "\n"
        else
            cache[tostring(t)] = true
            if type(t) == 'table' then
                for pos, val in pairs(t) do
                    if type(val) == 'table' then
                        result = result .. indent .. "[" ..pos.. "] => " ..tostring(t) .. " {\n"
                        _print_impl(val,indent..string.rep(" ", string.len(pos)+8))
                        result = result .. indent .. string.rep(" ", string.len(pos) + 6) .. "}\n"
                    elseif type(val) == 'string' then
                        result = result .. indent .. "[" ..pos.. '] => "' ..val.. '"\n'
                    else
                        result = result .. indent.."["..pos.."] => "..tostring(val) .. "\n"
                    end
                end
            else
                result = result .. indent .. tostring(t) .. "\n"
            end
        end
    end
    if type(t) == 'table' then
        result = result .. tostring(t).." {\n"
        _print_impl(t, "  ")
        result = result .. "}\n"
    else
        _print_impl(t, "  ")
    end
    return result
end

local function Hulua_TableShallowCopy(t)
  local new
  if type(t) == 'table' then
    new = {}
    for k, v in pairs(t) do
      new[k] = v
    end
  else
    new = t
  end
  return new
end

local function Hulua_TableDeepCopy(t)
  local lookup = {}
  local function _copy_impl(o)
    if type(o) ~= 'table' then
      return o
    elseif lookup[o] then
      return lookup[o]
    end
    local new = {}
    lookup[o] = new
    for k, v in pairs(o) do
      new[_copy_impl(k)] = _copy_impl(v)
    end
    return setmetatable(new, getmetatable(o))
  end
  return _copy_impl(t)
end

local function Hulua_TableCopy(t, deep_copy)
  if deep_copy then
    return Hulua_TableDeepCopy(t)
  else 
    return Hulua_TableShallowCopy(t)
  end
end

local function Hulua_TableArrayFind(t, element_or_func)
  if type(element_or_func) == "function" then
    for k, v in ipairs(t) do
      if element_or_func(k, v) then
        return k
      end
    end
  else
    for k, v in ipairs(t) do
      if v == element_or_func then
        return k
      end
    end
  end
end
local function Hulua_TableArrayRemove(t, element_or_func)
  local key = Hulua_TableArrayFind(t, element_or_func)
  if key then
    table.remove(t, key)
  end
end
function Hulua_StringSplit(sep)
  local sep, fields = sep or "\t", {}
  local pattern = string.format("([^%s]+)", sep)
  self:gsub(pattern, function(c) fields[#fields+1] = c end)
  return fields
end

local function Hulua_ConnectArgs(...)
  local str = ""
  for k, v in ipairs({...}) do
    str = str .. tostring(v) .. " "
  end
  return str
end

local render_log_counter = 0
local function Hulua_LOGRClear()
  render_log_counter = 0
end

local function Hulua_LOGDR(...)
  local log = Hulua_ConnectArgs(...)
  hl.ScriptVM_LOGR(0, render_log_counter, hl.ScriptVM_kWrite, log)
  render_log_counter = render_log_counter + 1
end

local function Hulua_LOGIR(...)
  local log = Hulua_ConnectArgs(...)
  hl.ScriptVM_LOGR(0, render_log_counter, hl.ScriptVM_kGreen, log)
  render_log_counter = render_log_counter + 1
end

local function Hulua_LOGWR(...)
  local log = Hulua_ConnectArgs(...)
  hl.ScriptVM_LOGR(0, render_log_counter, hl.ScriptVM_kYellow, log)
  render_log_counter = render_log_counter + 1
end

local function Hulua_LOGER(...)
  local log = Hulua_ConnectArgs(...)
  hl.ScriptVM_LOGR(0, render_log_counter, hl.ScriptVM_kRed, log)
  render_log_counter = render_log_counter + 1
end

local function Hulua_LOGD(...)
  local log = Hulua_ConnectArgs(...)
  hl.ScriptVM_LOGD(log)
end
local function Hulua_LOGI(...)
  local log = Hulua_ConnectArgs(...)
  hl.ScriptVM_LOGI(log)
end
local function Hulua_LOGW(...)
  local log = Hulua_ConnectArgs(...)
  hl.ScriptVM_LOGW(log)
end
local function Hulua_LOGE(...)
  local log = Hulua_ConnectArgs(...)
  hl.ScriptVM_LOGE(log)
end

local function Hulua_CHECK(cond, ...)
  local log = Hulua_ConnectArgs(...)
  hl.ScriptVM_CHECK(cond, log)
end

function HuluaClass(classname, super)
  local superType = type(super)  
  local cls  

  if superType ~= "function" and superType ~= "table" then  
      superType = nil  
      super = nil  
  end

  if superType == "function" or (super and super.__ctype == 1) then  
      -- inherited from native C++ Object  
      cls = {}  

      if superType == "table" then  
          -- copy fields from super  
          for k,v in pairs(super) do cls[k] = v end

          cls.__create = super.__create  
          cls.super    = super  
      else  
        cls.__create = super  
        cls.ctor = function() end  
      end  

      cls.__cname = classname  
      cls.__ctype = 1  
      function cls.New(...)  
          local instance = cls.__create(...)
          -- copy fields from class to native object  
          for k,v in pairs(cls) do instance[k] = v end  
          
          instance.class = cls
          instance:ctor(...)
          return instance
      end
  else  
      -- inherited from Lua Object  
      if super then  
          cls = {}  
          setmetatable(cls, {__index = super})  
          cls.super = super
      else
          cls = { ctor = function() end }
      end

      cls.__cname = classname  
      cls.__ctype = 2 -- lua  
      cls.__index = cls

      function cls:New(...)
          local instance = setmetatable({}, cls)  
          instance.class = cls  
          instance:ctor(...)  
          return instance
      end  
  end  

  return cls  
end  

local HuluaClassify = {}

local Hulua_Import
local function Hulua_Define(name, o)
    if type(name) ~= 'string' then
        Hulua_LOGE("Class name type error: " .. "[" .. tostring(name) .. "]")
        return
    end
    local cls = HuluaClassify[name]
    if cls then
        return cls
    end

    if type(o) == 'string' then
      o = Hulua_Import(o)
    end

    cls = HuluaClass(name, o)
    HuluaClassify[name] = cls
    return cls
end
Hulua_Import = function (name)
    return Hulua_Define(name)
end
local function Hulua_UnDef(name)
    if HuluaClassify[name] then
        HuluaClassify[name] = nil
        package.loaded[name] = nil
    end
end
local G_GameFramework

function Initialize()
    local ok = G_GameFramework:Initialize()
    if ok then
        LOGD("script initialize success")
    else
        LOGD("script initialize failed")
    end
    return ok
end

function Frame(t)
    local ok = G_GameFramework:Update(t)
    return ok
end

function Event(msg)
    return G_GameFramework:Event(msg) or 0
end

function Shutdown()
    G_GameFramework:Shutdown()
    LOGD("script Shutdown success")
end

local Hulua_G_Mt = {}
Hulua_G_Mt.__newindex = function(t, k, v) Hulua_LOGE("Global variables cannot be used: ", k) end
Hulua_G_Mt.__index = function(t, k) Hulua_LOGE("Global variables cannot be used: ", k) end

import = Hulua_Import
define = Hulua_Define
undef = Hulua_UnDef
LOGD = Hulua_LOGD
LOGI = Hulua_LOGI
LOGW = Hulua_LOGW
LOGE = Hulua_LOGE
LOGRClear = Hulua_LOGRClear
LOGDR = Hulua_LOGDR
LOGIR = Hulua_LOGIR
LOGWR = Hulua_LOGWR
LOGER = Hulua_LOGER
CHECK = Hulua_CHECK
print = LOGI
error = LOGE

-- lua 5.2+
getfenv = false
setfenv = false
loadstring = false

table.copy = Hulua_TableCopy
table.tostring =  Hulua_TableToString
table.find = Hulua_TableArrayFind
table.erase = Hulua_TableArrayRemove
string.split = Hulua_StringSplit
io.load = Hulua_FileLoad

if hl.IsEditorMode() then
    Ed = {}
end

setmetatable(_G, Hulua_G_Mt)

if hl.IsEditorMode() then
    require("mobdebug").start()
end

require "InstanceCall"
G_GameFramework = require "GameFramework"
require "Action"
require "GamePlayRecord"
require "UI/Window"
require "Node"
require "NodeLayer"
require "ServiceManager"
require "RootNode"
require "NodeManager"
require "VideoPlayer"
local networkManager = require "NetworkManager"
networkManager:Init()

if hl.IsPlatformDesktop() then
    require "ViewerDesktop"
else
    require "ViewerMobile"
    require "ScaleGestureDetector"
    require "RotationGestureDetector"
end


LOGD("Hulua StartupTime", hl.Timer_StartupTime())

-- local function ShowRenderInfo()
-- 	local stats = hl.GPRenderDrive_GetRenderStats();

-- 	LOGDR(string.format("Frame %0.3f [ms], %0.3f FPS", 
-- 			stats.cpu_frameMS, 
-- 			1000.0 / stats.cpu_frameMS))
-- 	LOGDR(string.format("Submit CPU %0.3f [ms], GPU %0.3f [ms] (GPU driver latency: %d)", 
-- 			stats.cpu_timeMS, 
-- 			stats.gpu_timeMS,
-- 			stats.max_gpu_latency))
-- 	LOGDR(string.format("Backbuffer %dW x %dH in pixels, debug text %dW x %dH in characters", 
-- 			stats.width,
-- 			stats.height, 
-- 			stats.text_width,
-- 			stats.text_height))
-- 	LOGDR(string.format("DrawCall %d",
-- 			stats.num_draw))
-- 	LOGDR(string.format("Triangle %u", 
-- 			hl.GPRenderDrive_GetDrawPrimitiveCount(hl.GPIndexPrim_kTriangle)))

--     local view_stats
--     for i = 0, hl.GPRenderDrive_GetViewStatsCount() -1 do
-- 		view_stats = hl.GPRenderDrive_GetViewStats(i);
-- 		LOGDR("View [%d], Cpu %0.3f [ms], GPU %0.3f [ms]", 
-- 				view_stats.view, 
-- 				view_stats.cpu_elapsedMS, 
-- 				view_stats.gpu_elapsedMS)
-- 	end
-- end