--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================

local NetworkRequest = define("NetworkRequest")

local function UrlEncode(s)
    s = string.gsub(s, "([^%w%.%- ])", function(c) return string.format("%%%02X", string.byte(c)) end)
    return string.gsub(s, " ", "+")
end

local function UrlDecode(s)
    s = string.gsub(s, '%%(%x%x)', function(h) return string.char(tonumber(h, 16)) end)
    return s
end

-- TODO: make as enum
function NetworkRequest:SetScheme(scheme)
    self.scheme = scheme
    return self
end

function NetworkRequest:AsHttp()
    self.scheme = "http"
    return self
end

function NetworkRequest:SetHost(host)
    self.host = host
    return self
end

function NetworkRequest:SetPort(port)
    self.port = port
    return self
end

function NetworkRequest:SetPath(path)
    self.path = path
    return self
end

function NetworkRequest:AddQuery(key, value)
    if self.queries == nil then
        self.queries = {}
    end

    --self.queries[UrlEncode(key)] = UrlEncode(value)
    self.queries[key] = value
    return self
end

function NetworkRequest:SetFragment(fragment)
    --self.fragment = UrlEncode(fragment)
    self.fragment = fragment
end

function NetworkRequest:BuildUrl()
    if self.scheme == nil or
            self.host == nil then
        CHECK(false, "NetworkRequest's scheme/host is nil !!!", self.scheme, self.host)
        return nil
    end

    self.url = self.scheme .. "://" .. self.host
    if self.port ~= nil then
        self.url = self.url .. ":" .. self.port
    end

    if self.path ~= nil then
        self.url = self.url .. self.path
    end

    if self.queries ~= nil then
        local queries_str = ""
        for key, value in pairs(self.queries) do
            queries_str = queries_str .. key .. "=" .. value
        end

        self.url = self.url .. "?" .. queries_str
    end

    if self.fragment ~= nil then
        self.url = self.url .. "#" .. self.fragment
    end

    return self.url
end

local NetworkManager = define("NetworkManager")

function NetworkManager:Init()
    self.request_infos = {}
    self.request_id = 0
    LOGD("NetworkManager: Init ...")
end

local kReturnCodeSuccess = 0

function NetworkManager:OnResponse(request_id, response_code, response)
    if self.request_infos[request_id] == nil then
        LOGW("NetworkManager: OnResponse: request id:", request_id, "does not exist, may has been removed ...")
        return
    end

    local request_info = self.request_infos[request_id]
    self.request_infos[request_id] = nil

    local listener = request_info.listener

    if listener == nil then
        LOGD("NetworkManager: OnResponse: request id:", request_id, "has no listener, just ignore ...")
        return
    end

    if response_code == kReturnCodeSuccess then
        if listener.OnSuccess ~= nil then
            listener:OnSuccess(response)
        end
    else
        if listener.OnError ~= nil then
            listener:OnError(response_code, response)
        end
    end
end

function NetworkManager:RemoveListener(request_id)
    if self.request_infos[request_id] == nil then
        LOGW("NetworkManager: RemoveListener: request id:", request_id, "does not exist !!!")
        return
    end

    local request = self.request_infos[request_id].request
    LOGD("NetworkManager: RemoveListener: request id:", request_id, "url: ", request.url)
    self.request_infos[request_id] = nil
end

function NetworkManager:SendRequest(request, listener)
    local url = request:BuildUrl()
    if url == nil then
        return nil
    end

    local request_id = self.request_id
    self.request_id = self.request_id + 1

    if self.request_infos[request_id] ~= nil then
        CHECK(false, "NetworkManager: SendRequest:", self.url, "uuid is dup !!!, this should be impossible !!!")
        return nil
    end

    local request_info = {}
    request_info.listener = listener
    request_info.request = request
    LOGD("NetworkManager: SendRequest: url:", request.url)
    LOGD("NetworkManager: SendRequest: request id:", request_id)
    self.request_infos[request_id] = request_info

    local msg = hl.MessageQueue_MakeAsync()
    msg:PushBegin(hl.Hl_MSG_SCRIPT_SEND_NETWORK_REQUEST)
    msg:PushInt64(request_id)
    msg:PushString(request.url)
    msg:PushEnd()
    hl.MessageQueue_PostAndFreeAsync(msg)
end

return NetworkManager
