--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local Node = define "Node"

local Window = import "Window"
local GamePlayRecord = import "GamePlayRecord"
local NetworkManager = import "NetworkManager"

Node.kDefaultNodeOrder = 10000

function Node:GetNodeContext()
    return self.node_context
end

Node.kNormalMode = 1
Node.kPassiveMode = 2

function Node:Construct(type, id)
    if self.constructed then
        CHECK(false, "Node ", self.name, " has already constructed !!!")
        return
    end

    self.mode = self.kNormalMode

    self.type = type
    self.id = id

    self.node_layer_pos = -1

    self.constructed = true

    self.inited = false
    self.awaked = false
    self.name = self.type .. ":" .. self.id

    self:_InitLayerEnums()

    self.root_actor = hl.Actor_New()
    self.root_actor:SetName(hl.NamedView(self.name))
    hl.Class_Retain(self.root_actor)

    self.root_actor_group = hl.ActorGroup_New()
    self.root_actor_group:Collect(self.root_actor)
    hl.Class_Retain(self.root_actor_group)

    self.node_context = {}

    self.stash = {}

    self.listeners = {}

    self.enabled_services = {}

    self.allocated_camera_filter_masks = {}

    self.local_position = hl.Vec3(0, 0, 0)
    self.local_rotation = hl.Quat(0, 0, 0, 0)
    self.local_scale = hl.Vec3(1, 1, 1)

    self.fonts = {}
    self.system_texts = {}

    self.pending_network_requests = {}

    self.window = nil
end

function Node:InitWindow(renderable_layer)
    if self.window == nil then
        self.window = Window:New(self, "Window", renderable_layer)
    end
end

function Node:SendRequest(request, listener)
    local request_id = NetworkManager:SendRequest(request, listener)

    if request_id ~= nil then
        table.insert(self.pending_network_requests, request_id)
    end
end

function Node:GetSystemText()
    local res = hl.SysText_New()
    table.insert(self.system_texts, res)
    return res
end

function Node:GetFont(font_file_name, size, border, embolden)
    local font_key = font_file_name .. size .. border .. embolden
    local font = self.fonts[font_key]
    if font == nil then
        font =  hl.Font_New()
        if font:InitWithSource(self.node_work_dir / font_file_name,
                size, 1024, border, embolden, hl.Vec2_kOne, hl.Vec2_kZero) then
            self.fonts[font_key] = font
            hl.Class_Retain(font)
        else
            CHECK(false, "Node:", self:GetName(), "font:", font_key, "init fail !!!")
            hl.Class_Discard(font)
            font = nil
        end
    end

    return font
end

function Node:SetMode(mode)
    self.mode = mode
end

function Node:SetNodeLayerPos(node_layer_pos)
    self.node_layer_pos = node_layer_pos
end

function Node:GetNodeLayerPos()
    return self.node_layer_pos
end

function Node:GetRootActor()
    return self.root_actor
end

function Node:AllocCameraFilterMask()
    local new_mask = self.node_context.node_manager:AllocCameraFilterMask()
    if new_mask ~= nil then
        LOGD("Node ", self.name, "is allocated a filter mask", new_mask)
        CHECK(self.allocated_camera_filter_masks[new_mask] == nil,
                "Node ", self.name, "mask ", mask, "is already allocated !!!")
        local this = self
        self.allocated_camera_filter_masks[new_mask] = function ()
                                                this:DeallocCameraFilterMask(new_mask)
                                                    end
    end

    return new_mask
end

function Node:DeallocCameraFilterMask(mask)
    CHECK(self.allocated_camera_filter_masks[mask] ~= nil,
            "Node ", self.name, "mask ", mask, "is not allocated to this node!!!")
    self.allocated_camera_filter_masks[mask] = nil
    self.node_context.node_manager:DeallocCameraFilterMask(mask)
end

function Node:Initialize(node_context)
    if self.inited then
        CHECK(false, "Node ", self.name, " has already inited !!!")
        return
    end

    self.inited = true
    self.awaked = false

    self.node_context = node_context

    self:InitRWContext()

    if false == self:_OnInitialize() then
        CHECK(false, "Node ", self.name, "initialize fail !!!")
        return false
    end

    LOGD("Node ", self.name, "initialize ...")

    return true
end

function Node:EnableService(service_enum, enable)
    if enable then
        if self.enabled_services[service_enum] then
            LOGI("Node:", self.name, ", enable service:",
                    service_enum, "has already enabled ...")
            return
        else
            LOGD("Node", self.name, "enable service", service_enum)
            self.enabled_services[service_enum] = true
        end
    else
        if self.enabled_services[service_enum] == nil or self.enabled_services[service_enum] == false then
            LOGI("Node:", self.name, ", disable service:",
                    service_enum, " is not enabled yet ...")
            return
        else
            LOGD("Node", self.name, "disable service", service_enum)
            self.enabled_services[service_enum] = false
        end
    end

    self.node_context.node_manager:EnableService(service_enum, enable)
end

function Node:AddListener(event, listener, holder)
    if  self.listeners[event] ~= nil then
        CHECK(false, "Node ", self.name, "event", event, "listener already added !!!")
        return
    end

    LOGD("Node ", self.name, "event", event, "listener added ...")
    self.listeners[event] = listener

    self.node_context.game_framework:AddListener(event, listener, holder)
end

function Node:RemoveListener(event)
    if self.listeners[event] == nil then
        CHECK(false, "Node ", self.name, "event", event, "listener not added yet !!!")
        return
    end

    LOGD("Node ", self.name, "event", event, "listener removed ...")

    self.node_context.game_framework:RemoveListener(event, self.listeners[event])
    self.listeners[event] = nil
end

function Node:SetRenderLayer(renderable, layer)
    CHECK(self.node_context ~= nil, "Node ", self.name, "Set renderable layer, invalid node context!!!")
    CHECK(type(renderable) == "userdata", "Node ", self.name, "Set renderable layer, invalid renderable!!!")

    if layer == self.LAYER_BG_2D then
        renderable:SetFilterMask(self.node_context.camera_2d_bg:GetFilterMask())
    elseif layer == self.LAYER_3D_0 then
        renderable:SetFilterMask(self.node_context.camera_3d:GetFilterMask())
    elseif layer == self.LAYER_OVERLAY_2D then
        renderable:SetFilterMask(self.node_context.camera_2d_overlay:GetFilterMask())
    else
        CHECK(false, "Node ", self.name, "Set renderable layer, invalid layer!!!", layer)
    end
end

function Node:_InitLayerEnums()
    self.LAYER_BG_2D = 0
    self.LAYER_3D_0 = 1
    self.LAYER_OVERLAY_2D = 2
end

function Node:CreateSystem(system_ids)
    local world = self.node_context.node_manager:GetWorld()
    for _, system_id in pairs(system_ids) do
        world:CreateSystem(system_id)
    end
end

function Node:GetService(service_enum)
   return self.node_context.node_manager:GetService(service_enum)
end

function Node:GetDefaultTexture()
    return self.node_context.node_manager:GetDefaultTexture()
end

function Node:GetDefaultBlackTexture()
    return self.node_context.node_manager:GetDefaultBlackTexture()
end

local function _Image_Texture_Creator(image_path, ...)
    local new_tex = hl.ImageTexture_New()
    local args = table.pack(...)
    local init_res = false
    if args.n == 0 then
        init_res = new_tex:InitWithSource(image_path)
    elseif args.n == 1 then
        init_res = new_tex:InitWithSource(image_path, args[1])
    elseif args.n == 2 then
        init_res = new_tex:InitWithSource(image_path, args[1], args[2])
    else
        CHECK(false, "Node: _Image_Texture_Creator invalid args num !!!", args.n)
    end

    if init_res == false then
        LOGE("Node: _Image_Texture_Creator: init", image_path:CStr(), " fail ...")
        hl.Class_Discard(new_tex)
        new_tex = nil
    end

    return new_tex
end

function Node:NodeEvent(msg, message_source)
    return self:_OnNodeEvent(msg, message_source)
end

function Node:AllocImageTexture(name, path, ...)
    return self:_AllocTextureImpl(name, path, _Image_Texture_Creator, ...).obj
end

function Node:_AllocTextureImpl(name, path, texture_creator, ...)
    CHECK(type(name) == "userdata", "Node ", self.name, "_AllocTextureImpl invalid name !!!")
    CHECK(type(path) == "userdata", "Node ", self.name, "_AllocTextureImpl invalid path !!!")

    local stashed = self.stash[name]
    if nil == stashed then
        local new_tex = texture_creator(path, ...)
        if nil == new_tex then
            LOGE("Node:", self.name, ": _AllocTextureImpl: texture_creator: fail ...")
            return { obj = nil }
        end
        new_tex:SetName(name:View())
        hl.Class_Retain(new_tex)
        local new_stashed = { obj = new_tex }
        self.stash[name] = new_stashed
        return new_stashed
    elseif stashed.obj:TypeID() ~= hl.Texture_ID() then
        CHECK(false, "Node ", self.name, "stashed texture type", stashed.obj:TypeID(), "mismatch !!!")
        return { obj = nil }
    else
        CHECK(false, "Node ", self.name, "_AllocTextureImpl duplicate name !!!", name)
        return { obj = nil }
    end
end

function Node:_AllocActor(name)
    CHECK(type(name) == "userdata", "Node ", self.name, "AllocActor invalid name !!!")

    local stashed = self.stash[name]
    if nil == stashed then
        local new_actor = hl.Actor_New()
        new_actor:SetName(name:View())
        hl.Class_Retain(new_actor)
        local new_stashed = { obj = new_actor }
        self.stash[name] = new_stashed
        return new_stashed
    elseif stashed.obj:TypeID() ~= hl.Actor_ID() then
        CHECK(false, "Node ", self.name, "stashed actor type", stashed.obj:TypeID(), "mismatch !!!")
        return { obj = nil }
    else
        CHECK(false, "Node ", self.name, "_AllocActor duplicate name !!!", name)
        return { obj = nil }
    end
end

function Node:_Clone(name, obj, deep_copy)
    CHECK(type(name) == "userdata", "Node ", self.name, "_CloneImpl invalid name !!!")

    deep_copy = deep_copy or false

    if self.stash[name] == nil then
        local clone_obj = hl.Class_Clone(obj, deep_copy)
        hl.Class_Retain(clone_obj)
        local new_stashed = { obj = clone_obj }
        self.stash[name] = new_stashed
        return new_stashed
    else
        CHECK(false, "Node ", self.name, "_Clone duplicate name !!!", name)
        return { obj = nil }
    end
end

function Node:CloneActorGroup(name, actor_group, deep_copy)
    CHECK(actor_group:TypeID() == hl.ActorGroup_ID(), "Node ", self.name, "CloneActorGroup invalid actor_group !!!")

    local clone_stash = self:_Clone(name, actor_group, deep_copy)
    return hl.ActorGroup_Cast(clone_stash.obj)
end

function Node:CloneActor(name, actor, deep_copy)
    CHECK(actor:TypeID() == hl.Actor_ID(), "Node ", self.name, "CloneActor invalid actor !!!")

    local clone_stash = self:_Clone(name, actor, deep_copy)
    return hl.Actor_Cast(clone_stash.obj)
end

function Node:AllocActor(name)
    local stashed = self:_AllocActor(name)
    if stashed.obj == nil then
        return nil
    else
        return stashed.obj
    end
end

function Node:AllocSpriteActorWithTexture(name, texture)
    CHECK(texture ~= nil, "Node ", self.name, "AllocSpriteActorWithTexture invalid texture !!!")

    local material = hl.Sprite_CreateSimpleMaterial()
    material:SetBaseColorTex(texture)

    return self:AllocSpriteActor(name, material)
end

function Node:AllocSpriteActor(name, material)
    local stashed_sprite_actor = self:_AllocActor(name)

    if stashed_sprite_actor.obj == nil then
        return nil
    else
        local sprite = nil
        sprite = hl.Sprite_New()
        stashed_sprite_actor.obj:AddComponent(sprite)
        if material == nil then
            material = hl.Sprite_CreateSimpleMaterial()
            material:SetBaseColorTex(self.node_context.node_manager:GetDefaultTexture())
        end
        sprite:SetMaterial(material)
        return stashed_sprite_actor.obj, sprite
    end
end

local function FBX_ActorGroup_Creator(fbx_path)
    return hl.ActorGroup_CreateWithFBX(fbx_path)
end

local function DragonBone_ActorGroup_Creator(dragon_bone_path, ...)
    local args = table.pack(...)
    CHECK(args.n == 2, "DragonBone_ActorGroup_Creator: invalid args num", args.n)
    LOGD("DragonBone_ActorGroup_Creator", dragon_bone_path, args[1], args[2])
    return hl.ActorGroup_CreateWithDragonBone(dragon_bone_path, args[1], args[2])
end

local function Normal_ActorGroup_Creator()
    return hl.ActorGroup_New()
end

function Node:AllocActorGroup(name)
    CHECK(type(name) == "userdata", "Node ", self.name, "AllocActorGroup invalid name !!!")
    return self:_AllocActorGroup(name, nil, Normal_ActorGroup_Creator)
    
end

function Node:AllocActorGroupWithDragonBone(name, dragon_bone_path, filter_mask, layer)
    CHECK(type(dragon_bone_path) == "userdata", "Node ", self.name, "AllocActorGroupWithDragonBone invalid name !!!")
    return self:_AllocActorGroup(name, dragon_bone_path, DragonBone_ActorGroup_Creator, filter_mask, layer)
end

function Node:AllocActorGroupWithFBX(name, fbx_path, layer, material)
    layer = layer or 0
    CHECK(type(fbx_path) == "userdata", "Node ", self.name, "AllocActorGroupWithFBX invalid name !!!")
    local actor_group = self:_AllocActorGroup(name, fbx_path, FBX_ActorGroup_Creator)

    if nil == actor_group then
        return nil, nil
    else
        for i=0, actor_group:GetMemberCount()-1 do
            local actor = actor_group:GetMember(i)
            local com = actor:QueryComponent(hl.Renderable_ID(), true)
            if com then
                local mesh_com = hl.Renderable_Cast(com)
                if material ~= nil then
                    mesh_com:SetMaterial(material)
                end
                mesh_com:SetLayer(layer)
            end
        end
        return actor_group, material
    end
end

function Node:_AllocActorGroup(name, path, actor_group_creator, ...)
    local stashed = self:_AllocActorGroupImpl(name, path, actor_group_creator, ...)
    return stashed.obj
end

function Node:_AllocActorGroupImpl(name, path, actor_group_creator, ...)
    CHECK(type(name) == "userdata", "Node ", self.name, "AllocActorGroup invalid name !!!")
    
    local stashed = self.stash[name]
    if not stashed then
        local new_actor_group = actor_group_creator(path, ...)
        new_actor_group:SetName(name:View())
        hl.Class_Retain(new_actor_group)
        local new_stashed = {obj = new_actor_group}
        self.stash[name] = new_stashed
        return new_stashed
    elseif stashed.obj:TypeID() ~= hl.ActorGroup_ID() then
        CHECK(false, "Node ", self.name, "stashed actor group type", stashed.obj:TypeID(), "mismatch !!!")
        return { obj = nil }
    else
        CHECK(false, "Node ", self.name, "AllocActorGroup duplicate name !!!", name)
        return { obj = nil }
    end
end

function Node:Deallocate(name)
    if self.stash[name] == nil then
        CHECK(false, "Node:", self.name, "Deallocate:", name:CStr(), "does not exist")
        return
    else
        local stashed = self.stash[name];
        hl.Class_Release(stashed.obj)
        stashed.obj = nil
        self.stash[name] = nil
        LOGD("Node:", self.name, "Deallocate:", name:CStr())
    end
end

function Node:InitRWContext()
    self.json_context = hl.ClassRWContext()
    self.json_context.version = hl.Hl_ENGINE_VERSION
    self.json_context.min_version = hl.Hl_ENGINE_MINIMUM_VERSION
    self.json_context.work_dir = self:GetWorkDir()
    self.json_context.package_name = hl.Named(self:GetType())
    self.json_context.json = true

    self.context = hl.ClassRWContext()
    self.context.version = hl.Hl_ENGINE_VERSION
    self.context.min_version = hl.Hl_ENGINE_MINIMUM_VERSION
    self.context.work_dir = self:GetWorkDir()
    self.context.package_name = self.json_context.package_name
    self.context.json = false

    self.node_work_dir = self.context.work_dir / self.type
end

function Node:GetWorkDir()
    return self.node_context.node_manager:GetWorkDir()
end

function Node:GetRuntimeBuiltinDir()
    return self.node_context.node_manager:GetRuntimeBuiltinDir()
end

function Node:GetType()
    return self.type
end

function Node:GetName()
    return self.name
end

function Node:SetLocalPosition(position)
    self.local_position = position
    self.root_actor:GetTransform():SetLocalPosition(position)
end

function Node:SetLocalRotation(rotation)
    self.local_rotation = rotation
    self.root_actor:GetTransform():SetLocalRotation(rotation)
end

function Node:SetLocalScale(scale)
    self.local_scale = scale
    self.root_actor:GetTransform():SetLocalScale(scale)
end

function Node:GetLocalPosition()
    return self.local_position
end

function Node:GetLocalRotation()
    return self.local_rotation
end

function Node:GetLocalScale()
    return self.local_scale
end

function Node:GetId()
    return self.id
end

function Node:RemoveFromWorld(actor)
    local actor_class_id = actor:TypeID()
    if actor_class_id == hl.Actor_ID() then
        if self.awaked then
            self.node_context.node_manager:GetWorld():RemoveActor(actor)
        end
        actor:GetTransform():SetParent(nil)
    elseif actor_class_id == hl.ActorGroup_ID() then
        if self.awaked then
            actor:RemoveFromWorld(self.node_context.node_manager:GetWorld())
        end
        actor:GetRoot():GetTransform():SetParent(nil)
    else
        CHECK(false, "Node ", self.name,"remove invalid actor !!!")
        LOGE("Node ", self.name,"remove invalid actor !!!")
    end
end

function Node:AddToWorld(actor)
    local actor_class_id = actor:TypeID()
    if actor_class_id == hl.Actor_ID() then
        if self.awaked then
            self.node_context.node_manager:GetWorld():AddActor(actor)
        end
        actor:GetTransform():SetParent(self.root_actor:GetTransform())

    elseif actor_class_id == hl.ActorGroup_ID() then
        if self.awaked then
            actor:AddToWorld(self.node_context.node_manager:GetWorld())
        end
        actor:GetRoot():GetTransform():SetParent(self.root_actor:GetTransform())
    else
        CHECK(false, "Node ", self.name, "add invalid actor !!!")
    end
end

function Node:SwitchState(new_state)
    self:_OnSwitchState(new_state)
end

Node.kStringNameRecordType = 0
Node.kNumberNameRecordType = 1

function Node:StaticReadName(msg)
    local name_record_type = msg:PopInt8()

    local node_type = nil
    local node_id = nil
    if name_record_type == Node.kStringNameRecordType then
        node_type = msg:PopString()
        node_id = msg:PopInt32()
    elseif name_record_type == Node.kNumberNameRecordType then
        node_type = msg:PopInt32()
        node_type = tostring(node_type)
        node_id = msg:PopInt32()
    else
        CHECK(false, "Node: StaticReadName: unknown name_record_type:", name_record_type, "!!!")
    end

    return node_type, node_id
end

function Node:StaticWriteName(msg, node)
    msg:PushInt8(Node.kStringNameRecordType)
    msg:PushString(node:GetType())
    msg:PushInt32(node:GetId())
end

function Node:Snapshot()
    local record_message = self.node_context.node_manager:AllocRecordMessage(hl.NamedView(self.name))
    record_message:PushBegin(hl.Hl_MSG_SCRIPT_NODE_SNAPSHOT)
    Node:StaticWriteName(record_message, self)
    self:_OnSnapshot(record_message)
    record_message:PushEnd()
end

function Node:RestoreFromSnapshot(snapshot_msg)
    self:_OnRestoreFromSnapshot(snapshot_msg)
end

function Node:RequestRecordMessage()
    if self.node_context.state == GamePlayRecord.kState_Recording or
            self.node_context.state == GamePlayRecord.kState_PlayRecord then
        local record_message = self.node_context.node_manager:AllocRecordMessage(hl.NamedView(self.name))
        record_message:PushBegin(hl.Hl_MSG_SCRIPT_NODE_LOGIC)
        record_message:PushString(self.type)
        record_message:PushInt32(self.id)
        return record_message
    else
        return nil
    end
end

function Node:Update(dt)
    if self.mode == self.kNormalMode then
        self:_OnUpdate(dt)
    end

    self:_OnRefresh(dt)
    if self.window ~= nil then
        self.window:Update(dt)
    end
end

function Node:Awake(node_context)
    if self.awaked then
        return
    end

    self.awaked = true
    self.node_context = node_context

    self:_OnAwake()

    LOGD("Node ", self.name, "awake ...")
end

function Node:Sleep()
    if not self.awaked then
        return
    end

    self:_OnSleep()

    self.awaked = false

    self:DisableServices()

    LOGD("Node ", self.name, "sleep ...")
end

function Node:DisableServices()
    for service_enum, enabled in pairs(self.enabled_services) do
        if enabled then
            self:EnableService(service_enum, false)
            LOGD("Node", self.name, "DisableServices: disable service", service_enum)
        end
    end

    self.enabled_services = {}
end

function Node:Shutdown()
    self:_OnShutdown()
end

function Node:Destroy()
    if not self.constructed then
        return
    end

    self.constructed = false
    self.inited = false

    self:_ReleaseWindow()

    self:_ReleaseListeners()

    self:_ReleaseCameraFilterMasks()

    self:_ReleaseStash()

    self:_ReleaseFonts()

    self:_ReleaseSystemTexts()

    self.root_actor_group:Clear()
    hl.Class_Release(self.root_actor_group)
    hl.Class_Release(self.root_actor)
    self.root_actor = nil
    self.root_actor_group = nil

    self.node_layer_pos = -1

    self.node_context = nil

    LOGD("Node ", self.name, "shutdown ...")
end

function Node:_ReleaseWindow()
    if self.window ~= nil then
        self.window:Destroy()
        self.window = nil
    end
end

function Node:_ReleasePendingNetworkRequest()
    for _, request_id in ipairs(self.listeners) do
        NetworkManager:RemoveListener(request_id)
    end

    self.pending_network_requests = {}
end

function Node:_ReleaseSystemTexts()
    for _, text in ipairs(self.system_texts) do
        hl.Class_Discard(text)
    end

    self.system_texts = {}
end

function Node:_ReleaseFonts()
    for font_key, font in pairs(self.fonts) do
        LOGD("Node ", self.name, "release font:", font_key)

        hl.Class_Release(font)
    end

    self.fonts = {}
end

function Node:GetNodeOrder()
    return self.kDefaultNodeOrder
end

function Node:_ReleaseCameraFilterMasks()
    for mask, dealloc in pairs(self.allocated_camera_filter_masks) do
        if dealloc ~= nil then
            dealloc()
            LOGD("Node ", self.name, "release camera filter mask: ", mask)
        end
    end

    self.allocated_camera_filter_masks = nil
end

function Node:_ReleaseListeners()
    for event, listener in pairs(self.listeners) do
        if listener ~= nil then
            LOGD("Node ", self.name," remove event", event, "listener")
            self.node_context.game_framework:RemoveListener(event, listener)
        end
    end

    self.listeners = nil
end

function Node:_ReleaseStash()
    for name, stashed in pairs(self.stash) do
        LOGD("Node ", self.name," release stashed obj:",
                name:CStr())
        hl.Class_Release(stashed.obj)
        stashed.obj = nil
    end
    self.stash = {}

    return true
end

function Node:_OnInitialize()
    return true
end

function Node:_OnAwake()
end

function Node:_OnSleep()
end

function Node:_OnShutdown()
end

function Node:_OnSwitchState(new_state)
end

function Node:_OnNodeEvent(msg, message_source)
    return true
end

function Node:_OnMouseEvent()
end

function Node:_OnTouchEvent()
end

function Node:_OnViewEvent()
end

function Node:_OnUpdate(dt)
end

function Node:_OnSnapshot(snapshot_msg)
end

function Node:_OnRestoreFromSnapshot(snapshot_msg)
end

function Node:_OnRefresh(dt)
end

return Node