--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local Node = import ("Node")
local NodeLayer = define ("NodeLayer")

LOGD("NodeLayer has been defined ...")

function NodeLayer:SetName(name)
    self.name = name
end

function NodeLayer:GetName()
    return self.name
end

function NodeLayer:Initialize(node_manager, node_collection)
    self.node_manager = node_manager
    self.node_context = self.node_manager:GetNodeContext()

    self.ordered_node_list = {}
    self.node_collection = node_collection

    self:_OnInitialize()
    LOGD("NodeLayer:", self.name, "Initialize ...")
end

function NodeLayer:Shutdown()
    self:_OnShutdown()

    --NOTE: we can not delete item of self.ordered_node_list when iterating it
    --so just make a shallow copy
    local tmp_node_list = table.copy(self.ordered_node_list, false)

    for _, node in ipairs(tmp_node_list) do
        self:RemoveNode(node:GetType(), node:GetId())
    end

    self.ordered_node_list = {}
    tmp_node_list = nil

    LOGD("NodeLayer:", self.name, "Shutdown ...")
end

function NodeLayer:_RequireNode(node_type, node_id)
    hl.SetPackageName(hl.NamedView(node_type))
    undef(node_type)
    local node_script = require(node_type)

    if node_script == nil then
        CHECK(false, "NodeLayer: _RequireNode: require:", node_type, "fail !!!")
        return nil
    end

    local new_node = node_script:New()
    new_node:Construct(node_type, node_id)

    return new_node
end

function NodeLayer:AddNode(node_type, node_id)
    LOGD("NodeLayer:", self.name, "Try AddNode:", node_type, node_id)
    return self:_OnAddNode(node_type, node_id)
end

function NodeLayer:RemoveNode(node_type, node_id)
    LOGD("NodeLayer:", self.name, "Try RemoveNode:", node_type, node_id)
    return self:_OnRemoveNode(node_type, node_id)
end

function NodeLayer:ClearNodes()
    local tmp_node_list = table.copy(self.ordered_node_list)

    --NOTE: we can not delete item of self.ordered_node_list when iterating it
    --so just make a shallow copy
    for _, node in ipairs(tmp_node_list) do
        LOGD("NodeLayer:", self.name ,"ClearNodes: Remove:", node:GetType(), node:GetId())
        self:RemoveNode(node:GetType(), node:GetId())
    end

    self.ordered_node_list = {}
    tmp_node_list = nil
end

function NodeLayer:UpdateNodesPos()
    for pos, node in ipairs(self.ordered_node_list) do
        node:SetNodeLayerPos(pos)
    end
end

function NodeLayer:DisableServices()
    for _, node in ipairs(self.ordered_node_list) do
        node:DisableServices()
    end
end

function NodeLayer:DispatchNodeEvent(node_type, node_id, message_source)
    local node = self:QueryNode(node_type, node_id)
    if node == nil then
        return false
    end

    return node:NodeEvent(msg, message_source)
end

function NodeLayer:AddNodeToWorld(node)
    self.node_manager:AddNodeToWorld(node)
end

function NodeLayer:RemoveNodeFromWorld(node)
    self.node_manager:RemoveNodeFromWorld(node)
end

function NodeLayer:Update(dt)
    for _, node in ipairs(self.ordered_node_list) do
        node:Update(dt)
    end
end

function NodeLayer:SwitchState(new_state)
    self:_OnSwitchState(new_state)
end

function NodeLayer:GetNodeNum()
    return #self.ordered_node_list
end

function NodeLayer:QueryNode(node_type, node_id)
    return self.node_manager:QueryNode(node_type, node_id)
end

function NodeLayer:Snapshot(msg)
    for _, node in ipairs(self.ordered_node_list) do
        Node:StaticWriteName(msg, node)
        node:Snapshot()
    end
end

function NodeLayer:_OnInitialize()
end

function NodeLayer:_OnAddNode(node_type, node_id)
end

function NodeLayer:_OnRemoveNode(node_type, node_id)
end

function NodeLayer:_OnSwitchState(new_state)
end

function NodeLayer:_OnShutdown()
end

require "NormalNodeLayer"
require "EventRecordPlayNodeLayer"
require "StateRecordPlayNodeLayer"

return NodeLayer
