--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local NodeManager = define "NodeManager"
local ServiceManager = import("ServiceManager")
local RootNode = import("RootNode")
local NormalNodeLayer = import("NormalNodeLayer")
local EventRecordPlayNodeLayer = import("EventRecordPlayNodeLayer")
local StateRecordPlayNodeLayer = import("StateRecordPlayNodeLayer")
local GameFramework = import "GameFramework"
local GamePlayRecord = import "GamePlayRecord"
local Service = import "Service"

NodeManager.kForegroundNodeLayerEnum = 1
NodeManager.kStateRecordPlayNodeLayerEnum = 2
NodeManager.kEventRecordPlayNodeLayerEnum = 3
NodeManager.kBackgroundNodeLayerEnum = 4

function NodeManager:GetWorld()
    return self.world
end

function NodeManager:GetService(service_enum)
    return self.service_manager:GetService(service_enum)
end

function NodeManager:Initialize(game_play_record)
    LOGD("NodeManager Initialize ...")

    self.game_play_record = game_play_record
    CHECK(self.game_play_record ~= nil, "NodeManager: Initialize: game play record nil !!!")

    self.tag = hl.Named("NodeManager")

    self.message_recorder = nil
    self.state = GamePlayRecord.kState_Preview

    self.need_capture = false

    self.snapshot_name = hl.Named("Snapshot")

    self:_InitFixedTextures()
    self:_PrepareBuiltInResource()
    self:_PrepareMessageProcessor()

    self.world = hl.World_New()
    self.actor_group = hl.ActorGroup_New()
    self:_PrepareBuiltInSystems()

    hl.World_Awake(self.world)

    self.node_collection = {}

    self.service_manager = ServiceManager:New()

    self.root_node = RootNode:New()
    self.root_node:Construct(hl.Hl_SCRIPT_ROOT_NAME, hl.Hl_SCRIPT_ROOT_ID)
    self.root_node:AttachServiceManager(self.service_manager)
    self.node_collection[hl.Hl_SCRIPT_ROOT_NAME] = {num = 1, nodes = {}}
    self.node_collection[hl.Hl_SCRIPT_ROOT_NAME].nodes[hl.Hl_SCRIPT_ROOT_ID] = {node = self.root_node,
                                                                                pos = -1, node_layer = nil}

    self:_PrepareCameraFilterMaskAllocator()
    self:_PrepareDefaultCameras()
    self:_ConstructNodeContext()

    self.root_node:Initialize(self.node_context)
    self:AddNodeToWorld(self.root_node)

    self.node_layers = {}
    self.node_layer_map = {}
    self.foreground_node_layer = NormalNodeLayer:New()
    self.foreground_node_layer:SetName("Foreground")
    self.state_record_play_node_layer = StateRecordPlayNodeLayer:New()
    self.state_record_play_node_layer:SetName("StateRecordPlay")
    self.event_record_play_node_layer = EventRecordPlayNodeLayer:New()
    self.event_record_play_node_layer:SetName("EventRecordPlay")
    self.background_node_layer = NormalNodeLayer:New()
    self.background_node_layer:SetName("Background")

    table.insert(self.node_layers, self.background_node_layer)
    table.insert(self.node_layers, self.state_record_play_node_layer)
    table.insert(self.node_layers, self.event_record_play_node_layer)
    table.insert(self.node_layers, self.foreground_node_layer)
    self.node_layer_map[self.kForegroundNodeLayerEnum] = self.foreground_node_layer
    self.node_layer_map[self.kStateRecordPlayNodeLayerEnum] = self.state_record_play_node_layer
    self.node_layer_map[self.kEventRecordPlayNodeLayerEnum] = self.event_record_play_node_layer
    self.node_layer_map[self.kBackgroundNodeLayerEnum] = self.background_node_layer

    for _, node_layer in ipairs(self.node_layers) do
        node_layer:Initialize(self, self.node_collection)
    end

    LOGD("NodeManager Initialize end ...")
end

function NodeManager:AddNode(node_type, node_id, node_layer_enum)
    local add_success = false
    local added_node = nil

    local node_layer = self.node_layer_map[node_layer_enum]

    if node_layer == nil then
        CHECK(false, "NodeManager: AddNode:", node_type, node_id, node_layer_enum, "fail, node_layer is nil !!!")
        return false, nil
    end

    local node = self:QueryNode(node_type, node_id)
    if node ~= nil then
        LOGW("NodeManager: AddNode:", node_type, node_id, node_layer_enum, "fail, node has been already added !!!")
        return false, nil
    end

    add_success, added_node = node_layer:AddNode(node_type, node_id)

    CHECK(add_success, "NodeManager: AddNode:", node_type, node_id, node_layer_enum, "fail !!!")

    return add_success, added_node
end

function NodeManager:AddNodeImpl(new_node, node_layer_pos, node_layer)
    self:CollectNode(new_node, node_layer_pos, node_layer)
    new_node:Initialize(self.node_context)
    self:AddNodeToWorld(new_node)
end

function NodeManager:RemoveNodeImpl(node_type, node_id)
    local node_type_info = self.node_collection[node_type]
    if nil == node_type_info or node_type_info.num == 0 then
        CHECK(false, "NodeManager: RemoveNodeImpl, Node's type:", node_type, "can not be found, or has no instance!!!")
        return nil
    end

    local node_info = node_type_info.nodes[node_id]

    if nil == node_info then
        CHECK(false, "NodeManager: RemoveNodeImpl, Node:", node_type, "can not be found!!!")
        return nil
    end

    node_info.node:Shutdown()

    self:RemoveNodeFromWorld(node_info.node)

    local res_pos = node_info.node:GetNodeLayerPos()

    node_info.node:Destroy()

    node_type_info.nodes[node_id] = nil
    node_type_info.num = node_type_info.num - 1

    if node_type_info.num == 0 then
        package.loaded[node_type] = nil
        LOGD("NodeManager: RemoveNodeImpl: remove", node_type,  "from package.loaded")
    end

    return res_pos
end

function NodeManager:RemoveNode(node_type, node_id)

    local node, node_info = self:QueryNode(node_type, node_id)

    if node == self.root_node then
        CHECK(false, "NodeManager: RemoveNode:", node_type, node_id, ", root node can not be removed !!!")
        return false
    end

    if node == nil then
        LOGW("NodeManager: RemoveNode:", node_type, node_id, "fail, node does not exist !!!")
        return true
    end

    return node_info.node_layer:RemoveNode(node_type, node_id)
end

function NodeManager:GetWorkDir()
    return hl.GetWorkDir()
end

function NodeManager:GetRuntimeBuiltinDir()
    local builtin_class_rwContext = hl.GetBuiltinClassRWContext()
    return builtin_class_rwContext.work_dir / builtin_class_rwContext.package_name:CStr()
end

function NodeManager:_PrepareMessageProcessor()
    self.message_processors = {
        [hl.Hl_MSG_SCRIPT_INITIALIZE] = { func = self.InitializeScript, can_record = true },

        [hl.Hl_MSG_SCRIPT_SHUTDOWN] = { func = self.ShutdownScript, can_record = true },

        [hl.Hl_MSG_SCRIPT_RESET] = { func = self.Reset, can_record = true },

        --[hl.Hl_MSG_SCRIPT_HIERARCHY_CHANGED] = { func = self.ChangeScriptNodeHierarchy, can_record = true },

        [hl.Hl_MSG_SCRIPT_TRANSFORM_CHANGED] = { func = self.ChangeScriptNodeTransform, can_record = true },

        [hl.Hl_MSG_SCRIPT_POSITION_CHANGED] = { func = self.ChangeScriptNodePosition, can_record = true },

        [hl.Hl_MSG_SCRIPT_ROTATION_CHANGED] = { func = self.ChangeScriptNodeRotation, can_record = true },

        [hl.Hl_MSG_SCRIPT_SCALE_CHANGED] = { func = self.ChangeScriptNodeScale, can_record = true },

        [hl.Hl_MSG_SCRIPT_NODE_LOGIC] = { func = self.DispatchNodeEvent, can_record = false }
    }
end

function NodeManager:_PrepareCameraFilterMaskAllocator()
    self.camera_filter_mask_allocator = {}
    self.camera_filter_mask_num = 32
    local new_mask
    for index = 1, self.camera_filter_mask_num do
        new_mask = 1 << index
        self.camera_filter_mask_allocator[new_mask] = { mask = new_mask, allocated = false }
    end
    new_mask = nil
end

function NodeManager:AllocCameraFilterMask()
    local allocated_mask = nil

    for _, mask_info in pairs(self.camera_filter_mask_allocator) do
        if mask_info.allocated == false then
            mask_info.allocated = true
            allocated_mask = mask_info.mask
            break
        end
    end

    if nil == allocated_mask then
        CHECK(false, "NodeManager: no filter mask to alloc")
    else
        LOGD("NodeManager: alloc camera filter mask:", allocated_mask)
    end

    return allocated_mask
end

function NodeManager:DeallocCameraFilterMask(filter_mask)
    if self.camera_filter_mask_allocator[filter_mask] == nil or
            self.camera_filter_mask_allocator[filter_mask].allocated == false then
        CHECK(false, "NodeManager: dealloc invalid camera filter mask", filter_mask, "!!!")
        return
    end

    LOGD("NodeManager: dealloc camera filter mask:", filter_mask)
    self.camera_filter_mask_allocator[filter_mask].allocated = false
end

function NodeManager:EnableService(service_enum, enable)
    CHECK(self.root_node ~= nil, "NodeManager: EnableService: root node is nil !!!")
    self.service_manager:EnableService(service_enum, enable)
end

function NodeManager:_ConstructNodeContext()
    self.node_context = {}
    self.node_context.camera_2d_bg = self.default_bg_camera2d
    self.node_context.camera_3d = self.default_camera3d
    self.node_context.camera_2d_overlay = self.default_camera2d
    self.node_context.node_manager = self
    self.node_context.game_framework = GameFramework
end

function NodeManager:GetNodeContext()
    return self.node_context
end

function NodeManager:GetDefaultBGCamera2d()
    return self.default_bg_camera2d, self.default_bg_camera2d_actor
end

function NodeManager:GetDefaultCamera2d()
    return self.default_camera2d, self.default_camera2d_actor
end

function NodeManager:GetDefaultCamera3d()
    return self.default_camera3d, self.default_camera3d_actor
end

function NodeManager:GetDefaultTexture()
    return self.default_texture
end

function NodeManager:GetDefaultBlackTexture()
    return self.default_black_texture
end

function NodeManager:_PrepareBuiltInSystems()
    self.world:CreateSystem(hl.RenderSystem_ID())
end

function NodeManager:_PrepareDefaultCameras()
    self.default_bg_camera2d_name = hl.Named("Default_BG_Camera2D")
    self.default_bg_filter2d = self:AllocCameraFilterMask()
    CHECK(self.default_bg_filter2d ~= nil, "default_bg_filter2d get nil camera filter mask!!!")
    LOGD("default_bg_filter2d get camera filter mask ", self.default_bg_filter2d)
    self.default_bg_camera2d_order = 5000

    self.default_bg_camera2d_actor = self.root_node:AllocActor(self.default_bg_camera2d_name)
    self.root_node:AddToWorld(self.default_bg_camera2d_actor)
    self.default_bg_camera2d = hl.Camera_New()
    self.default_bg_camera2d:SetEnableOrthoMode(true)
    self.default_bg_camera2d:SetOrder(self.default_bg_camera2d_order)
    self.default_bg_camera2d:SetFilterMask(self.default_bg_filter2d)
    self.default_bg_camera2d_actor:AddComponent(self.default_bg_camera2d)
    self.default_bg_camera2d_actor:GetTransform():LookAtLocal(
            hl.Vec3_kAxisZ, hl.Vec3_kZero, hl.Vec3_kAxisY)
    
    self.default_camera2d_name = hl.Named("Default_Camera2D")
    self.default_filter2d = self:AllocCameraFilterMask()
    CHECK(self.default_filter2d ~= nil, "default_filter2d get nil camera filter mask!!!")
    LOGD("default_filter2d get camera filter mask ", self.default_filter2d)
    self.default_camera2d_order = 20000

    self.default_camera2d_actor = self.root_node:AllocActor(self.default_camera2d_name)
    self.root_node:AddToWorld(self.default_camera2d_actor)
    self.default_camera2d = hl.Camera_New()
    self.default_camera2d:SetEnableOrthoMode(true)
    self.default_camera2d:SetOrder(self.default_camera2d_order)
    self.default_camera2d:SetFilterMask(self.default_filter2d)
    self.default_camera2d_actor:AddComponent(self.default_camera2d)
    self.default_camera2d_actor:GetTransform():LookAtLocal(
            hl.Vec3_kAxisZ, hl.Vec3_kZero, hl.Vec3_kAxisY)

    self.default_camera3d_name = hl.Named("Default_Camera3D")
    self.default_camera3d_order = 10000
    LOGD("default_camera3d get camera filter mask 0")
    self.default_camera3d_clear_color = 0x00000000

    self.default_camera3d_actor = self.root_node:AllocActor(self.default_camera3d_name)
    self.root_node:AddToWorld(self.default_camera3d_actor)
    self.default_camera3d = hl.Camera_New()
    self.default_camera3d:SetOrder(self.default_camera3d_order)

    self.default_camera3d_actor:GetTransform():LookAtLocal(
            hl.Vec3_kAxisZ * 5.0, hl.Vec3_kZero, hl.Vec3_kAxisY)
    self.default_camera3d_actor:AddComponent(self.default_camera3d)

    if hl.IsPlatformDesktop() == false then
        local render_target = hl.RenderTarget_New()
        render_target:SetColorAttachment(0, self:GetFixedTexture(self.OUTPUT_TEXTURE_0, false))
        render_target:SetEnableDepth(true)
        self.default_bg_camera2d:SetRenderTarget(render_target)
        self.default_camera2d:SetRenderTarget(render_target)
        self.default_camera3d:SetRenderTarget(render_target)
    end
end

function NodeManager:_PrepareBuiltInResource()
    self.default_texture_name = hl.Named("default_texture")
    --TODO: tmp path
    local default_texture_path = hl.GetBootstrapPath() / "/dummy.png"
    self.default_texture = hl.ImageTexture_New()
    self.default_texture:InitWithSource(default_texture_path)
    hl.Class_Retain(self.default_texture)

    self.default_black_texture_name = hl.Named("default_black_texture")
    local default_black_texture_path = hl.GetBootstrapPath() / "/black.png"
    self.default_black_texture = hl.ImageTexture_New()
    self.default_black_texture:InitWithSource(default_black_texture_path)
    hl.Class_Retain(self.default_black_texture)

end

function NodeManager:_InitFixedTextures()
    self.INPUT_TEXTURE_0 = 0
    self.INPUT_TEXTURE_1 = 1
    self.INPUT_TEXTURE_2 = 2
    self.INPUT_TEXTURE_3 = 3
    self.OUTPUT_TEXTURE_0 = 0
    self.OUTPUT_TEXTURE_1 = 1
    self.OUTPUT_TEXTURE_2 = 2
    self.OUTPUT_TEXTURE_3 = 3
    self.fixed_input_textures = {}
    self.fixed_output_textures = {}
end

function NodeManager:QueryNode(node_type, node_id)
    local node_type_info = self.node_collection[node_type]

    if nil == node_type_info or node_type_info.num == 0 then
        return nil, nil
    end

    local node_info = node_type_info.nodes[node_id]
    if node_info == nil then
        return nil, nil
    end

    return node_info.node, node_info
end

function NodeManager:GetValidInstanceNum(type)
    if self.node_collection[type] == nil then
        return 0
    end

    return self.node_collection[type].num
end

function NodeManager:GetValidTypeNum()
    local valid_type_num = 0

    for _, type_info in pairs(self.node_collection) do
        if type_info.num > 0 then
            valid_type_num = valid_type_num + 1
        end
    end

    if valid_type_num > 0 then
        return valid_type_num - 1
    else
        return 0
    end
end

function NodeManager:NeedUpdate()
    return self.service_manager:GetStartedServiceNum() > 0 or self:GetValidTypeNum() > 0
end

function NodeManager:AddNodeToWorld(node)
    self.actor_group:Collect(node:GetRootActor())
    self.actor_group:AddToWorld(self.world)
    self.actor_group:Clear()
    if self.root_node ~= node then
        node:GetRootActor():GetTransform():SetParent(self.root_node:GetRootActor():GetTransform())
    end
    node:Awake(self.node_context)
end

function NodeManager:RemoveNodeFromWorld(node)
    node:Sleep()
    self.actor_group:Collect(node:GetRootActor())
    self.actor_group:RemoveFromWorld(self.world)
    self.actor_group:Clear()
end

function NodeManager:GetFixedTexture(fixed_texture_enum, is_input)
    local texture = nil

    if is_input == true then
        texture = self.fixed_input_textures[fixed_texture_enum]
        if texture == nil then
            texture = hl.FixedTexture_New()
            texture:SetID(fixed_texture_enum, is_input)
            hl.Class_Retain(texture)
            self.fixed_input_textures[fixed_texture_enum] = texture
        end
    else
        texture = self.fixed_output_textures[fixed_texture_enum]
        if texture == nil then
            texture = hl.FixedTexture_New()
            texture:SetID(fixed_texture_enum, is_input)
            hl.Class_Retain(texture)
            self.fixed_output_textures[fixed_texture_enum] = texture
        end
    end

    CHECK(texture ~= nil, "NodeManager: GetFixedTexture: fixed_texture_enum:",
            fixed_texture_enum, "is_input:", is_input, "get nil texture !!!")

    return texture
end

function NodeManager:Shutdown()
    LOGD("NodeManager Shutdown ...")

    self:_ShutdownNodeLayers()

    self:_ReleaseDefaultCameras()
    self:RemoveNodeFromWorld(self.root_node)
    self.root_node:Shutdown()
    self.root_node:Destroy()
    self.root_node = nil
    self.node_collection = nil

    hl.World_Sleep(self.world)
    self.actor_group:Clear()
    hl.Class_Discard(self.actor_group)
    self.actor_group = nil
    self.world = nil

    self:_ReleaseCameraFilterMask()
    self:_ReleaseBuiltInResource()
    self:_ReleaseFixedTextures()
    self.message_recorder = nil
end

function NodeManager:_ShutdownNodeLayers()
    for _, node_layer in ipairs(self.node_layers) do
        node_layer:Shutdown()
    end
end

function NodeManager:_ReleaseCameraFilterMask()
    for mask, mask_info in pairs(self.camera_filter_mask_allocator) do
        if mask_info.allocated == true then
            CHECK(false,"NodeManager: not all camera filter mask is deallocated !!!")
        end
    end

    self.camera_filter_mask_allocator = nil
end

function NodeManager:_ReleaseDefaultCameras()
    self:DeallocCameraFilterMask(self.default_camera2d:GetFilterMask())
    self.default_camera2d = nil
    self.default_camera2d_actor = nil
    self:DeallocCameraFilterMask(self.default_bg_camera2d:GetFilterMask())
    self.default_bg_camera2d = nil
    self.default_bg_camera2d_actor = nil

    self.default_camera3d = nil
    self.default_camera3d_actor = nil
end

function NodeManager:_ReleaseBuiltInResource()
    hl.Class_Release(self.default_texture)
    self.default_texture = nil

    hl.Class_Release(self.default_black_texture)
    self.default_black_texture = nil
end

function NodeManager:_ReleaseFixedTextures()
    for enum, texture in pairs(self.fixed_input_textures) do
        hl.Class_Release(texture)
        self.fixed_input_textures[enum] = nil
    end
    self.fixed_input_textures = nil

    for enum, texture in pairs(self.fixed_output_textures) do
        hl.Class_Release(texture)
        self.fixed_output_textures[enum] = nil
    end
    self.fixed_output_textures = nil
end

function NodeManager:SetState(new_state)
    LOGD("NodeManager: SetState:", new_state)
    local state_change = self.node_context.state ~= new_state
    local old_state = self.node_context.state

    if state_change then
        LOGD("NodeManager: SetState: state change: ->", new_state)

        local old_state_need_record = (old_state == GamePlayRecord.kState_Recording or old_state == GamePlayRecord.kState_PlayRecord)
        local new_state_need_record = (new_state == GamePlayRecord.kState_Recording or new_state == GamePlayRecord.kState_PlayRecord)

        if not old_state_need_record and new_state_need_record then
            LOGD("NodeManager: State switch in recording/play_record, enable face track and pose 3d ...")

            self.root_node:EnableService(Service.SERVICE_FACE_TRACK, true)
            local face_track_service = self:GetService(Service.SERVICE_FACE_TRACK)
            face_track_service:EnablePose3D(true)
        elseif old_state_need_record and not new_state_need_record then
            LOGD("NodeManager: State switch out recording/play_record, disable face track and pose 3d ...")

            self.root_node:EnableService(Service.SERVICE_FACE_TRACK, false)
            local face_track_service = self:GetService(Service.SERVICE_FACE_TRACK)
            face_track_service:EnablePose3D(false)
        else
        end
    end

    self.root_node:SwitchState(new_state)

    for _, node_layer in ipairs(self.node_layers) do
        node_layer:SwitchState(new_state)
    end

    self.node_context.state = new_state
end

function NodeManager:SetMessageRecorder(recorder)
    self.message_recorder = recorder
end

function NodeManager:OnA2SEvent(msg)

    local message_id = msg:PopBegin()
    local processor = self.message_processors[message_id]
    if processor == nil then
        LOGD("NodeManager: OnA2SEvent: Message: id:", message_id, "has no processor ..., just ignore it")
        return
    end

    processor.func(self, msg, GameFramework.kMessage_A2S)
end

function NodeManager:OnN2SEvent(msg)

    local message_id = msg:PopBegin()
    local processor = self.message_processors[message_id]
    if processor == nil then
        LOGD("NodeManager: OnN2SEvent: Message: id:", message_id, "has no processor ..., just ignore it")
        return
    end

    processor.func(self, msg, GameFramework.kMessage_N2S)
end

function NodeManager:OnRecordedEvent(msg)
    if self.game_play_record.play_type == GamePlayRecord.kMessageType0 then
        self.event_record_play_node_layer:OnRecordedEvent(msg)
    elseif  self.game_play_record.play_type == GamePlayRecord.kMessageType1 then
        self.state_record_play_node_layer:OnRecordedEvent(msg)
    else
       CHECK(false, "NodeManager: OnRecordedEvent: invalid game_play_record play type:",
               self.game_play_record.play_type, "!!!")
    end
end

function NodeManager:CollectNode(node, node_layer_pos, node_layer)
    local new_node_type = node:GetType()
    local new_node_id = node:GetId()

    local node_type_info = self.node_collection[new_node_type]

    if nil == node_type_info then
        node_type_info = {num = 0, nodes = {}}
        self.node_collection[new_node_type] = node_type_info
    end

    local node_info = node_type_info.nodes[new_node_id]

    if nil == node_info then
        node_info = {node = nil, node_layer = nil}
        node_type_info.nodes[new_node_id] = node_info
    end

    node_info.node = node
    node_info.node:SetNodeLayerPos(node_layer_pos)
    node_info.node_layer = node_layer
    node_type_info.num = node_type_info.num + 1
end

function NodeManager:AllocRecordMessage(name)
    return self.message_recorder:AllocMessage(name)
end

function NodeManager:Snapshot()
    local snapshot_msg = self:AllocRecordMessage(self.snapshot_name:View())

    local node_num = 0
    for _, node_layer in ipairs(self.node_layers) do
        node_num = node_num + node_layer:GetNodeNum()
    end
    snapshot_msg:PushBegin(hl.Hl_MSG_SCRIPT_SNAPSHOT)
    snapshot_msg:PushInt32(node_num)

    for _, node_layer in ipairs(self.node_layers) do
        node_layer:Snapshot(snapshot_msg)
    end

    snapshot_msg:PushEnd()
end

function NodeManager:UpdateFrame(dt)
    self.root_node:BeginFrame(dt)

    self.root_node:Update(dt)

    for _, node_layer in ipairs(self.node_layers) do
        node_layer:Update(dt)
    end

    self.root_node:EndFrame(dt)

    if self.node_context.state == GamePlayRecord.kState_Recording or
            self.node_context.state == GamePlayRecord.kState_PlayRecord then
       self:Snapshot()
    end
end

function NodeManager:DispatchNodeEvent(msg, message_source)
    local node_type = msg:PopString()
    local node_id = msg:PopInt32()

    local node = self:QueryNode(node_type, node_id)
    if node == nil then
        return false
    end

    return node:NodeEvent(msg, message_source)
end

function NodeManager:ChangeScriptNodeScale(msg, message_source)
    local node_type = msg:PopString()
    local node_id = msg:PopInt32()

    local node = self:QueryNode(node_type, node_id)
    if node == nil then
        return false
    end

    node:SetLocalScale(hl.Vec3(msg:PopFloat(), msg:PopFloat(), msg:PopFloat()))

    return true
end

function NodeManager:ChangeScriptNodeRotation(msg, message_source)
    local node_type = msg:PopString()
    local node_id = msg:PopInt32()

    local node = self:QueryNode(node_type, node_id)
    if node == nil then
        return false
    end

    node:SetLocalRotation(hl.Quat(hl.Vec3(msg:PopFloat(), msg:PopFloat(), msg:PopFloat())))

    return true
end

function NodeManager:ChangeScriptNodePosition(msg, message_source)
    local node_type = msg:PopString()
    local node_id = msg:PopInt32()

    local node = self:QueryNode(node_type, node_id)
    if node == nil then
        return false
    end

    node:SetLocalPosition(hl.Vec3(msg:PopFloat(), msg:PopFloat(), msg:PopFloat()))

    return true
end

function NodeManager:ChangeScriptNodeTransform(msg, message_source)
    local node_type = msg:PopString()
    local node_id = msg:PopInt32()

    local node = self:QueryNode(node_type, node_id)
    if node == nil then
        return false
    end

    local pos_x = msg:PopFloat()
    local pos_y = msg:PopFloat()
    local pos_z = msg:PopFloat()
    local rot_x = msg:PopFloat()
    local rot_y = msg:PopFloat()
    local rot_z = msg:PopFloat()
    local scl_x = msg:PopFloat()
    local scl_y = msg:PopFloat()
    local scl_z = msg:PopFloat()

    node:SetLocalPosition(hl.Vec3(pos_x, pos_y, pos_z))
    node:SetLocalRotation(hl.Quat(hl.Vec3(rot_x, rot_y, rot_z)))
    node:SetLocalScale(hl.Vec3(scl_x, scl_y, scl_z))

    return true
end

function NodeManager:Reset(msg, message_source)
    local reset_flags = msg:PopInt64()
    LOGD("NodeManager: Reset: reset_flags:", reset_flags)

    if reset_flags&hl.Hl_MSG_SCRIPT_RESET_CLEAR_SCRIPTS_FLAG ~= 0 then
        LOGD("NodeManager: Reset: Hl_MSG_SCRIPT_RESET_CLEAR_SCRIPTS_FLAG ...")

        for _, node_layer in ipairs(self.node_layers) do
            node_layer:ClearNodes()
        end
    end

    if reset_flags&hl.Hl_MSG_SCRIPT_RESET_STOP_SERVICES_FLAG ~= 0 then
        LOGD("NodeManager: Reset: Hl_MSG_SCRIPT_RESET_STOP_SERVICES_FLAG ...")

        for _, node_layer in ipairs(self.node_layers) do
            node_layer:DisableServices()
        end

        self.root_node:DisableServices()

        --TODO: show_bg service is always on ...
        self.root_node:EnableService(Service.SERVICE_SHOW_BG, true)
    end

    return true
end

function NodeManager:ShutdownScript(msg, message_source)
    local node_type = msg:PopString()
    local node_id = msg:PopInt32()
    LOGD("NodeManager: ShutdownScript", node_type, node_id)

    return self:RemoveNode(node_type, node_id)
end

function NodeManager:InitializeScript(msg, message_source)
    local node_num = msg:PopUint32()

    LOGD("NodeManager: InitializeScript: will add", node_num, "nodes ...")
    CHECK(node_num > 0, "NodeManager: InitializeScript: node_num:", node_num, "must > 0 !!!")

    local node_type = nil
    local node_id = nil
    local node_layer_enum = nil

    local tmp_node = nil
    local added_nodes = {}

    local add_success = true

    local pos_x = nil
    local pos_y = nil
    local pos_z = nil
    local rot_x = nil
    local rot_y = nil
    local rot_z = nil
    local scl_x = nil
    local scl_y = nil
    local scl_z = nil

    for node_index = 1, node_num do
        node_type = msg:PopString()
        node_id = msg:PopInt32()
        node_layer_enum = msg:PopInt32()

        add_success, tmp_node = self:AddNode(node_type, node_id, node_layer_enum)

        if not add_success then
            LOGW("NodeManager: InitializeScript:", node_type, node_id, "add fail !!!")
            break
        else
            table.insert(added_nodes, tmp_node)
        end

        pos_x = msg:PopFloat()
        pos_y = msg:PopFloat()
        pos_z = msg:PopFloat()
        rot_x = msg:PopFloat()
        rot_y = msg:PopFloat()
        rot_z = msg:PopFloat()
        scl_x = msg:PopFloat()
        scl_y = msg:PopFloat()
        scl_z = msg:PopFloat()

        tmp_node:SetLocalPosition(hl.Vec3(pos_x, pos_y, pos_z))
        tmp_node:SetLocalRotation(hl.Quat(hl.Vec3(rot_x, rot_y, rot_z)))
        tmp_node:SetLocalScale(hl.Vec3(scl_x, scl_y, scl_z))
    end

    if not add_success then
        LOGW("NodeManager: InitializeScript: fail, remove added nodes")

        for _, node in ipairs(added_nodes) do
            self:RemoveNode(node:GetType(), node:GetId())
        end
    else
        LOGD("NodeManager: InitializeScript: success, add", node_num, "nodes ")
    end

    added_nodes = nil

    return add_success
end

return NodeManager