--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local RotationGestureDetector = define "RotationGestureDetector"

function RotationGestureDetector:Initialize(listener)
	assert(listener)
	self.listener = listener

	self:_ResetState()

	self.pointNum = 0
end

function RotationGestureDetector:IsProgress()
	return self.progress
end

function RotationGestureDetector:Shutdown()
	self.listener = nil
end

function RotationGestureDetector:GetTotalRotation()
	return -self.curRotation
end

function RotationGestureDetector:GetRotation()
	return -self.curAverageRotationDiff
end

function RotationGestureDetector:_ResetState()
	self.curAverageRotationDiff = 0.0
	self.curRotation = 0.0
	self.curPointRotations = {}
	self.touches = {}
	self.progress = false
end

function RotationGestureDetector:_RefreshAverageRotation(touches)
	local sumX, sumY = 0, 0

	for _, touch in pairs(touches) do
		sumX = sumX + touch.x
		sumY = sumY + touch.y
	end
	local focusX = sumX / self.pointNum
	local focusY = sumY / self.pointNum

	local rotationDiffSum = 0

	local availablePointNum = 0

	for _, touch in pairs(touches) do
		local prevPointRotation = self.curPointRotations[touch.id]
		if nil ~= prevPointRotation then
			availablePointNum = availablePointNum + 1

			local curPointRotation = self:_getPointRadius(touch.x - focusX, touch.y - focusY)

			local curPointRotationDiff = curPointRotation - self.curPointRotations[touch.id]

			if curPointRotationDiff > math.pi or curPointRotationDiff < -math.pi then
				if curPointRotationDiff > math.pi then
					curPointRotationDiff = curPointRotationDiff - math.pi * 2
				else
					curPointRotationDiff = math.pi * 2 + curPointRotationDiff
				end
			end

			rotationDiffSum = rotationDiffSum + curPointRotationDiff

			self.curPointRotations[touch.id] = curPointRotation
		end
	end

	self.curAverageRotationDiff = rotationDiffSum / availablePointNum
	self.curRotation = self.curRotation + self.curAverageRotationDiff

end

function RotationGestureDetector:_getPointRadius(x, y)
	local radius = math.atan(y, x)

	if radius < 0 then
		radius = radius + math.pi * 2
	end

	return radius
end

function RotationGestureDetector:_RefreshPointsRotation(touches)
	assert(self.pointNum > 0,
			"RotationGestureDetector: _RefreshPointsRotation fail, point num is " .. self.pointNum)

	local sumX, sumY = 0, 0

	for _, touch in pairs(touches) do
		sumX = sumX + touch.x
		sumY = sumY + touch.y
	end

	local focusX = sumX / self.pointNum
	local focusY = sumY / self.pointNum

	for _, touch in pairs(touches) do
		self.curPointRotations[touch.id] = self:_getPointRadius(touch.x - focusX, touch.y - focusY)
	end
end

function RotationGestureDetector:TouchEvent(action, touches)
	if action == hl.Hl_ACTION_TOUCH_BEGAN then
		self:_TouchBegan(touches)
	elseif action == hl.Hl_ACTION_TOUCH_MOVED then
		self:_TouchMoved(touches)
	elseif action == hl.Hl_ACTION_TOUCH_ENDED then
		self:_TouchEnded(touches)
	elseif action == hl.Hl_ACTION_TOUCH_CANCELLED then
		self:_TouchCancelled(touches)
	end
end
function RotationGestureDetector:_TouchBegan(touches)
	assert(self.pointNum >= 0,
			"RotationGestureDetector: touch begin, but the point num is " .. self.pointNum)

	for _, touch in pairs(touches) do
		if nil == self.touches[touch.id] then
			self.touches[touch.id] = { id = touch.id, x = touch.x, y = touch.y }
		else
			self.touches[touch.id].id = touch.id
			self.touches[touch.id].x = touch.x
			self.touches[touch.id].y = touch.y
		end
		self.pointNum = self.pointNum + 1
	end

	if self.pointNum > 1 then
		self:_RefreshPointsRotation(self.touches)
		if true ~= self.progress then
			self.progress = self.listener:OnRotateBegin(self)
		end
	end
end

function RotationGestureDetector:_TouchMoved(touches)
	self:_UpdateTouches(touches)

	if self.progress then
		self:_RefreshAverageRotation(self.touches)
		self.listener:OnRotate(self)
	end
end

function RotationGestureDetector:_TouchEnded(touches)
	assert(self.pointNum > 0,
			"RotationGestureDetector: touch end, but the point num is " .. self.pointNum)

	for _, touch in pairs(touches) do
		self.touches[touch.id] = nil
		self.curPointRotations[touch.id] = nil

		self.pointNum = self.pointNum - 1
	end

	if self.progress then
		if self.pointNum <= 1 then
			self.listener:OnRotateEnd(self)
			self:_ResetState()
		else
			self:_RefreshPointsRotation(self.touches)
		end
	end
end

function RotationGestureDetector:_UpdateTouches(touches)
	for _, touch in pairs(touches) do
		if nil == self.touches[touch.id] then
			self.touches[touch.id] = { id = touch.id, x = touch.x, y = touch.y }
		else
			self.touches[touch.id].id = touch.id
			self.touches[touch.id].x = touch.x
			self.touches[touch.id].y = touch.y
		end
	end
end

function RotationGestureDetector:_TouchCancelled(touches)
	self.listener:OnRotateEnd(self)
	self:_ResetState()
	self.pointNum = 0
end

return RotationGestureDetector