--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local ScaleGestureDetector = define "ScaleGestureDetector"

function ScaleGestureDetector:Initialize(listener)
	assert(listener)
	self.listener = listener

	self.focusX = nil
	self.focusY = nil
	self.curr_spanX = 0
	self.curr_spanY = 0
	self.curr_span = 0

	self.prev_spanX = 0
	self.prev_spanY = 0
	self.prev_span = 0
	self.curr_time = 0
	self.prev_time = 0
	self.progress = false
	self.touch_num_change = false
	self.touches = {}
	self.last = {}
end

function ScaleGestureDetector:Shutdown()
	self.listener = nil
end

function ScaleGestureDetector:IsProgress()
	return self.progress
end

function ScaleGestureDetector:GetScaleFactor()
	if self.prev_span > 0 then 
		return self.curr_span / self.prev_span
	else
		return 1
	end
end

function ScaleGestureDetector:TouchEvent(action, touches)
	if action == hl.Hl_ACTION_TOUCH_BEGAN then
		self:_TouchBegan(touches)
	elseif action == hl.Hl_ACTION_TOUCH_MOVED then
		self:_TouchMoved(touches)
	elseif action == hl.Hl_ACTION_TOUCH_ENDED then
		self:_TouchEnded(touches)
	elseif action == hl.Hl_ACTION_TOUCH_CANCELLED then
		self:_TouchCancelled(touches)
	end
end
function ScaleGestureDetector:_TouchBegan(touches)
	for _, touch in pairs(touches) do
		table.insert(self.touches, { id = touch.id, x = touch.x, y = touch.y })
	end
	if #self.touches > 1 then
		if self.progress then
			self.touch_num_change = true
		else
			self:_ComputeSpan(self.touches)
			self.prev_spanX = self.curr_spanX
			self.prev_spanY = self.curr_spanY
			self.prev_span = self.curr_span
			self.progress = self.listener:OnScaleBegin(self)
		end
	end
end

function ScaleGestureDetector:_TouchMoved(touches)
	if self.progress then
		self:_ComputeSpan(touches)
		local update_prev_span = true
		if self.touch_num_change then
			self.prev_spanX = self.curr_spanX
			self.prev_spanY = self.curr_spanY
			self.prev_span = self.curr_span
			self.touch_num_change = false
			update_prev_span = false
		end
	
		self.listener:OnScale(self)
	
		if update_prev_span then
			self.prev_spanX = self.curr_spanX
			self.prev_spanY = self.curr_spanY
			self.prev_span = self.curr_span
		end
	end
end

function ScaleGestureDetector:_TouchEnded(touches)
	for _, touch in ipairs(touches) do
		for i, cached_touch in ipairs(self.touches) do
			if touch.id == cached_touch.id then
				table.remove(self.touches, i)
				break
			end
		end
	end

	if self.progress and #self.touches <= 1 then
		self:_ComputeSpan(self.touches)
		self.prev_spanX = self.curr_spanX
		self.prev_spanY = self.curr_spanY
		self.prev_span = self.curr_span
		self.listener:OnScaleEnd(self)
		self.progress = false
	else
		self.touch_num_change = true
	end
end

function ScaleGestureDetector:_TouchCancelled(touches)
	if self.progress then
		self:_ComputeSpan(touches)
		self.prev_spanX = self.curr_spanX
		self.prev_spanY = self.curr_spanY
		self.prev_span = self.curr_span
		self.listener:OnScaleEnd(self)
		self.progress = false
	end
	self.touches = {}
end

function ScaleGestureDetector:_ComputeSpan(touches)
	local sumX, sumY, div = 0, 0, #touches
	for _, touch in ipairs(touches) do
		sumX = sumX + touch.x
		sumY = sumY + touch.y
	end
	local focusX = sumX / div
	local focusY = sumY / div

	local dev_sumX, dev_sumY = 0, 0
	for _, touch in ipairs(touches) do
		dev_sumX = dev_sumX + math.abs(touch.x - focusX)
		dev_sumY = dev_sumY + math.abs(touch.y - focusY)
	end
	local devX = dev_sumX / div
	local devY = dev_sumY / div

	local spanX = devX * 2
	local spanY = devY * 2
	local span = math.sqrt(spanX * spanX + spanY * spanY)

	self.curr_spanX = spanX
	self.curr_spanY = spanY
	self.curr_span = span

	self.focusX = focusX
	self.focusY = focusY
end

return ScaleGestureDetector