--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local Service = define "Service"

LOGD("Service has been defined ...")

Service.SERVICE_NONE = -1
Service.SERVICE_SHOW_BG = 0
Service.SERVICE_FACE_TRACK = 1
Service.SERVICE_FILTER = 2
Service.SERVICE_BEAUTY = 3
Service.SERVICE_FACE_RESHAPE = 4
Service.SERVICE_OCCLUDER = 5
Service.SERVICE_MULTI_OUTPUT = 6
Service.SERVICE_FONT = 7

Service.MAX_FACE_NUM = 8

function Service:Initialize(name, node)
    self.user_count = 0
    self.name = name
    self.attach_node = node
    self.stopped = true
    self.service_type = self.SERVICE_NONE
    self.enabled_services = {}
    self.listeners = {}
    self.node_manager = self.attach_node.node_context.node_manager
    LOGD("Service:", self.name:CStr(), "initialize ...")
    self:_OnInitialize()
end

function Service:Enable(enable)
    if enable then
        self.user_count = self.user_count + 1
        if self.user_count == 1 then
            LOGD("Service", self.name:CStr(), "start ...")
            self.stopped = false
            self:_OnStart()
        end
    else
        self.user_count = self.user_count - 1
        if self.user_count == 0 then
            LOGD("Service", self.name:CStr(), "stop ...")
            self.stopped = true
            self:_OnStop()
            self:_ReleaseListeners();
        end
    end

    LOGD("Service", self.name:CStr(), "enable:", enable, "user count:", self.user_count)

    if self.user_count < 0 then
        CHECK(false, "Service", self.name:CStr(), "user_count < 0 error !!!", self.user_count)
        self.user_count = 0
    end
end

function Service:Awake()
    LOGD("Service:", self.name:CStr(), "awake, user count:", self.user_count)
    self:_OnAwake()
end

function Service:GetService(service_enum)
    return self.attach_node.node_context.node_manager:GetService(service_enum)
end

function Service:EnableService(service_enum, enable)
    if service_enum == self.service_type then
        CHECK(false, "Service:", self.service_type, "can not enable/disable itself !!!")
        return
    end

    if enable then
        if self.enabled_services[service_enum] then
            LOGI("Service:", self.name:CStr(), ", enable service:",
                    service_enum, "has already enabled ...")
            return
        else
            LOGD("Service", self.name:CStr(), "enable service", service_enum)
            self.enabled_services[service_enum] = true
        end
    else
        if self.enabled_services[service_enum] == nil or self.enabled_services[service_enum] == false then
            LOGI("Service:", self.name:CStr(), ", disable service:",
                    service_enum, " is not enabled yet ...")
            return
        else
            LOGD("Service", self.name:CStr(), "disable service", service_enum)
            self.enabled_services[service_enum] = false
        end
    end

    self.attach_node.node_context.node_manager:EnableService(service_enum, enable)
end

function Service:DisableServices()
    for service_enum, enabled in pairs(self.enabled_services) do
        if enabled then
            self:EnableService(service_enum, false)
            LOGD("Service", self.name:CStr(), "DisableServices: disable service", service_enum)
        end
    end

    self.enabled_services = {}
end

function Service:AllocRecordMessage()
    CHECK(self.service_type ~= self.attach_node.node_context.node_manager.SERVICE_NONE, "Service:", self.name:CStr(), "type is SERVICE_NONE !!!")

    local record_message = self.attach_node.node_context.node_manager:AllocRecordMessage(self.name:View())
    record_message:PushBegin(hl.Hl_MSG_SCRIPT_NODE_LOGIC)
    record_message:PushString(self.attach_node:GetType())
    record_message:PushInt32(self.attach_node:GetId())
    record_message:PushInt32(hl.Hl_MSG_SCRIPT_NODE_LOGIC_CONFIGURE_SERVICE)
    record_message:PushInt32(self.service_type)

    return record_message
end

function Service:IsStopped()
    return self.stopped
end

function Service:Update(dt)
    if not self.stopped then
        self:_OnUpdate(dt)
    end
end

function Service:SwitchState(new_state)
    self:_OnSwitchState(new_state)
end

function Service:NodeEvent(msg, message_source)
    self:_OnNodeEvent(msg, message_source)
end

function Service:AddToWorld(actor)
    self.attach_node:AddToWorld(actor)
end

function Service:RemoveFromWorld(actor)
    self.attach_node:RemoveFromWorld(actor)
end

function Service:GetWorkDir()
    return self.attach_node:GetWorkDir()
end

function Service:AllocActor(name)
    return self.attach_node:AllocActor(name)
end

function Service:GetRuntimeBuiltinDir()
    return self.attach_node:GetRuntimeBuiltinDir()
end

function Service:SetRenderLayer(renderable, layer)
    self.attach_node:SetRenderLayer(renderable, layer)
end

function Service:CreateSystem(system_ids)
    self.attach_node:CreateSystem(system_ids)
end

function Service:AddListener(event, listener, holder)
    if  self.listeners[event] ~= nil then
        CHECK(false, "Service ", self.name, "event", event, "listener already added !!!")
        return
    end

    LOGD("Service ", self.name, "event", event, "listener added ...")
    self.listeners[event] = listener

    self.attach_node.node_context.game_framework:AddListener(event, listener, holder)
end

function Service:RemoveListener(event)
    if self.listeners[event] == nil then
        CHECK(false, "Service ", self.name, "event", event, "listener not added yet !!!")
        return
    end

    LOGD("Service ", self.name, "event", event, "listener removed ...")

    self.attach_node.node_context.game_framework:RemoveListener(event, self.listeners[event])
    self.listeners[event] = nil
end

function Service:_ReleaseListeners()
    LOGD("Service ", self.name," release listeners ...")

    for event, listener in pairs(self.listeners) do
        if listener ~= nil then
            self:RemoveListener(event)
        end
    end

    self.listeners = {}
end

function Service:Sleep()
    LOGD("Service:", self.name:CStr(), "sleep, user count:", self.user_count)
    self:_OnSleep()

    self:DisableServices()
end

function Service:Shutdown()
    CHECK(self.user_count == 0, "Service", self.name:CStr(), "is still enabled by someone, pls disable !!!")
    LOGD("Service:", self.name:CStr(), "shutdown ..., user_count:", self.user_count)

    self.stopped = true

    self:_OnShutdown()

    self.attach_node = nil
    self.user_count = 0
end

function Service:_OnInitialize()
end

function Service:_OnAwake()
end

function Service:_OnSleep()
end

function Service:_OnStart()
end

function Service:_OnStop()
end

function Service:_OnUpdate(dt)
end

function Service:_OnShutdown()
end

function Service:_OnNodeEvent(msg, message_source)
end

function Service:_OnSwitchState(new_state)
end

require "ShowBGService"
require "FaceTrackService"
require "FilterService"
require "BeautyService"
require "FaceReshapeService"
require "OccluderService"
require "MultiOutputService"
require "FontService"

return Service