--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================

local ServiceManager = define "ServiceManager"

LOGD("ServiceManager has been defined ...")

require "Service"

local Service = import "Service"
local ShowBGService = import "ShowBGService"
local FaceTrackService = import "FaceTrackService"
local FilterService = import "FilterService"
local BeautyService = import "BeautyService"
local FaceReshapeService = import "FaceReshapeService"
local OccluderService = import "OccluderService"
local MultiOutputService = import "MultiOutputService"
local FontService = import "FontService"

function ServiceManager:Initialize(attach_node)
    self.attach_node = attach_node
    self.node_context = attach_node:GetNodeContext()
    LOGD("ServiceManager initialized ...")
    self:_PrepareServices()

    self:Awake()

    return true
end

function ServiceManager:_PrepareServices()
    self.services = {}
    self.ordered_services = {}

    self.show_bg_service = ShowBGService:New()
    self.services[Service.SERVICE_SHOW_BG] = self.show_bg_service
    self.show_bg_service:Initialize(hl.Named("ShowBGService"), self.attach_node)

    self.face_track_service = FaceTrackService:New()
    self.services[Service.SERVICE_FACE_TRACK] = self.face_track_service
    self.face_track_service:Initialize(hl.Named("FaceTrackService"), self.attach_node)

    self.filter_service = FilterService:New()
    self.services[Service.SERVICE_FILTER] = self.filter_service
    self.filter_service:Initialize(hl.Named("FilterService"), self.attach_node)

    self.beauty_service = BeautyService:New()
    self.services[Service.SERVICE_BEAUTY] = self.beauty_service
    self.beauty_service:Initialize(hl.Named("BeautyService"), self.attach_node)

    self.face_reshape_service = FaceReshapeService:New()
    self.services[Service.SERVICE_FACE_RESHAPE] = self.face_reshape_service
    self.face_reshape_service:Initialize(hl.Named("FaceReshapeService"), self.attach_node)

    self.occluder_service = OccluderService:New()
    self.services[Service.SERVICE_OCCLUDER] = self.occluder_service
    self.occluder_service:Initialize(hl.Named("OccluderService"), self.attach_node)

    self.multi_output_service = MultiOutputService:New()
    self.services[Service.SERVICE_MULTI_OUTPUT] = self.multi_output_service
    self.multi_output_service:Initialize(hl.Named("MultiOutputService"), self.attach_node)

    self.font_service = FontService:New()
    self.services[Service.SERVICE_FONT] = self.font_service
    self.font_service:Initialize(hl.Named("FontService"), self.attach_node)

    self.monitored_services = {[Service.SERVICE_FILTER] = self.filter_service,
                                  [Service.SERVICE_BEAUTY] = self.beauty_service,
                                  [Service.SERVICE_FACE_RESHAPE] = self.face_reshape_service,
                                  [Service.SERVICE_MULTI_OUTPUT] = self.multi_output_service}

    table.insert(self.ordered_services, self.face_track_service)
    table.insert(self.ordered_services, self.beauty_service)
    table.insert(self.ordered_services, self.face_reshape_service)
    table.insert(self.ordered_services, self.filter_service)
    table.insert(self.ordered_services, self.occluder_service)
    table.insert(self.ordered_services, self.multi_output_service)
    table.insert(self.ordered_services, self.show_bg_service)
    table.insert(self.ordered_services, self.font_service)
end

function ServiceManager:SwitchState(new_state)
    for _, service in ipairs(self.ordered_services) do
        service:SwitchState(new_state)
    end
end

function ServiceManager:Awake()
    for _, service in ipairs(self.ordered_services) do
        service:Awake()
    end
end

function ServiceManager:Sleep()
    for _, service in ipairs(self.ordered_services) do
        service:Sleep()
    end
end

function ServiceManager:Shutdown()
    self:Sleep()

    for _, service in ipairs(self.ordered_services) do
        service:Shutdown()
    end

    self.show_bg_service = nil
    self.face_track_service = nil
    self.filter_service = nil
    self.beauty_service = nil
    self.face_reshape_service = nil
    self.occluder_service = nil
    self.multi_output_service = nil

    self.services = nil
    self.ordered_services = nil
end

function ServiceManager:OnEvent(action, msg, message_source)
    local event_consumed = false

    if action == hl.Hl_MSG_SCRIPT_NODE_LOGIC_CONFIGURE_SERVICE then
        local service_type = msg:PopInt32()

        local service = self.services[service_type]
        if service == nil then
            CHECK(false, "ServiceManager: OnEvent: Hl_MSG_SCRIPT_NODE_LOGIC_CONFIGURE_SERVICE: service: id:", service_type, "can not find !!!")
            return false
        end

        service:NodeEvent(msg, message_source)
        event_consumed = true
    end

    return event_consumed
end

function ServiceManager:GetStartedServiceNum()
    local num = 0
    for _, service in ipairs(self.ordered_services) do
        if service:IsStopped() == false then
            num = num + 1
        end
    end

    return num
end

function ServiceManager:GetService(service_enum)
    local service = self.services[service_enum]

    if service == nil then
        CHECK(false, "RootNode: GetService: service:", service_enum, "does not exist !!!")
        return
    end

    return service
end

function ServiceManager:EnableService(service_enum, enable)
    local service = self.services[service_enum]

    if service == nil then
        CHECK(false, "RootNode: EnableService: service:", service_enum, "does not exist !!!")
        return
    end

    service:Enable(enable)

    self:_RebuildServiceChain(service_enum, service, enable)
end

function ServiceManager:_RebuildServiceChain(changed_service_enum, changed_service, enable)
    if self.monitored_services[changed_service_enum] == nil then
        return
    end

    LOGD("RootNode: _RebuildServiceChain: changed_service:", changed_service.name:CStr(),
            "enable:", enable)

    local input_texture = self.node_context.node_manager:GetFixedTexture(
            self.node_context.node_manager.INPUT_TEXTURE_0, true)

    local last_started_service = nil

    self.beauty_service:SetExternalOutputTexture(nil)
    self.face_reshape_service:SetExternalOutputTexture(nil)
    self.filter_service:SetExternalOutputTexture(nil)

    if self.beauty_service:IsStopped() == false then
        self.beauty_service:SetInputTexture(input_texture)
        input_texture = self.beauty_service:GetOutputTexture()
        last_started_service = self.beauty_service
    end

    if self.face_reshape_service:IsStopped() == false then
        self.face_reshape_service:SetInputTexture(input_texture)
        input_texture = self.face_reshape_service:GetOutputTexture()
        last_started_service = self.face_reshape_service
    end

    if self.filter_service:IsStopped() == false then
        self.filter_service:SetInputTexture(input_texture)
        input_texture = self.filter_service:GetOutputTexture()
        last_started_service = self.filter_service
    end

    if self.multi_output_service:IsStopped() then
        if last_started_service == nil then
            self.show_bg_service:SetBGTexture(self.attach_node.node_context.node_manager:GetFixedTexture(
                    self.attach_node.node_context.node_manager.INPUT_TEXTURE_0, true))
        else
            self.show_bg_service:SetBGTexture(last_started_service:GetOutputTexture())
        end
    else
        if last_started_service ~= nil then
            last_started_service:SetExternalOutputTexture(self.node_context.node_manager:GetFixedTexture(
                    self.node_context.node_manager.OUTPUT_TEXTURE_1, false))
            self.multi_output_service:SetNeedRender(false)
        else
            self.multi_output_service:SetNeedRender(true)
        end
    end
end

function ServiceManager:BeginFrame(dt)
    for _, service in ipairs(self.ordered_services) do
        service:Update(dt)
    end
end

function ServiceManager:EndFrame(dt)
end

function ServiceManager:Update(dt)
end

return ServiceManager