--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local View = define "View"

View.kDefaultZ = 0.0

local kNormalBGIndex = 1
local kPressedBGIndex = 2
local kDisabledBGIndex = 3
local kCheckeddBGIndex = 4

View.kNoDraggableFlag = 0
View.kXDraggableFlag = 1
View.kYDraggableFlag = 2
View.kXYDraggableFlag = View.kXDraggableFlag | View.kYDraggableFlag

View.kGravityUndefined = 1
View.kGravityCenter = 2

local max_distance = 1000000

function View:ctor(node, name, window)
    LOGD("View:ctor", self)

    if window == nil then
        window = node.window
    end

    CHECK(name ~= nil, "View: ctor: name is nil !!!");
    CHECK(node ~= nil, "View: ctor: node is nil !!!");
    CHECK(window ~= nil, "View: ctor: window is nil !!!");

    self.window = window

    self.attach_node = node
    self.name = name

    LOGD("View:", self.name, "attach to", self.attach_node:GetName())

    self.name = name

    self.width = 0
    self.height = 0
    self.half_width = 0
    self.half_height = 0
    self.center_x = 0
    self.center_y = 0

    self.alpha = 1

    self.invalidated = true

    self.actor = self.attach_node:AllocActor(hl.Named(self.name .. "_main_actor"))
    self.bg_actor = self.attach_node:AllocActor(hl.Named(self.name .. "_bg_actor"))
    self.fg_actor = self.attach_node:AllocActor(hl.Named(self.name .. "_fg_actor"))

    self.parent = nil

    self.attach_node:AddToWorld(self.actor)
    self.attach_node:AddToWorld(self.bg_actor)
    self.attach_node:AddToWorld(self.fg_actor)
    self.bg_actor:GetTransform():SetParent(self.actor:GetTransform())
    self.fg_actor:GetTransform():SetParent(self.actor:GetTransform())

    self.child_views = {}
    self.has_invalidated_child = false

    self.rotation = 0

    self.touchable = false

    self.intercept_touch = false

    self.touch_targets = {}

    self.enabled = true

    self.downed = false
    self.downed_pointer_id = -1

    self.listener = nil

    self.visible = true
    self.parent_visible = true

    self.background_sprite = nil

    self.renderables = {}

    self.background_textures = {}
    self.foreground_texture = nil
    self.foreground_scale = 1.0

    self.draggable_flag = View.kNoDraggableFlag
    self.draggable_max_x = max_distance
    self.draggable_min_x = -max_distance
    self.draggable_max_y = max_distance
    self.draggable_min_y = -max_distance

    self.last_drag_x_delta = 0.0
    self.last_drag_y_delta = 0.0
    self.last_drag_pointer_id = -1

    self.checked = false

    self.margin_left = 0
    self.margin_right = 0
    self.margin_top = 0
    self.margin_bottom = 0

    self.padding_left = 0
    self.padding_right = 0
    self.padding_top = 0
    self.padding_bottom = 0

    self.background_sprite = hl.Sprite_New()
    local bg_material = hl.Sprite_CreateSimpleMaterial()
    bg_material:SetAlpha(self.alpha)
    local bg_render_state = bg_material:GetRenderState()
    bg_render_state.blend_src = hl.GPBlendMode_kSrcAlpha
    bg_render_state.blend_dst = hl.GPBlendMode_kSrcAlphaInv
    self.background_sprite:SetMaterial(bg_material)
    self.background_sprite:SetWidth(self.width)
    self.background_sprite:SetHeight(self.height)
    self.background_sprite:SetLayer(self.window.renderable_layer)
    self.bg_actor:AddComponent(self.background_sprite)
    self.attach_node:SetRenderLayer(self.background_sprite, self.attach_node.LAYER_OVERLAY_2D)
    table.insert(self.renderables, self.background_sprite)

    self.foreground_sprite = hl.Sprite_New()
    local fg_material = hl.Sprite_CreateSimpleMaterial()
    fg_material:SetAlpha(self.alpha)
    local fg_render_state = fg_material:GetRenderState()
    fg_render_state.blend_src = hl.GPBlendMode_kSrcAlpha
    fg_render_state.blend_dst = hl.GPBlendMode_kSrcAlphaInv
    self.foreground_sprite:SetMaterial(fg_material)
    self.foreground_sprite:SetWidth(self.width * self.foreground_scale)
    self.foreground_sprite:SetHeight(self.height * self.foreground_scale)
    self.foreground_sprite:SetLayer(self.window.renderable_layer)
    self.fg_actor:AddComponent(self.foreground_sprite)
    self.attach_node:SetRenderLayer(self.foreground_sprite, self.attach_node.LAYER_OVERLAY_2D)
    table.insert(self.renderables, self.foreground_sprite)

    self.gravity = View.kGravityUndefined
end

function View:SetGravity(gravity)
    self.gravity = gravity
    self:Invalidate()
end

function View:GetGravity(gravity)
   return self.gravity
end

function View:GetAlpha()
    return self.alpha
end

function View:SetAlpha(alpha)
    self:Invalidate()
    self.alpha = alpha
end

function View:SetChecked(checked)
    self.checked = checked
    self:UpdateBackground()
end

function View:IsChecked()
    return self.checked
end

function View:SetDraggableFlag(draggable_flag)
    self.draggable_flag = draggable_flag
end

function View:__SetParentVisible(parent_visible)
    self.parent_visible = parent_visible

    local viewReallyVisible = self:__IsViewReallyVisible()

    for _, child_view in ipairs(self.child_views) do
        child_view:__SetParentVisible(viewReallyVisible)
    end
end

function View:IsVisible()
    return self.visible
end

function View:SetVisible(visible)
    self.visible = visible

    for _, child_view in ipairs(self.child_views) do
        child_view:__SetParentVisible(self:__IsViewReallyVisible())
    end

    self:Invalidate()
end

function View:OrderSelf(begin_order)
    self.background_sprite:SetOrder(begin_order)
    self.foreground_sprite:SetOrder(begin_order + 1)

    return begin_order + 2
end

function View:Order(begin_order)
    local end_order = self:OrderSelf(begin_order)

    for _, view in ipairs(self.child_views) do
        end_order = view:Order(end_order)
    end

    return end_order
end

function View:SetForegroundTexture(texture)
    self.foreground_texture = texture
    self:Invalidate()
end

function View:SetForegroundScale(scale)
    self.foreground_scale = scale
    self:Invalidate()
end

function View:SetNormalBackgroundTexture(texture)
    --CHECK(texture ~= nil, "View:", self.name, "SetNormalBackgroundTexture: texture is nil !!!")

    self:SetBackgroundTexture(texture, self.background_textures[kPressedBGIndex],
            self.background_textures[kDisabledBGIndex], self.background_textures[kCheckeddBGIndex])
end

function View:SetPressedBackgroundTexture(texture)
    --CHECK(texture ~= nil, "View:", self.name, "SetPressedBackgroundTexture: texture is nil !!!")

    self:SetBackgroundTexture(self.background_textures[kNormalBGIndex], texture,
            self.background_textures[kDisabledBGIndex], self.background_textures[kCheckeddBGIndex])
end

function View:SetDisabledBackgroundTexture(texture)
    --CHECK(texture ~= nil, "View:", self.name, "SetDisabledBackgroundTexture: texture is nil !!!")

    self:SetBackgroundTexture(self.background_textures[kNormalBGIndex], self.background_textures[kPressedBGIndex],
            texture, self.background_textures[kCheckeddBGIndex])
end

function View:SetCheckedBackgroundTexture(texture)
    --CHECK(texture ~= nil, "View:", self.name, "SetCheckedBackgroundTexture: texture is nil !!!")

    self:SetBackgroundTexture(self.background_textures[kNormalBGIndex], self.background_textures[kPressedBGIndex],
            self.background_textures[kDisabledBGIndex], texture)
end

function View:SetBackgroundTexture(normal, pressed, disabled, checked)
    --CHECK(normal ~= nil, "View:", self.name, "SetBackgroundTexture: normal is nil !!!")

    self.background_textures[kNormalBGIndex] = normal
    self.background_textures[kPressedBGIndex] = pressed == nil and normal or pressed
    self.background_textures[kDisabledBGIndex] = disabled == nil and normal or disabled
    self.background_textures[kCheckeddBGIndex] = checked == nil and normal or checked

    self:UpdateBackground()
end

-- TODO: should put in Draw() ...
function View:UpdateBackground()
    if self.background_sprite == nil then
        return
    end

    if not self.enabled and self.background_textures[kDisabledBGIndex] ~= nil then
        self.background_sprite:GetMaterial():SetBaseColorTex(self.background_textures[kDisabledBGIndex])
    elseif self.downed and self.background_textures[kPressedBGIndex] ~= nil then
        self.background_sprite:GetMaterial():SetBaseColorTex(self.background_textures[kPressedBGIndex])
    elseif self.checked and self.background_textures[kCheckeddBGIndex] ~= nil then
        self.background_sprite:GetMaterial():SetBaseColorTex(self.background_textures[kCheckeddBGIndex])
    elseif self.background_textures[kNormalBGIndex] ~= nil then
        self.background_sprite:GetMaterial():SetBaseColorTex(self.background_textures[kNormalBGIndex])
    end

    self:SetVisible(self.visible)
end

function View:Enable(enable)
    self.enabled = enable

    self:UpdateBackground()
end

function View:SetListener(listener)
    self.listener = listener
end

function View:InterceptTouch(intercept)
    self.intercept_touch = intercept
end

function View:SetTouchable(touchable)
    self.touchable = touchable
end

function View:AddChildView(view)
    CHECK(view ~= self, "View:", self.name, "try add self !!!")
    CHECK(view ~= nil, "View:", self.name, "try add nil child view !!!")
    CHECK(view.parent == nil, "View:", view.name, "'s parent is not nil, can not be added !!!")

    if table.find(self.child_views, view) then
        CHECK(false, "View:", self.name, "has added child view:", self.view.name)
    end

    table.insert(self.child_views, view)
    view:OnAdd(self)

    self.window:OrderView()

    self:Invalidate()
end

function View:OnAdd(parent_view)
    self.parent = parent_view
    self.parent_visible = self.parent:IsVisible()
    self.actor:GetTransform():SetParent(parent_view.actor:GetTransform())
end

function View:OnRemove()
    self:SetVisible(false)
    self.actor:GetTransform():SetParent(nil)
    self.parent = nil
end

function View:RemoveChildView(view)
    CHECK(view ~= self, "View:", self.name, "try remove self !!!")
    CHECK(view ~= nil, "View:", self.name, "try remove nil child view !!!")
    CHECK(view.parent == self, "View:", view.name, "'s parent is not View:", self.name,
            "can not remove it !!!")

    if not table.find(self.child_views, view) then
        CHECK(false, "View:", self.name, "has no child view:", self.view.name)
    end

    table.erase(self.child_views, view)
    view:OnRemove()

    self.window:OrderView()

    self:Invalidate()
end

function View:ClearChildView()
    for _, view in ipairs(self.child_views) do
        view:OnRemove()
    end

    self.child_views = {}

    self.window:OrderView()

    self:Invalidate()
end

function View:SetRotation(rotation)
    self.rotation = rotation * hl.kDeg2Rad

    self:Invalidate()
end

function View:SetCenterX(x)
    self.center_x = x
    self:Invalidate()
end

function View:SetCenterY(y)
    self.center_y = y
    self:Invalidate()
end

function View:SetWidth(width)
    self.width = width
    self.half_width = width/2
    self:Invalidate()
end

function View:SetHeight(height)
    self.height = height
    self.half_height = height/2
    self:Invalidate()
end

function View:MarkInvalidated()
    self.invalidated = true
end

function View:Invalidate()
    self.invalidated = true

    local parent = self.parent

    while parent ~= nil do
        if parent.has_invalidated_child then
            break
        end

        parent.has_invalidated_child = true
        parent = parent.parent
    end
end

function View:Layout(dt)
    if (not self.invalidated) and (not self.has_invalidated_child) then
        return
    end

    self:__PreLayout(dt)

    for _, child_view in ipairs(self.child_views) do
        child_view:Layout(dt)
    end

    self:__PostLayout(dt)
end

function View:__PreLayout(dt)
end

function View:__PostLayout(dt)
end

function View:Draw(dt)
    if (not self.invalidated) and (not self.has_invalidated_child) then
        return
    end

    self:__PreDraw(dt)

    for _, child_view in ipairs(self.child_views) do
        child_view:Draw(dt)
    end

    self:__PostDraw(dt)

    self.invalidated = false
    self.has_invalidated_child = false
end

function View:__PreDraw(dt)
end

function View:__IsViewReallyVisible()
    return self.parent_visible and self.visible
end

function View:__PostDraw(dt)
    local viewReallyVisible = self:__IsViewReallyVisible()

    for _, renderable in ipairs(self.renderables) do
        renderable:SetVisible(viewReallyVisible)
    end

    local bg_material = self.background_sprite:GetMaterial()
    if bg_material:GetTexture(bg_material:GetTextureIndex(hl.Material_kBaseColorTex)) == nil then
        self.background_sprite:SetVisible(false)
    else
        self.background_sprite:SetVisible(viewReallyVisible)
    end

    if self.foreground_texture == nil then
        self.foreground_sprite:SetVisible(false)
    else
        self.foreground_sprite:SetVisible(viewReallyVisible)
    end

    self.foreground_sprite:GetMaterial():SetBaseColorTex(self.foreground_texture)

    self.actor:GetTransform():SetLocalPosition(hl.Vec3(self.center_x, self.center_y, View.kDefaultZ))
    self.actor:GetTransform():SetLocalRotation(hl.Quat(self.rotation, hl.Vec3(0, 0, 1)))

    if self.background_sprite ~= nil then
        self.background_sprite:SetWidth(self.width)
        self.background_sprite:SetHeight(self.height)
        self.background_sprite:GetMaterial():SetAlpha(self.alpha)
    end

    if self.foreground_sprite ~= nil then
        self.foreground_sprite:SetWidth(self.width * self.foreground_scale)
        self.foreground_sprite:SetHeight(self.height * self.foreground_scale)
        self.foreground_sprite:GetMaterial():SetAlpha(self.alpha)
    end
end

-- TDOO: current no need handle message directly
--function View:ConvertTouchMessageToEvent(message)
--    local touch_event = {}
--    touch_event.points = {}
--    touch_event.action = message:PopInt32()
--    touch_event.num = message:PopInt32()
--    touch_event.points = {}
--
--    local pointer_id = -1
--    local x = 0.0
--    local y = 0.0
--
--    local half_runtime_width = Hl.RuntimeWidth(hl.kSCREEN_HALF)
--    local half_runtime_height = Hl.RuntimeHeight(hl.kSCREEN_HALF)
--
--    for i = 0, touch_event.num do
--        pointer_id = message:PopInt32()
--        x = message:PopFloat() - half_runtime_width
--        y = half_runtime_height - message:PopFloat()
--        table.insert(touch_event.points, {pointer_id = pointer_id, x = x, y = y})
--    end
--
--    return touch_event
--end
--
--function View:ConvertMouseMessageToEvent(message)
--    local action = message:PopInt32()
--
--    if (hl.Hl_ACTION_MOUSE_ENTER_LEFT ~= action and
--            hl.Hl_ACTION_MOUSE_LEAVE_LEFT ~= action and
--            hl.Hl_ACTION_MOUSE_MOVED ~= action) then
--        return nil
--    end
--
--    local touch_event = {}
--    touch_event.points = {}
--
--    if action == hl.Hl_ACTION_MOUSE_ENTER_LEFT then
--        touch_event.action = hl.Hl_ACTION_TOUCH_BEGAN
--    elseif action == hl.Hl_ACTION_MOUSE_LEAVE_LEFT then
--        touch_event.action = hl.Hl_ACTION_TOUCH_ENDED
--    elseif action == hl.Hl_ACTION_MOUSE_MOVED then
--        touch_event.action = hl.Hl_ACTION_TOUCH_MOVED
--    end
--
--    touch_event.num = 1
--
--    local half_runtime_width = Hl.RuntimeWidth(hl.kSCREEN_HALF)
--    local half_runtime_height = Hl.RuntimeHeight(hl.kSCREEN_HALF)
--
--    local pointer_id = message:PopInt32()
--    local x = message:PopFloat() - half_runtime_width
--    local y = half_runtime_height - message:PopFloat()
--    table.insert(touch_event.points, {pointer_id = pointer_id, x = x, y = y})
--end
--
--function View:OnMessage(message)
--    CHECK(message ~= nil, "View:", self.name, "OnMessage: message is nil !!!")
--
--    local message_id = message:PopBegin()
--
--    local touch_event = nil
--
--    if message_id == hl.HL_MSG_TOUCH then
--        touch_event = self:ConvertTouchMessageToEvent(message)
--    elseif message_id == hl.Hl_MSG_MOUSE then
--        touch_event = self:ConvertMouseMessageToEvent(message)
--    end
--
--    if touch_event ~= nil then
--        self:DispatchTouch(touch_event)
--    end
--end

function View:OnTouchEvent(action, touches)
    local touch_event = {}
    touch_event.points = {}
    touch_event.action = action
    touch_event.num = #touches

    local x = 0.0
    local y = 0.0

    local half_runtime_width = hl.RuntimeWidth(hl.kSCREEN_HALF)
    local half_runtime_height = hl.RuntimeHeight(hl.kSCREEN_HALF)

    for _, touch in ipairs(touches) do
        x = touch.x - half_runtime_width
        y = half_runtime_height - touch.y
        table.insert(touch_event.points, {pointer_id = touch.id, x = x, y = y})
    end

    self:DispatchTouch(touch_event)
end

function View:OnMouseEvent(action, x, y)
    if (hl.Hl_ACTION_MOUSE_ENTER_LEFT ~= action and
            hl.Hl_ACTION_MOUSE_LEAVE_LEFT ~= action and
            hl.Hl_ACTION_MOUSE_MOVED ~= action) then
        return
    end

    local touch_event = {}
    touch_event.points = {}

    if action == hl.Hl_ACTION_MOUSE_ENTER_LEFT then
        touch_event.action = hl.Hl_ACTION_TOUCH_BEGAN
    elseif action == hl.Hl_ACTION_MOUSE_LEAVE_LEFT then
        touch_event.action = hl.Hl_ACTION_TOUCH_ENDED
    elseif action == hl.Hl_ACTION_MOUSE_MOVED then
        touch_event.action = hl.Hl_ACTION_TOUCH_MOVED
    end

    touch_event.num = 1

    local half_runtime_width = hl.RuntimeWidth(hl.kSCREEN_HALF)
    local half_runtime_height = hl.RuntimeHeight(hl.kSCREEN_HALF)

    x = x - half_runtime_width
    y = half_runtime_height - y
    table.insert(touch_event.points, {pointer_id = 0, x = x, y = y})

    self:DispatchTouch(touch_event)
end

function View:PreProcessTouchTargets()
    local delete_touch_target_ids = {}

    for id, touch_target in pairs(self.touch_targets) do
        if touch_target.view.parent ~= self then
            table.insert(delete_touch_target_ids, id)
        else
            touch_target.dispatched = false
        end
    end

    for _, id_to_delete in ipairs(delete_touch_target_ids) do
        self.touch_targets[id_to_delete] = nil
    end
end

function View:FindTouchTargetView(touch_event)
    local view = nil
    for index = #self.child_views, 1, -1 do
        view = self.child_views[index]
        if view:DispatchTouch(touch_event) then
            return view
        end
    end

    return nil
end

function View:DispatchTouch(touch_event)
    self:TransformTouchEvents(touch_event)

    if self.intercept_touch then
        return self:Touch(touch_event)
    end

    local action = touch_event.action

    local pointer_id
    local x, y

    local handled = false
    local handled_by_child = false

    self:PreProcessTouchTargets()

    for _, point in ipairs(touch_event.points) do
        handled_by_child = false

        pointer_id = point.pointer_id
        x = point.x
        y = point.y

        local touch_target = self.touch_targets[pointer_id]

        if action == hl.Hl_ACTION_TOUCH_BEGAN then
            if touch_target ~= nil then
                handled_by_child = touch_target.view:DispatchTouch(touch_event)
                touch_target.dispatched = true
            else
                local view = self:FindTouchTargetView(touch_event)
                if view ~= nil then
                    local new_touch_target = {}
                    new_touch_target.dispatched = true
                    new_touch_target.view = view
                    self.touch_targets[pointer_id] = new_touch_target
                    handled_by_child = true
                end
            end
        elseif action == hl.Hl_ACTION_TOUCH_MOVED then
            if touch_target ~= nil then
                if not touch_target.dispatched then
                    handled_by_child = touch_target.view:DispatchTouch(touch_event)
                    touch_target.dispatched = true
                end
            end
        elseif action == hl.Hl_ACTION_TOUCH_ENDED then
            if touch_target ~= nil then
                if not touch_target.dispatched then
                    handled_by_child = touch_target.view:DispatchTouch(touch_event)
                    touch_target.dispatched = true
                end

                self.touch_targets[pointer_id] = nil
            end
        elseif action == hl.Hl_ACTION_TOUCH_CANCELLED then
            for _, cur_touch_target in pairs(self.touch_targets) do
                if not cur_touch_target.dispatched then
                    handled_by_child = cur_touch_target.view:DispatchTouch(touch_event)
                    cur_touch_target.dispatched = true
                end
            end

            self:ClearTouchTargets()
        else
            CHECK(false, "View:", self.name, "can not handle action", action)
        end

        if handled_by_child then
            handled = true
        else
            if self:Touch(touch_event) then
                handled = true
            end
        end
    end

    self:DeTransformTouchEvents(touch_event)

    return handled
end

function View:ClearTouchTargets()
    self.touch_targets = {}
end

function View:ResetTouchState()
    self.downed = false
    self.downed_pointer_id = -1

    for _, touch_target in pairs(self.touch_targets) do
        touch_target.view:ResetTouchState()
    end

    self:ClearTouchTargets()

    self:UpdateBackground()
end

function View:Touch(touch_event)

    if not self.enabled then
        return false
    end

    if not self.touchable then
        return false
    end

    if not self.visible then
        return false
    end

    local action = touch_event.action
    local num = touch_event.num

    if num <= 0 then
        self:ResetTouchState()
        return false
    end

    local handled = true

    if action == hl.Hl_ACTION_TOUCH_BEGAN then
        if not self.downed then
           local pointer_id = -1

            for _, point in ipairs(touch_event.points) do
                pointer_id = point.pointer_id

                if self:IsPointInView(point.x, point.y) then
                    self.downed_pointer_id = pointer_id
                    self.downed = true

                    self:OnPressed()

                    if self.last_drag_pointer_id == -1 then
                        self:OnDragBegin(point.x, point.y, pointer_id)
                    end
                    --self:ScheduleLongPressCheck();
                end
            end

            if self.downed_pointer_id == -1 then
                handled = false
            end
        end
    elseif action == hl.Hl_ACTION_TOUCH_MOVED then
        for _, point in ipairs(touch_event.points) do
            local pointer_id = point.pointer_id

            if self.last_drag_pointer_id == pointer_id then
                if self.draggable_flag ~= View.kNoDraggableFlag then
                    self:OnDrag(point.x, point.y)
                end
            end

            if self.downed then
                if self.downed_pointer_id == pointer_id then
                    if not self:IsPointInView(point.x, point.y) then
                        self:OnReleased()
                    end
                end
            end
        end
    elseif action == hl.Hl_ACTION_TOUCH_ENDED then
        for _, point in ipairs(touch_event.points) do
            local pointer_id = point.pointer_id

            if self.downed_pointer_id == pointer_id then
                if self.downed then
                    if self:IsPointInView(point.x, point.y) then
                        self:OnClicked()
                    end

                    self:OnReleased()
                end
            end

            if pointer_id == self.last_drag_pointer_id then
                self:OnDragEnd()
            end
        end
    elseif action == hl.Hl_ACTION_TOUCH_CANCELLED then
        if self.downed then
            self:OnReleased()
        end
        self:OnDragEnd()
    else
        CHECK(false, "View:", self.name, "Touch: can not handle action:", action)
    end

    return handled
end

function View:TransformTouchEvents(touch_event)
    local cos = hl.Cos(self.rotation)
    local sin = hl.Sin(self.rotation)

    local x, y
    for _, point in ipairs(touch_event.points) do
        x = point.x
        y = point.y
        point.x = cos*(x - self.center_x) + sin*(y - self.center_y)
        point.y = -sin*(x - self.center_x) + cos*(y - self.center_y)
    end
end

function View:DeTransformTouchEvents(touch_event)
    local cos = hl.Cos(self.rotation)
    local sin = hl.Sin(self.rotation)

    local x, y
    for _, point in ipairs(touch_event.points) do
        x = point.x
        y = point.y
        point.x = cos*(x) - sin*(y) + self.center_x
        point.y = sin*(x) + cos*(y) + self.center_y
    end
end

function View:IsPointInView(x, y)
    return -self.half_width <= x and
            self.half_width >= x and
            -self.half_height <= y and
            self.half_height >= y
end

function View:OnPressed()
    self:UpdateBackground()

    if self.listener and self.listener.OnPressed then
        self.listener:OnPressed(self)
    end
end

function View:OnDrag(cur_x, cur_y)
    local transform = self.actor:GetTransform()
    local cur_pos = transform:GetLocalPosition()

    if self.draggable_flag & View.kXDraggableFlag ~= 0 then
        local new_center_x = cur_pos.x + cur_x - self.last_drag_x_delta
        self.center_x = hl.Clamp(new_center_x, self.draggable_min_x, self.draggable_max_x)
    end

    if self.draggable_flag & View.kYDraggableFlag ~= 0 then
        local new_center_y = cur_pos.y + cur_y - self.last_drag_y_delta
        self.center_y = hl.Clamp(new_center_y, self.draggable_min_y, self.draggable_max_y)
    end

    self.actor:GetTransform():SetLocalPosition(hl.Vec3(self.center_x, self.center_y, View.kDefaultZ))
end

function View:SetHorizontalDragRange(min_x, max_x)
    self.draggable_min_x = min_x
    self.draggable_max_x = max_x
end

function View:SetVerticalDragRange(min_y, max_y)
    self.draggable_min_y = min_y
    self.draggable_max_y = max_y
end

function View:ResetDragState()
    self.last_drag_x_delta = 0
    self.last_drag_y_delta = 0
    self.last_drag_pointer_id = -1
end

function View:OnDragBegin(begin_x, begin_y, pointer_id)
    self.last_drag_x_delta = begin_x
    self.last_drag_y_delta = begin_y
    self.last_drag_pointer_id = pointer_id

    if self.listener and self.listener.OnDragBegin then
        self.listener:OnDragBegin(self)
    end
end

function View:OnDragEnd()
    self:ResetDragState()

    if self.listener and self.listener.OnDragEnd then
        self.listener:OnDragEnd(self)
    end
end

function View:OnReleased()
    self:ResetTouchState()

    if self.listener and self.listener.OnReleased then
        self.listener:OnReleased(self)
    end
end

function View:OnClicked()
    if self.listener and self.listener.OnClicked then
        self.listener:OnClicked(self)
    end
end

return View
