--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local VideoPlayer = define ("VideoPlayer")

VideoPlayer.kFitModeNone = 0
VideoPlayer.kFitModeKeepWidth = 1
VideoPlayer.kFitModeKeepHeight = 2
VideoPlayer.kFitModeScale = 3
VideoPlayer.kFitModeFull = 4

VideoPlayer.kInvalidWidth = -1
VideoPlayer.kInvalidHeight = -1

VideoPlayer.kSecond2Milli = 1000

local VideoPlayerPlayID = -1
function VideoPlayer:Init(node, name)
    LOGD("VideoPlayer ", self.name, "Init ...")
    self.play_id = -1
    CHECK(node ~= nil, "VideoPlayer: Init: node is nil !!!")
    self.attach_node = node
    self.node_manager = self.attach_node.node_context.node_manager
    if name == nil then
        self.name = "video_player_" .. node:GetName()
    else
        self.name = name
    end

    self.video_frame_texture = self.node_manager:GetFixedTexture(
            self.node_manager.INPUT_TEXTURE_2, true)

    self.frame_original_width = VideoPlayer.kInvalidWidth
    self.frame_original_height = VideoPlayer.kInvalidHeight

    self.fit_mode = VideoPlayer.kFitModeKeepHeight

    self.listeners = {}

    self.paused = false

    self:_AddListener(hl.Hl_MSG_SCRIPT_PLAYER_ACTION, self._OnPlayAction, self)
    self:_AddListener(hl.Hl_MSG_SCRIPT_PLAYER_EVENT, self._OnPlayEvent, self)
    self:_AddListener(hl.Hl_MSG_RUNTIME_RESET, self._OnRuntimeReset, self)

    self.play_event_listeners = { __mode = "k" }

    self.duration = 0

    self:_InitUI()
end

function VideoPlayer:GetDuration()
    return self.duration
end

function VideoPlayer:SetFitMode(fit_mode)
    self.fit_mode = fit_mode
    self:_CalculateFrameSize(self.frame_original_width, self.frame_original_height)
end

function VideoPlayer:_OnRuntimeReset()
    self:_CalculateFrameSize(self.frame_original_width, self.frame_original_height)
end

function VideoPlayer:SetLayer(layer)
    self.layer = layer
    self.video_frame_sprite.bg_sprite:SetLayer(self.layer)
end

function VideoPlayer:_InitUI()
    self.layer = -1000
    self.root_actor = self.attach_node:AllocActor(hl.Named(self.name .. "_root_actor"))
    self.attach_node:AddToWorld(self.root_actor)

    self.video_frame_actor = self.attach_node:AllocActor(hl.Named(self.name .. "_video_frame_actor"))
    self.attach_node:AddToWorld(self.video_frame_actor)
    self.video_frame_actor:GetTransform():SetParent(self.root_actor:GetTransform())

    self.video_frame_sprite = hl.Sprite_New()
    local material = hl.Sprite_CreateSimpleMaterial()
    self.video_frame_sprite:SetMaterial(material)
    self.video_frame_sprite:SetWidth(0)
    self.video_frame_sprite:SetHeight(0)
    self.video_frame_sprite:SetVisible(true)
    self.video_frame_sprite:SetLayer(self.layer)
    self.video_frame_actor:AddComponent(self.video_frame_sprite)
    self.attach_node:SetRenderLayer(self.video_frame_sprite, self.attach_node.LAYER_OVERLAY_2D)
    self.video_frame_sprite:GetMaterial():SetBaseColorTex(self.video_frame_texture)

    self:_CalculateFrameSize(self.frame_original_width, self.frame_original_height)
end

function VideoPlayer:AddListener(listener)
    self.play_event_listeners[listener] = listener
end

function VideoPlayer:ClearListener()
    self.play_event_listeners = {}
end

function VideoPlayer:RemoveListener(listener)
    self.play_event_listeners[listener] = nil
end

function VideoPlayer:_OnPlayEvent(play_id, event, msg)
    LOGD("VideoPlayer:", self.name, "_OnPlayEvent: event:", event, "play id:", play_id)

    if self.play_id ~= play_id then
        LOGD("VideoPlayer:", self.name, "_OnPlayEvent: event:", event, "play_id:", play_id,
            " != cur play id", self.play_id, "just ignore it ...")
        return
    end

    if event == hl.Hl_MSG_SCRIPT_PLAY_BEGIN then
        -- FIXME: Play begin message is a little delayed, we detect it by ourself
        --self:_OnPlayBegin(play_id)
    elseif event == hl.Hl_MSG_SCRIPT_PLAY_END then
        self:_OnPlayEnd(play_id)
    elseif event == hl.Hl_MSG_SCRIPT_PLAY_FAIL then
        local error_desc = msg:PopInt32()
        self:_OnPlayError(play_id, error_desc)
    elseif event == hl.Hl_MSG_SCRIPT_PLAY_UPDATE then
        local pts = msg:PopDouble()
        local video_frame_width = msg:PopInt32()
        local video_frame_height = msg:PopInt32()
        local video_duration = msg:PopInt64()
        -- TODO: Not render
        self:_OnPlayBegin(play_id, video_frame_width, video_frame_height, video_duration)
        self:_OnPlayUpdate(play_id, pts)
    else
        CHECK(false, "VideoPlayer:", self.name, "_OnPlayEvent: can not handle event:", event, "play_id:", play_id)
    end
end

function VideoPlayer:_OnPlayBegin(play_id, video_frame_width, video_frame_height, video_duration)
    if self.wait_first_video_frame then
        self.frame_original_width = video_frame_width
        self.frame_original_height = video_frame_height
        self:_CalculateFrameSize(video_frame_width, video_frame_height)

        self.duration = video_duration

        LOGD("VideoPlayer:", self.name, "play begin: video width:", video_frame_width, "video height:", video_frame_height, "duration:", self.duration)

        for _, listener in pairs(self.play_event_listeners) do
            if listener.onPlayBegin then
                listener.onPlayBegin(play_id, self.duration)
            end
        end

        self.wait_first_video_frame = false
    end
end

function VideoPlayer:_CalculateFrameSize(video_frame_width, video_frame_height)
    local runtime_width = hl.RuntimeWidth(hl.kSCREEN_EQUAL)
    local runtime_height = hl.RuntimeHeight(hl.kSCREEN_EQUAL)

    LOGD("VideoPlayer:", self.name, "_CalculateFrameSize:", video_frame_width, video_frame_height,
            runtime_width, runtime_height, "fit mode:", self.fit_mode)

    if video_frame_width == VideoPlayer.kInvalidWidth or
            video_frame_height == VideoPlayer.kInvalidHeight then
        self.video_frame_sprite:SetWidth(0)
        self.video_frame_sprite:SetHeight(0)
        return
    end

    if self.fit_mode == VideoPlayer.kFitModeNone then

    elseif self.fit_mode == VideoPlayer.kFitModeKeepWidth then
        local scale_ratio = runtime_width / video_frame_width
        video_frame_width = runtime_width
        video_frame_height = scale_ratio * video_frame_height
    elseif self.fit_mode == VideoPlayer.kFitModeKeepHeight then
        local scale_ratio = runtime_height / video_frame_height
        video_frame_height = runtime_height
        video_frame_width = scale_ratio * video_frame_width
    elseif self.fit_mode == VideoPlayer.kFitModeScale then
        if video_frame_width > video_frame_height then
            local scale_ratio = runtime_width / video_frame_width
            video_frame_width = runtime_width
            video_frame_height = scale_ratio * video_frame_height
        else
            local scale_ratio = runtime_height / video_frame_height
            video_frame_height = runtime_height
            video_frame_width = scale_ratio * video_frame_width
        end
    elseif self.fit_mode == VideoPlayer.kFitModeFull then
        video_frame_width = runtime_width
        video_frame_height = runtime_height
    else
        CHECK(false, "VideoPlayer:", self.name, "_CalculateFrameSize: invalid fit mode: ", self.fit_mode, "!!!")
    end

    LOGD("VideoPlayer:", self.name, "set width/height to:", video_frame_width, video_frame_height)

    self.video_frame_sprite:SetWidth(video_frame_width)
    self.video_frame_sprite:SetHeight(video_frame_height)
end

function VideoPlayer:_OnPlayEnd(play_id)
    for _, listener in pairs(self.play_event_listeners) do
        if listener.onPlayEnd then
            listener.onPlayEnd(play_id, self.duration)
        end
    end
    self.duration = 0
end

function VideoPlayer:_OnPlayError(play_id, error)
    for _, listener in pairs(self.play_event_listeners) do
        if listener.onPlayFail then
            listener.onPlayFail(play_id, error, self.duration)
        end
    end
    self.duration = 0
end

function VideoPlayer:_OnPlayUpdate(play_id, pts)
    for _, listener in pairs(self.play_event_listeners) do
        if listener.onUpdate then
            listener.onUpdate(play_id, pts * VideoPlayer.kSecond2Milli, self.duration)
        end
    end
end

function VideoPlayer:_OnPlayAction(action, msg)
    LOGD("VideoPlayer:", self.name, "_OnPlayAction:", action)
    if action == hl.Hl_MSG_SCRIPT_PLAYER_PLAY then
        local video_url = msg:PopString()
        local loop = msg:PopInt8()
        local begin = msg:PopFloat()
        local ended = msg:PopFloat()
        if loop == 1 then
            self:Play(video_url, true, begin, ended)
        else
            self:Play(video_url, false, begin, ended)
        end
    elseif action == hl.Hl_MSG_SCRIPT_PLAYER_PAUSE then
        if hl.IsPlatformAndroid() then
            local play_id = msg:PopInt32()
            if play_id == self.play_id then
                self:Pause()
            else
                LOGI("VideoPlayer:", self.name, "_OnPlayAction pause, play id is",
                        play_id, "but video player's cur play id is", self.play_id, "just ignore ...")
            end
        else
            self:Pause()
        end
    elseif action == hl.Hl_MSG_SCRIPT_PLAYER_RESUME then
        if hl.IsPlatformAndroid() then
            local play_id = msg:PopInt32()
            if play_id == self.play_id then
                self:Resume()
            else
                LOGI("VideoPlayer:", self.name, "_OnPlayAction resume, play id is",
                        play_id, "but video player's cur play id is", self.play_id, "just ignore ...")
            end
        else
            self:Resume()
        end
    elseif action == hl.Hl_MSG_SCRIPT_PLAYER_SEEK then
        if hl.IsPlatformAndroid() then
            local play_id = msg:PopInt32()
            if play_id == self.play_id then
                local seek_pos = msg:PopDouble()
                self:Seek(seek_pos)
            else
                LOGI("VideoPlayer:", self.name, "_OnPlayAction seek, play id is",
                play_id, "but video player's cur play id is", self.play_id, "just ignore ...")
            end
        else
            local seek_pos = msg:PopDouble()
            self:Seek(seek_pos)
        end

    elseif action == hl.Hl_MSG_SCRIPT_PLAYER_STOP then
        if hl.IsPlatformAndroid() then
            local play_id = msg:PopInt32()
            if play_id == self.play_id then
                self:Stop()
            else
                LOGI("VideoPlayer:", self.name, "_OnPlayAction stop, play id is",
                play_id, "but video player's cur play id is", self.play_id, "just ignore ...")
            end
        else
            self:Stop()
        end
    else
        CHECK(false, "VideoPlayer:", self.name, "_OnPlayAction: can not handle action:", action)
    end
end

function VideoPlayer:Shutdown()
    LOGD("VideoPlayer ", self.name, "Shutdown ...")
    self:Stop()
    self:_ReleaseListeners()
    self:ClearListener()
    self.attach_node:RemoveFromWorld(self.video_frame_actor)
    self.attach_node:RemoveFromWorld(self.root_actor)
    LOGD("VideoPlayer Shutdown Finish.")
end

function VideoPlayer:_AddListener(event, listener, holder)
    if  self.listeners[event] ~= nil then
        CHECK(false, "VideoPlayer ", self.name, "event", event, "listener already added !!!")
        return
    end

    LOGD("VideoPlayer ", self.name, "event", event, "listener added ...")
    self.listeners[event] = listener

    self.attach_node.node_context.game_framework:AddListener(event, listener, holder)
end

function VideoPlayer:_RemoveListener(event)
    if self.listeners[event] == nil then
        CHECK(false, "VideoPlayer ", self.name, "event", event, "listener not added yet !!!")
        return
    end

    LOGD("VideoPlayer ", self.name, "event", event, "listener removed ...")

    self.attach_node.node_context.game_framework:RemoveListener(event, self.listeners[event])
    self.listeners[event] = nil
end

function VideoPlayer:_ReleaseListeners()
    for event, listener in pairs(self.listeners) do
        if listener ~= nil then
            LOGD("VideoPlayer ", self.name," remove event", event, "listener")
            self.attach_node.node_context.game_framework:RemoveListener(event, listener)
        end
    end

    self.listeners = nil
end

function VideoPlayer:Play(video_url, loop, begin, ended)
    LOGD("VideoPlayer:", self.name, "Play:", video_url, loop, self.play_id)
    if loop == nil or loop == false then
        loop = 0
    else
        loop = 1
    end
    begin = begin or 0
    ended = ended or - 1
    VideoPlayerPlayID = VideoPlayerPlayID + 1
    self.play_id = VideoPlayerPlayID
    local msg = hl.MessageQueue_MakeAsync()
    msg:PushBegin(hl.Hl_MSG_NATIVE_PLAYER_ACTION)
    msg:PushInt32(self.play_id)
    msg:PushInt32(hl.Hl_MSG_NATIVE_PLAYER_PLAY)
    msg:PushString(video_url)
    msg:PushInt8(loop)
    msg:PushFloat(begin)
    msg:PushFloat(ended)
    msg:PushEnd()
    hl.MessageQueue_PostAndFreeAsync(msg)
    self.paused = false
    self.wait_first_video_frame = true
    self.begin = begin
    self.ended = ended

    return self.play_id
end

function VideoPlayer:IsPaused()
    return self.paused
end

function VideoPlayer:Pause()
    LOGD("VideoPlayer:", self.name, "Pause ...")
    local msg = hl.MessageQueue_MakeAsync()
    msg:PushBegin(hl.Hl_MSG_NATIVE_PLAYER_ACTION)
    msg:PushInt32(self.play_id)
    msg:PushInt32(hl.Hl_MSG_NATIVE_PLAYER_PAUSE)
    msg:PushEnd()
    hl.MessageQueue_PostAndFreeAsync(msg)
    self.paused = true
end

function VideoPlayer:Resume()
    LOGD("VideoPlayer:", self.name, "Resume ...")
    local msg = hl.MessageQueue_MakeAsync()
    msg:PushBegin(hl.Hl_MSG_NATIVE_PLAYER_ACTION)
    msg:PushInt32(self.play_id)
    msg:PushInt32(hl.Hl_MSG_NATIVE_PLAYER_RESUME)
    msg:PushEnd()
    hl.MessageQueue_PostAndFreeAsync(msg)
    self.paused = false
end

function VideoPlayer:Stop()
    LOGD("VideoPlayer:", self.name, "Stop ...")
    local msg = hl.MessageQueue_MakeAsync()
    msg:PushBegin(hl.Hl_MSG_NATIVE_PLAYER_ACTION)
    msg:PushInt32(self.play_id)
    msg:PushInt32(hl.Hl_MSG_NATIVE_PLAYER_STOP)
    msg:PushEnd()
    hl.MessageQueue_PostAndFreeAsync(msg)
end

function VideoPlayer:Seek(seek_pos)
    LOGD("VideoPlayer:", self.name, "Seek", seek_pos)
    local msg = hl.MessageQueue_MakeAsync()
    msg:PushBegin(hl.Hl_MSG_NATIVE_PLAYER_ACTION)
    msg:PushInt32(self.play_id)
    msg:PushInt32(hl.Hl_MSG_NATIVE_PLAYER_SEEK)
    msg:PushDouble(seek_pos)
    msg:PushEnd()
    hl.MessageQueue_PostAndFreeAsync(msg)
end

return VideoPlayer