--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local Viewer = define "Viewer"
local GameFramework = import "GameFramework"

function Viewer:Initialize(camera)
	assert(camera)
	self.eye = camera
	self.rot_h = hl.Quat()
	self.rot_v = hl.Quat()
	self.pos_h = 0
	self.pos_v = 0
	self.width = 0
	self.height = 0
	self.begin_right_x = 0
	self.begin_right_y = 0
	self.last_right_x = 0
	self.last_right_y = 0
	self.dp = 0.2
	self.dq = -0.5 / hl.RuntimeHeight(hl.kSCREEN_EQUAL) * hl.kPi
	self.click_right = false
end

function Viewer:Shutdown()
end

function Viewer:Awake()
	GameFramework:AddListener(hl.Hl_MSG_RUNTIME_RESET, self.Reset, self)
	GameFramework:AddListener(hl.Hl_MSG_INPUT_KEY, self.KeyEvent, self)
	GameFramework:AddListener(hl.Hl_MSG_MOUSE, self.MouseEvent, self)
end

function Viewer:Sleep()
	GameFramework:RemoveListener(hl.Hl_MSG_RUNTIME_RESET, self.Reset)
	GameFramework:RemoveListener(hl.Hl_MSG_INPUT_KEY, self.KeyEvent)
	GameFramework:RemoveListener(hl.Hl_MSG_MOUSE, self.MouseEvent)
end

function Viewer:SetEye(camera)
	self.eye = camera
	if self.eye then
		self.eye:GetActor():GetTransform():SetGlobalRotation(self.rot_h * self.rot_v)
    end
end
    
function Viewer:Update(dt)
	if self.click_right and (self.pos_h ~= 0 or self.pos_v ~= 0) then
		local trans = self.eye:GetActor():GetTransform()
		local pos = trans:GetGlobalPosition()
		local rot = trans:GetGlobalRotation()
		trans:SetGlobalPosition(pos + rot * hl.Vec3(self.dp * self.pos_h, 0.0, self.dp * self.pos_v))
	end
end

function Viewer:MouseEvent(action, x, y)
	if action == hl.Hl_ACTION_MOUSE_ENTER_RIGHT then
		self:_MouseEnterRight(x, y)
	elseif action == hl.Hl_ACTION_MOUSE_LEAVE_RIGHT then
		self:_MouseLeaveRight(x, y)
	elseif action == hl.Hl_ACTION_MOUSE_MOVED then
		self:_MouseMoved(x, y)
	end
end
    
function Viewer:Reset(w, h)
	if w > 0 and h > 0 then
		self.width = w
		self.height = h
		self.dq = -0.5 / self.height * hl.kPi
    end
end

function Viewer:KeyEvent(action, key)
	if action == hl.Hl_ACTION_KEY_PRESS then
		self:_KeyPress(key)
	elseif action == hl.Hl_ACTION_KEY_RELEASE then
		self:_KeyRelease(key)
	end
end
    
function Viewer:_KeyPress(input)
	if input == hl.Hl_INPUT_KEY_W then
		self.pos_v = self.pos_v - 1
	elseif input == hl.Hl_INPUT_KEY_S then
        self.pos_v = self.pos_v + 1
	elseif input == hl.Hl_INPUT_KEY_A then
        self.pos_h = self.pos_h - 1
	elseif input == hl.Hl_INPUT_KEY_D then
		self.pos_h = self.pos_h + 1
	end
end

function Viewer:_KeyRelease(input)
    if input == hl.Hl_INPUT_KEY_W then
		self.pos_v = self.pos_v + 1
	elseif input == hl.Hl_INPUT_KEY_S then
        self.pos_v = self.pos_v - 1
	elseif input == hl.Hl_INPUT_KEY_A then
        self.pos_h = self.pos_h + 1
	elseif input == hl.Hl_INPUT_KEY_D then
        self.pos_h = self.pos_h - 1
	end
end

function Viewer:_MouseEnterRight(x, y)
	self.click_right = true
	self.begin_right_x = x
	self.begin_right_y = y
	self.last_right_x = x
	self.last_right_y = y
end
    
function Viewer:_MouseLeaveRight(x, y)
	self.click_right = false
end

function Viewer:_MouseMoved(x, y)
	if self.eye and self.click_right then
		local move_x = x - self.last_right_x
		local move_y = y - self.last_right_y
		self.rot_h = self.rot_h:Rotate(self.dq * move_x, hl.Vec3_kAxisY)
		self.rot_v = self.rot_v:Rotate(self.dq * move_y, hl.Vec3_kAxisX)
		self.eye:GetActor():GetTransform():SetGlobalRotation(self.rot_h * self.rot_v)
		self.last_right_x = x
		self.last_right_y = y
	end
end

return Viewer