--==================================================================================
--Copyright (c) 2018-2019 Hulua Group. All rights reserved.
--==================================================================================
local Viewer = define "Viewer"
local GameFramework = import "GameFramework"
local ScaleGestureDetector = import "ScaleGestureDetector"

function Viewer:Initialize(camera)
	assert(camera)
	self.eye = camera
	self.rot_h = hl.Quat()
	self.rot_v = hl.Quat()
	self.width = 0
	self.height = 0
	self.prevX = 0
	self.prevY = 0
	self.dp = 5.0
	self.dq = -0.5 / hl.RuntimeHeight(hl.kSCREEN_EQUAL) * hl.kPi
	self.single_touch = false
	self.scale_touch = false
	self.sgd = ScaleGestureDetector:New()
	self.sgd:Initialize(self)


end

function Viewer:Awake()
	GameFramework:AddListener(hl.Hl_MSG_RUNTIME_RESET, self.Reset, self)
	GameFramework:AddListener(hl.Hl_MSG_TOUCH, self.TouchEvent, self)
end

function Viewer:Sleep()
	GameFramework:RemoveListener(hl.Hl_MSG_RUNTIME_RESET, self.Reset)
	GameFramework:RemoveListener(hl.Hl_MSG_TOUCH, self.TouchEvent)
end

function Viewer:Shutdown()
	self.sgd:Shutdown()
end
    
function Viewer:Update(dt)
end
    
function Viewer:Reset(w, h)
	if w > 0 and h > 0 then
		self.width = w
		self.height = h
		self.dq = -0.5 / self.height * hl.kPi
    end
end

function Viewer:TouchEvent(action, touches)
	self.sgd:TouchEvent(action, touches)
	if action == hl.Hl_ACTION_TOUCH_BEGAN then
		self:_TouchBegan(touches)
	elseif action == hl.Hl_ACTION_TOUCH_MOVED then
		self:_TouchMoved(touches)
	elseif action == hl.Hl_ACTION_TOUCH_ENDED then
		self:_TouchEnded(touches)
	end
end

function Viewer:_TouchBegan(touches)
	if not self.scale_touch then
		for _, touch in ipairs(touches) do
			if touch.id == 0 then
				self.single_touch = true
				self.prevX = touch.x
				self.prevY = touch.y
				break
			end
		end
	end
end

function Viewer:_TouchMoved(touches)
	if self.eye and self.single_touch and not self.scale_touch then
		local x, y
		for _, touch in ipairs(touches) do
			if touch.id == 0 then
				x = touch.x
				y = touch.y
				break
			end
		end
		local move_x = x - self.prevX
		local move_y = y - self.prevY
		self.rot_h = self.rot_h:Rotate(self.dq * move_x, hl.Vec3_kAxisY)
		self.rot_v = self.rot_v:Rotate(self.dq * move_y, hl.Vec3_kAxisX)
		assert(self.eye)
		assert(self.eye:GetActor())
		assert(self.eye:GetActor():GetTransform())
		self.eye:GetActor():GetTransform():SetGlobalRotation(self.rot_h * self.rot_v)
		self.prevX = x
		self.prevY = y
	end
end

function Viewer:_TouchEnded(touches)
	if self.single_touch then
		for _, touch in ipairs(touches) do
			if touch.id == 0 then
				self.single_touch = false
				break
			end
		end
	end
end
    
function Viewer:OnScaleBegin(sgd)
	self.single_touch = false
	self.scale_touch = true
	return true
end

function Viewer:OnScaleEnd(sgd)
	self.scale_touch = false
end

function Viewer:OnScale(sgd)
	local scale = sgd:GetScaleFactor()
	local dis = 1.0 - scale
	local trans = self.eye:GetActor():GetTransform()
	local pos = trans:GetGlobalPosition()
	local rot = trans:GetGlobalRotation()
	trans:SetGlobalPosition(pos + rot * hl.Vec3(0.0, 0.0, self.dp * dis))
end

return Viewer